/*
 *  AppWindow  -- The main application document window for the AeroLS program.
 *
 *   Copyright (C) 2002-2024 by Joseph A. Huwaldt
 *   All rights reserved.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */

import jahuwaldt.aero.panelgeom.GeomComponent;
import jahuwaldt.aero.panelgeom.GeomElement;
import jahuwaldt.io.ExtFilenameFilter;
import jahuwaldt.io.FileUtils;
import jahuwaldt.js.ParamTools.ParamDatabase;
import jahuwaldt.js.ParamTools.ParamException;
import jahuwaldt.js.ParamTools.ParamInput;
import jahuwaldt.swing.*;
import jahuwaldt.tools.math.MathTools;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.MalformedURLException;
import javax.swing.*;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

/**
 * The main document window for the AeroLS lifting surface aerodynamic analysis
 * program. This window displays the geometry and allows the user to specify the
 * analysis that is to be performed.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: May 14, 2002
 * @version January 21, 2024
 */
public class AppWindow extends JFrame implements ParamNames {

    //	Initial application frame width & height.
	private static final short	FRM_WIDTH = 750;
	private static final short	FRM_HEIGHT = 600;

	// Definition of this application's menus.
	private static final
		String[][] FILEM_STRINGS ={		{ "New", "N", "handleNew" },
										{ "Open...", "O", "handleOpen" },
										{ null, null, null },
										{ "Close", "W", "handleClose" },
										{ "Save", "S", "handleSave" },
										{ "Save As...", null, "handleSaveAs" } };

	private static final
		String[][] EDITM_STRINGS = { 	{ "Undo", "Z", null },
										{ null, null, null },
										{ "Cut", "X", null },
										{ "Copy", "C", null },
										{ "Paste", "V", null } };

	private static final
		String[][] ANALM_STRINGS = { 	{ "Test Conditions", null, "handleTC" },
										{ null, null, null },
										{ "Analyze", null, "handleAnalyze" } };


	//	The extension used for data files.
	private static final String EXTENSION = ".alsx";

	//  The data used by this program.
	private ProjectData data;

    //  The panel displaying model information in the main window.
	private final InfoPanel infoPanel;

    //  The list view for our component list.
	private JList compList;

    //  The list model for our list of components.
	private ComponentListModel compListModel;

	//  The list view for our segment list.
	private JList segList;

	//  The list model for our list of segments in each component.
	private SegmentListModel segListModel;

	//  A list of view panels.
	private final ViewPanel[] views = new ViewPanel[3];

	//  A count of the number of windows that have been opened.
	private static int windowCount = 0;

	/**
	*  A reference to the file that this data is stored in.
	*/
	private File fileRef;


    /**
    *  Construct a new AeroLS application document window with the given title.
    */
    @SuppressWarnings("OverridableMethodCallInConstructor")
    public AppWindow() throws ParamException, Exception {
        super( "Untitled" + (windowCount > 0 ? " " + String.valueOf(windowCount) : "") );

        //  Define the data storage and intial values used by this program.
        data = new ProjectData();

		//	Set up this frame.
		this.setSize(FRM_WIDTH, FRM_HEIGHT);
		this.setResizable( true );

		//  Have the window dispose of itself when it closes.
		this.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

        //  Lay out this window's contents.
		Container cp = this.getContentPane();
		cp.setLayout(new BorderLayout());

		//  Create a tool bar for our window.
		JToolBar toolbar = new JToolBar();

		//  Create a references button and add to tool bar.
		ButtonL theBL = new ButtonL();
		JButton button = new JButton("Set Refs");
		button.addActionListener( theBL );
		button.setActionCommand( "References" );
		button.setToolTipText( "Edit aerodynamic reference quantities" );
		toolbar.add(button);

		//  Create add component button and add to tool bar.
		button = new JButton("Add Comp");
		button.addActionListener( theBL );
		button.setActionCommand( "Add Component" );
		button.setToolTipText( "Creates a new component for this vehicle" );
		toolbar.add(button);

		//  Create delete component button and add to tool bar.
		button = new JButton("Del Comp");
		button.addActionListener( theBL );
		button.setActionCommand( "Del Component" );
		button.setToolTipText( "Choose a component to delete or remove" );
		toolbar.add(button);

		//  Create add wing segment button and add to tool bar.
		button = new JButton("Add Wing Seg");
		button.addActionListener( theBL );
		button.setActionCommand( "Add Wing Segment" );
		button.setToolTipText( "Creates a new wing segment or network" );
		toolbar.add(button);

		//  Create delete wing segment button and add to tool bar.
		button = new JButton("Del Segment");
		button.addActionListener( theBL );
		button.setActionCommand( "Del Segment" );
		button.setToolTipText( "Choose a segment to delete or remove" );
		toolbar.add(button);

		//  Create zoom in button and add to tool bar.
		button = new JButton("+Zoom");
		button.addActionListener( theBL );
		button.setActionCommand( "Zoom In" );
		button.setToolTipText( "Increases the 3-view scale." );
		toolbar.add(button);

		//  Create zoom out button and add to tool bar.
		button = new JButton("-Zoom");
		button.addActionListener( theBL );
		button.setActionCommand( "Zoom Out" );
		button.setToolTipText( "Decreases the 3-view scale." );
		toolbar.add(button);

		//  Add the toolbar to the window's content pane.
		cp.add(toolbar, BorderLayout.NORTH);


		//  Create the display area panel.
		JPanel dispPanel = new JPanel(new BorderLayout());
		cp.add(dispPanel, BorderLayout.CENTER);

        //  Create a panel to contain the various lists the user can choose from.
		Box listBox = Box.createHorizontalBox();
        dispPanel.add(listBox, BorderLayout.NORTH);

        //  Create panel that displays a list of components.
        JPanel compList1 = createComponentListPanel();
        listBox.add(compList1);

        //  Create a panel that displays a list of segments in each component.
        compList1 = createSegmentListPanel();
        listBox.add(compList1);
        listBox.add(Box.createHorizontalStrut(200));
        listBox.add(Box.createGlue());

		//  Create the panel that will contain all of our geometry views.
		JPanel viewPanel = new JPanel(new GridLayout(2,2));
		dispPanel.add(viewPanel, BorderLayout.CENTER);

		//  Create each "view" panel.
        infoPanel = new InfoPanel();
        viewPanel.add(infoPanel);

        //  Create a top view.
        views[0] = new TopViewPanel(data.geometry);
		viewPanel.add(views[0]);

		//	Create a front view.
		views[1] = new FrontViewPanel(data.geometry);
		viewPanel.add(views[1]);

		//	Create a side view.
		views[2] = new SideViewPanel(data.geometry);
		viewPanel.add(views[2]);

		//  Create a menu bar for this window.
		JMenuBar menuBar = createMenuBar();
		this.setJMenuBar(menuBar);

		//  Position the window so that each new window can be seen.
		AppUtilities.positionWindow(this, FRM_WIDTH, FRM_HEIGHT );

		//  Increment the window count.
		++windowCount;
    }


    /**
     * Sets the title for this frame to the specified string. Also notifies the
     * main application class so that the "Windows" menu gets updated too.
     *
     * @param title The title to be displayed in the frame's border. A null
     * value is treated as an empty string, "".
     */
    @Override
    public void setTitle(String title) {
        super.setTitle(title);
        MDIApplication.windowTitleChanged(this);
    }

	/**
	*  Initializes the menus associated with this window.
	*/
	private JMenuBar createMenuBar() throws NoSuchMethodException {

		JMenuBar menuBar = new JMenuBar();

		// Set up the file menu.
		JMenu menu = AppUtilities.buildMenu( this, "File", FILEM_STRINGS );
		menuBar.add( menu );

        //  Add a Quit menu item (if it isn't automatically present).
        if (!QuitJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(MDIApplication.getInstance().getQuitJMenuItem());
        }

		// Set up the edit menu.
		menu = AppUtilities.buildMenu( this, "Edit", EDITM_STRINGS );
		menuBar.add( menu );

		// Set up the Analysis menu.
		menu = AppUtilities.buildMenu( this, "Analysis", ANALM_STRINGS );
		menuBar.add( menu );

		//  Create a Window's menu.
		menu = MDIApplication.newWindowsMenu("Windows");
		menuBar.add(menu);
        
        //  Create an about menu item (if not automatically present).
        if (!AboutJMenuItem.isAutomaticallyPresent()) {
            //  Create Help menu.
            menu = new JMenu("Help");
            menuBar.add(menu);
            
            //  Add the "About" item to the Help menu.
            menu.add(MDIApplication.getInstance().getAboutJMenuItem());
        }
        
		return menuBar;
	}



    /**
    *  Method that updates all the displays in this program
    *  to reflect current model information.
    */
    void update() {
        //  Update the info panel region.
        infoPanel.update();

        //  Update our segment list.
        int idx = compList.getSelectedIndex();
        compList.clearSelection();
        compList.setSelectedIndex(idx);

		//  Repaint any visible geometry.
		for (int i=0; i < 3; ++i)
			if (views[i] != null)   views[i].repaint();
    }

    /**
    *  Method that returns the current model name in this window.
    */
    String getModelName() {
        return data.modelName;
    }

    /**
    *  Method that sets the current model name used by this window.
    */
    void setModelName(String name) {
        data.modelName = name;
        data.geometry.setName(name);
    }


	/**
	*  Handle the user requesting a new document window.
	*  Creates a new instance of this application's main document window.
	*
	*  @param  event  The event that caused this method to be called.  Not used and may be "null".
	*  @returns A reference to the window that was created or null if one was not created.
	*/
	public Frame handleNew( ActionEvent event) {
		return ((AeroLS)MDIApplication.getInstance()).handleNew(event);
	}


    /**
     * Method that reads in the specified file and creates a new window for displaying
     * it's contents.
     *
     * @param theFile The file to be loaded and displayed. If null is passed,
     *                 this method will simply create an empty window.
     * @return A reference to the new window that is created and displayed.
     * @throws Exception if there is a problem creating the window
     * instance.
     */
    public static AppWindow newWindowFromDataFiles(File theFile) throws Exception {

        // Create an instance of this window.
        AppWindow window = new AppWindow();

        //  Read in the file.
        window.openFile(theFile);
        
        //  Set the window name.
        String windowName = FileUtils.getFileNameWithoutExtension(theFile);
        MDIApplication.addWindow(window);
        window.setName(windowName);
        window.setVisible(true);

        return window;
    }

	/**
	*  Handles the user choosing "Open" from the File menu.
    * 
	*  @param  event  The event that caused this method to be called.  Not used and may be "null".
	*/
    public void handleOpen(ActionEvent event) {
        // Does this data already have a file name/path associated with it?
        String dir = null;
        if (fileRef != null)
            dir = fileRef.getParent();

        File theFile = AppUtilities.selectFile(this, FileDialog.LOAD, "Choose an input file", dir,
                "", MDIApplication.getFilenameFilter());

        if (theFile != null)
            // Go open the file and add its data to this window.
            openFile(theFile);

    }

    /**
     * Function that opens the specified data file and adds it to the current window.
     *
     * @param theFile The data file to read in.
     */
    private void openFile(File theFile) {
        
        // Read the file and add it to the current window.
        // Create an input stream to the file.
        try (InputStream input = new FileInputStream(theFile)) {
            String extension = FileUtils.getExtension(theFile);
            
            //  Read in the data from the file.
            if ("vl0".equals(extension))
                //  Read geometry in from the specified VL0 file.
                data = VL0Reader.read(input, data.pdb);

            else if ("ggp".equals(extension))
                //  Read geometry in from the specified GGP file.
                data = GGPReader.read(input, data.pdb);

            else
                //  Read the project data in from the specified XML file stream.
                data = XMLInterface.read(input, data.pdb);
            
            // Change the name of this window to match the new file name.
            String fileName = FileUtils.getFileNameWithoutExtension(theFile);
            this.setTitle(fileName);
            data.modelName = fileName;
            data.geometry.setName(fileName);
            fileRef = theFile;

            //  Change the geometry viewed by the view panels.
            for (int i = 0; i < 3; ++i)
                if (views[i] != null)
                    views[i].setGeometry(data.geometry);

            //  Update the main window display.
            this.update();

        } catch (InterruptedException e) {
            //  If user cancels, just continue.
        } catch (IOException e) {
            JOptionPane.showMessageDialog(this, e.getMessage(), "I/O Error", JOptionPane.ERROR_MESSAGE);

        } catch (Exception e) {
            e.printStackTrace();
            AppUtilities.showException(null, "Unexpected Error",
                    "Copy the following message and e-mail it to the author:", e);
        }

    }

	/**
	*  Handle the user choosing "Close" from the File menu.  This implementation
	*  dispatches a "Window Closing" event to this window.
	*/
	public void handleClose( ActionEvent event ) {
		this.dispatchEvent( new WindowEvent( this, WindowEvent.WINDOW_CLOSING ) );
	}

	/**
	*  Handle the user choosing "Save" from the File menu.  Saves
	*  the currently loaded geometry as well as any analysis parameters
	*  to the last saved or opened file.
	*/
	public void handleSave( ActionEvent event ) {

	    //  If there is no current file, ask the user for one.
        if (fileRef == null)
            handleSaveAs(null);

        else
            writeOutData(fileRef);
	}


	/**
	*  Handle the user choosing "Save As" from the File menu.  Saves
	*  the currently loaded geometry as well as any analysis parameters.
	*/
	public void handleSaveAs( ActionEvent event ) {

		// Does this data already have a file name/path associated with it?
        String fileName = "Untitled";
        String dir = null;
        if (fileRef != null) {
            dir = fileRef.getParent();
            fileName = fileRef.getName();
        }
        int idx = fileName.lastIndexOf(".");
        if (idx > 0)
            fileName = fileName.substring(0, idx);
        fileName += "." + EXTENSION;
        
        //	Ask the user to select a file for saving.
        ExtFilenameFilter filter = new ExtFilenameFilter(EXTENSION, "AeroLS Analysis File");
        String fileExistsMsgFmt = "<html>A file with the name \"{0}\" already exists.<br>Do you want to replace it?</html>";
        File chosenFile = AppUtilities.selectFile4Save(AppWindow.this,
                "Choose a name for this AeroLS Analysis File", dir, fileName, filter,
                EXTENSION, fileExistsMsgFmt, "Warning");

        if (chosenFile != null) {
            if (chosenFile.exists() && !chosenFile.canWrite()) {
                String canNotWrite2File = "<html>Can not write to the selected file.<br>Make sure it is not locked or in use by another program.</html>";
                JOptionPane.showMessageDialog(AppWindow.this,
                        canNotWrite2File, "I/O Error", JOptionPane.ERROR_MESSAGE);
                return;
            }

		    writeOutData(chosenFile);
            fileRef = chosenFile;
		}

	}


	/**
	*  Method that saves the project's data to a disk file.
	*/
	private void writeOutData(File theFile) {

        if (theFile != null) {

	        // Create an output stream to the file.
			try (OutputStream out = new FileOutputStream(theFile)) {

		        //	Convert the project data to XML and write it out to the output stream.
		        XMLInterface.outputToXML(out, data);

				// Change the name of this guiApp's windows to match the new file name.
		        String fileName = theFile.getName();
				fileName = fileName.substring(0, fileName.lastIndexOf("."));
				this.setTitle(fileName);
				MDIApplication.windowTitleChanged(this);

			} catch( IOException e ) {
				JOptionPane.showMessageDialog( this, e.getMessage(), "I/O Error", JOptionPane.ERROR_MESSAGE );
			}
        }

	}


	/**
	*  Handle the user choosing "Test Conditions" from the Analysis menu.
	*/
	public void handleTC( ActionEvent event ) {

        //  Display a dialog the allows the user to define the test conditions.
        JDialog dialog = new TestConditionsDialog(this, "Test Conditions", data);
        dialog.pack();

        //  Position the dialog.
        dialog.setLocation(AppUtilities.dialogPosition(dialog));
        dialog.setVisible(true);

	}


	/**
	*  Handle the user choosing "Analyze" from the Analysis menu.
	*/
	public void handleAnalyze( ActionEvent event ) {

		//	Create and launch our analysis thread.
	    Thread analysisThread = new AnalysisThread(data, this);
		analysisThread.start();

	}


	/**
	*  Handles the user clicking on the Add Component button in
	*  the toolbar.
	*/
	private void handleNewComponent() {

	    //  Ask the user to name this component.
	    String name = JOptionPane.showInputDialog(this, "Please input the name for this component:");
	    if (name != null && !name.equals("")) {

	        //  User input a valid name, so create the new component.
	        GeomComponent comp = new GeomComponent(name);
	        data.geometry.add(comp);

	        //  Update our on-screen list of components.
            compListModel.elementAdded();
	    }

	}

	/**
	*  Handles the user clicking on the Del Component button in
	*  the toolbar.
	*/
	private void handleDelComponent() {

	    int size = data.geometry.size();
	    if (size > 0) {
            int idx = compList.getSelectedIndex();
            if (idx > -1) {
                String name = ((GeomElement)data.geometry.get(idx)).getName();

                //  Ask the user if they are sure.
                int pick = JOptionPane.showConfirmDialog(this,
                                "Are you sure you want to delete the component:  " + name,
                                "Confirm Delete", JOptionPane.YES_NO_OPTION);
                if (pick == JOptionPane.OK_OPTION) {

                    //  The user confirmed the delete operation.

                    //  Update the on-screen components list.
                    compListModel.removeComponent(idx);

                    //  Remove the component from the geometry list.
                    data.geometry.remove(idx);
                }
            }
	    }
	}

	/**
	*  Handles the user clicking on the Add Wing Segment button in
	*  the toolbar.
	*/
	private void handleNewWingSeg() {

	    //  First make sure there is at least one component available.
	    if (data.geometry.size() < 1)
	        handleNewComponent();

	    //  Get a index of the selected component.
        int idx = compList.getSelectedIndex();
	    if (idx < 0) {

	        //  No component is selected, ask the user to pick one.
	        int size = data.geometry.size();
	        if (size > 1) {
    	        String[] names = new String[size];
    	        for (int i=0; i < size; ++i)
    	            names[i] = ((GeomElement)data.geometry.get(i)).getName();

    	        String selValue = (String)JOptionPane.showInputDialog(this,
    	                    "Choose the component to add this segment to:",
    	                    "Pick Component", JOptionPane.PLAIN_MESSAGE, null,
    	                    names, names[0]);

    	        //  Figure out which component was selected.
    	        for (int i=0; i < size; ++i) {
    	            if (names[i].equals(selValue)) {
    	                idx = i;
    	                break;
    	            }
    	        }

    	        compList.setSelectedIndex(idx);

	        } else if (size == 1) {
	            idx = 0;
	            compList.setSelectedIndex(idx);
	        }
	    }

        if (idx > -1 && !data.geometry.isEmpty()) {
            //  Get the component we are adding a segment to.
    	    GeomComponent comp = (GeomComponent)data.geometry.get(idx);


            //  Display a dialog the allows the user to define a wing segment.
    	    try {

                //  Show the set/change references dialog window.
    	        JDialog dialog = new NewWingSegDialog(this, "New Wing Segment", comp, data.pdb);
    	        dialog.pack();
    	        dialog.setSize(new Dimension(700,540));

                //  Position the dialog.
                dialog.setLocation(AppUtilities.dialogPosition(dialog));
    	        dialog.setVisible(true);

    	    } catch (MalformedURLException err) {
    		    JOptionPane.showMessageDialog( null, err.getMessage(), "Improper URL",
    								    JOptionPane.ERROR_MESSAGE );
                err.printStackTrace();
    	    }
        }

	}

	/**
	*  Handles the user clicking on the Del Wing Segment button in
	*  the toolbar.
	*/
	private void handleDelSegment() {

	    if (!data.geometry.isEmpty()) {
            int idx = compList.getSelectedIndex();
            GeomComponent comp = (GeomComponent)data.geometry.get(idx);

            if (!comp.isEmpty()) {
                idx = segList.getSelectedIndex();

                if (idx > -1) {
                    String name = ((GeomElement)comp.get(idx)).getName();

                    //  Ask the user if they are sure.
                    int pick = JOptionPane.showConfirmDialog(this,
                                "Are you sure you want to delete the segment:  " + name,
                                "Confirm Delete", JOptionPane.YES_NO_OPTION);
                    if (pick == JOptionPane.OK_OPTION) {
                        //  The user confirmed the delete operation.

                        //  Update the on-screen components list.
                        segListModel.removeElement(idx);

                        //  Remove the segment from the component list.
                        comp.remove(idx);

                        //  Update the display.
                        this.getContentPane().repaint();
                    }
                }
            }
	    }
	}

	/**
	*  Handles the user clicking on the references button in
	*  the toolbar.
	*/
	private void handleReferences() {

	    try {

            //  Show the set/change references dialog window.
	        JDialog dialog = new ReferencesDialog(this, "Set References", data.pdb);
	        dialog.pack();

            //  Position the dialog.
            dialog.setLocation(AppUtilities.dialogPosition(dialog));
	        dialog.setVisible(true);

	    } catch (MalformedURLException err) {
		    JOptionPane.showMessageDialog( null, err.getMessage(), "Improper URL",
								    JOptionPane.ERROR_MESSAGE );
            err.printStackTrace();
	    }
	}


	/**
	*  Handles the user clicking on the zoom in button in
	*  the toolbar.
	*/
	private void handleZoomIn() {
        int denom = views[0].getScale();

        int p10 = (int)(Math.log10(denom)+.1);
        int value = (int)((double)denom/MathTools.pow10(p10));
        switch (value) {
            case 1:
                value = 5;
                --p10;
                break;
            case 2:
                value = 1;
                break;
            default:
                value = 2;
                break;
        }

        denom = (int)(value*MathTools.pow10(p10));
        for (int i=0; i < 3; ++i) {
            views[i].setScale(denom);
            views[i].repaint();
        }

	}


	/**
	*  Handles the user clicking on the zoom out button in
	*  the toolbar.
	*/
	private void handleZoomOut() {

        int denom = views[0].getScale();

        int p10 = (int)(Math.log10(denom)+.1);
        int value = (int)((double)denom/MathTools.pow10(p10));
        switch (value) {
            case 1:
                value = 2;
                break;
            case 2:
                value = 5;
                break;
            default:
                value = 1;
                ++p10;
                break;
        }

        denom = (int)(value*MathTools.pow10(p10));
        for (int i=0; i < 3; ++i) {
            views[i].setScale(denom);
            views[i].repaint();
        }

	}


    /**
    *  Method that creates a panel containing a list of
    *  the component objects in this model.
    */
    private JPanel createComponentListPanel() {
        JPanel panel = new JPanel(new BorderLayout());

        //  Put a label at the top.
        JLabel label = new JLabel("Components:");
        panel.add(label, BorderLayout.NORTH);


        //  Create a list to display the components.
        compListModel = new ComponentListModel();
        JList list = new JList(compListModel);
        compList = list;
        list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        list.setSelectedIndex(0);
        list.setPrototypeCellValue("Component Name");
        list.setVisibleRowCount(4);

        //  Create a selection listener for this class that updates the
        //  wing segment list appropriately.
        ListSelectionListener selListener = new ListSelectionListener() {
            @Override
            public void valueChanged(ListSelectionEvent e) {
                if (e.getValueIsAdjusting())    return;

                JList theList = (JList)e.getSource();
                if (theList.isSelectionEmpty()) {
                    //  Tell the segment list to go blank.
                    segListModel.setData(null);

                } else {
                    //  Tell the segment list to display all the segments
                    //  for the specified component.
                    int index = theList.getSelectedIndex();
                    if (!data.geometry.isEmpty())
                        segListModel.setData((java.util.List)data.geometry.get(index));
                    else
                        segListModel.setData(null);
                }
            }
        };
        list.addListSelectionListener(selListener);

        //  Listen for double-clicks on list items.
        MouseListener mouseListener = new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
                if (e.getClickCount() == 2) {
                    int index = compList.locationToIndex(e.getPoint());

                    //  Ask the user for a new component name.
                    GeomElement comp = (GeomElement)data.geometry.get(index);
                    String name = comp.getName();
	                name = JOptionPane.showInputDialog(AppWindow.this, "Rename component \""
	                                    + name + "\":");
	                if (name != null && !name.equals(""))
                        comp.setName(name);
                }
            }
        };
        list.addMouseListener(mouseListener);

        //  Put the list in a scroll pane so that we have scroll bars.
        JScrollPane listScrollPane = new JScrollPane(list);
        listScrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);


        //  Set a minimum size for the list.
        Dimension minimumSize = new Dimension(50, 50);
        listScrollPane.setMinimumSize(minimumSize);


        //  Add the list to the panel.
        panel.add(listScrollPane, BorderLayout.CENTER);

        return panel;
    }


    /**
    *  Method that creates a panel containing a list of
    *  the segment objects in the selected component.
    */
    private JPanel createSegmentListPanel() {
        JPanel panel = new JPanel(new BorderLayout());

        //  Put a label at the top.
        JLabel label = new JLabel("Segments:");
        panel.add(label, BorderLayout.NORTH);


        //  Create a list to display the segments.
        segListModel = new SegmentListModel();
        JList list = new JList(segListModel);
        segList = list;
        list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        list.setSelectedIndex(0);
        list.setPrototypeCellValue("Component Name");
        list.setVisibleRowCount(4);


        //  Listen for double-clicks on list items.
        MouseListener mouseListener = new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
                if (e.getClickCount() == 2) {
                    int index = segList.locationToIndex(e.getPoint());

                    //  Ask the user for a new segment name.
                    java.util.List comp = segListModel.getData();
                    if (comp != null) {
                        GeomElement seg = (GeomElement)comp.get(index);
                        String name = seg.getName();
	                    name = JOptionPane.showInputDialog(AppWindow.this, "Rename segment \""
	                                    + name + "\":");
	                    if (name != null && !name.equals(""))
                            seg.setName(name);
                    }
                }
            }
        };
        list.addMouseListener(mouseListener);

        //  Put the list in a scroll pane so that we have scroll bars.
        JScrollPane listScrollPane = new JScrollPane(list);
        listScrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);


        //  Set a minimum size for the list.
        Dimension minimumSize = new Dimension(50, 50);
        listScrollPane.setMinimumSize(minimumSize);


        //  Add the list to the panel.
        panel.add(listScrollPane, BorderLayout.CENTER);

        return panel;
    }


    /**
    *  Class that provides data to our component list as needed.  This object "models"
	*  the component data so that it can be presented in a list.
    */
    private class ComponentListModel extends AbstractListModel {

        @Override
        public int getSize() {
            int size = data.geometry.size();
            return size;
        }

        @Override
        public Object getElementAt(int index) {
            GeomElement comp = (GeomElement)data.geometry.get(index);
            return comp.getName();
        }

        /**
        *  This method is called to update the list after a new
        *  component has been added.
        */
        public void elementAdded() {
            int pos = data.geometry.size() - 1;
            fireIntervalAdded(this, pos, pos);
        }

        /**
        *  Removes the specified component from the component
        *  list.  This should only be called just before
        *  a component is removed from the geometry list.
        */
        public void removeComponent(int idx) {
            fireIntervalRemoved(this, idx, idx);
			segListModel.setData(null);
        }


    }


    /**
    *  A model of the data in our segment display list.  The data in this list changes
	*  dynamically when the user changes which component is selected in the component list.
    */
    private class SegmentListModel extends DynamicListModel {
        @Override
        public Object getElementAt(int index) {
            GeomElement element = (GeomElement)this.getData().get(index);
            return element.getName();
        }
    }


	/**
	*  Defines a button listener for our toolbar buttons.
	*/
	private class ButtonL implements ActionListener {

        @Override
		public void actionPerformed( ActionEvent evt ) {
			String cmd = evt.getActionCommand();
			
            switch (cmd) {
                case "References":
                    handleReferences();
                    break;
                case "Add Component":
                    handleNewComponent();
                    break;
                case "Del Component":
                    handleDelComponent();
                    break;
                case "Add Wing Segment":
                    handleNewWingSeg();
                    break;
                case "Del Segment":
                    handleDelSegment();
                    break;
                case "Zoom In":
                    handleZoomIn();
                    break;
                case "Zoom Out":
                    handleZoomOut();
                    break;
                default:
                    break;
            }

		}

	}

    /**
    *  A panel that displays information about the currently loaded model.
    */
    private class InfoPanel extends JPanel {

        //  Label for the model name field.
	    JLabel nameLbl;

        //  List of parameter text fields.
	    ParamInput[] paramFields = new ParamInput[6];


        /**
        *  Construct a new info panel object.
        */
        @SuppressWarnings("OverridableMethodCallInConstructor")
	    InfoPanel() {

            //  Get a reference to the parameter database.
	        ParamDatabase pdb = data.pdb;


            //  Layout the panel.
            Box box0 = Box.createHorizontalBox();
            this.add(box0);
            box0.add(Box.createGlue());

            Box box1 = Box.createVerticalBox();

            box1.add(Box.createGlue());

            nameLbl = new JLabel("Model Name = " + data.modelName);
            box1.add(nameLbl);

            box1.add(Box.createVerticalStrut(10));

            //  Create parameter input/display boxes for each parameter.
            ParamInput paramInput = new ParamInput(pdb.get(SREF));
            paramInput.setEditable(false);
            paramFields[0] = paramInput;
            box1.add(paramInput);
            box1.add(Box.createVerticalStrut(4));

            paramInput = new ParamInput(pdb.get(CREF));
            paramInput.setEditable(false);
            paramFields[1] = paramInput;
            box1.add(paramInput);
            box1.add(Box.createVerticalStrut(4));

            paramInput = new ParamInput(pdb.get(BREF));
            paramInput.setEditable(false);
            paramFields[2] = paramInput;
            box1.add(paramInput);
            box1.add(Box.createVerticalStrut(4));

            paramInput = new ParamInput(pdb.get(XMRC));
            paramInput.setEditable(false);
            paramFields[3] = paramInput;
            box1.add(paramInput);
            box1.add(Box.createVerticalStrut(4));

            paramInput = new ParamInput(pdb.get(YMRC));
            paramInput.setEditable(false);
            paramFields[4] = paramInput;
            box1.add(paramInput);
            box1.add(Box.createVerticalStrut(4));

            paramInput = new ParamInput(pdb.get(ZMRC));
            paramInput.setEditable(false);
            paramFields[5] = paramInput;
            box1.add(paramInput);
            box1.add(Box.createVerticalStrut(4));

            box1.add(Box.createGlue());

            box0.add(box1);
            box0.add(Box.createGlue());

            //  Add a pretty border to this panel.
            this.setBorder(BorderFactory.createEtchedBorder());

        }

        /**
        *  Method that updates the display of information in this panel.
        */
        public void update() {

            //  Update the model name.
            nameLbl.setText("Model Name = " + data.modelName);

            //  Update all the param input boxes.
            int length = paramFields.length;
            for (int i=0; i < length; ++i)
                paramFields[i].resetText();
        }
    }
}

