/*
 *  TablePlotWindow  -- An window that displays a plot of the specified analysis table.
 *
 *   Copyright (C) 2002-2024 by Joseph A. Huwaldt
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */

import jahuwaldt.js.tables.Float64Table;
import jahuwaldt.plot.*;
import jahuwaldt.swing.AboutJMenuItem;
import jahuwaldt.swing.AppUtilities;
import jahuwaldt.swing.MDIApplication;
import jahuwaldt.swing.QuitJMenuItem;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.List;
import javax.swing.*;
import javax.swing.border.Border;
import javolution.util.FastTable;
import org.jscience.mathematics.number.Integer64;
import org.jscience.mathematics.vector.Float64Vector;

/**
 * An application window for the application TableReader that displays a plot of
 * a selected n-dimensional table.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: December 21, 2000
 * @version January 21, 2024
 */
public class TablePlotWindow extends JFrame {

	// Definition of this window's menus.
	private static final 
		String[][] FILEM_STRINGS = {		{ "New", null, null },
										{ "Open...", null, null },
										{ null, null, null },
										{ "Close", "W", "handleClose" },
										{ "Save", null, null } };
	
	private static final 
		String[][] EDITM_STRINGS = {		{ "Undo", "Z", null },
										{ null, null, null },
										{ "Cut", "X", null },
										{ "Copy", "C", null },
										{ "Paste", "V", null } };

	//	References to the left/right buttons.
	private final JButton LEFT_BTN = new JButton("<");
	private final JButton RIGHT_BTN = new JButton(">");

	// The size of the plot symbols.
	private static final int SYM_SIZE = 6;
	
	// The number of plain symbols.
	private static final int NUMPLAINSYM = 22;
	
	// A list of plot symbols to use in this program.
	private static final PlotSymbol[] SYMBOLS = {	new CircleSymbol(),
													new SquareSymbol(),
													new DiamondSymbol(),
													new Triangle1Symbol(),
													new BoxLRSymbol(),
													new RTriangle1Symbol(),
													new TabUpSymbol(),
													new Triangle2Symbol(),
													new BoxLLSymbol(),
													new RTriangle2Symbol(),
													new TabRSymbol(),
													new Triangle3Symbol(),
													new BoxULSymbol(),
													new RTriangle3Symbol(),
													new TabDSymbol(),
													new Triangle4Symbol(),
													new BoxURSymbol(),
													new RTriangle4Symbol(),
													new TabLSymbol(),
													new ThinRect1Symbol(),
													new ThinRect2Symbol(),
													new XSymbol(),
													new CircleSymbol(),
													new SquareSymbol(),
													new DiamondSymbol(),
													new Triangle1Symbol(),
													new BoxLRSymbol(),
													new RTriangle1Symbol(),
													new TabUpSymbol(),
													new Triangle2Symbol(),
													new BoxLLSymbol(),
													new RTriangle2Symbol(),
													new TabRSymbol(),
													new Triangle3Symbol(),
													new BoxULSymbol(),
													new RTriangle3Symbol(),
													new TabDSymbol(),
													new Triangle4Symbol(),
													new BoxURSymbol(),
													new RTriangle4Symbol(),
													new TabLSymbol(),
													new ThinRect1Symbol(),
													new ThinRect2Symbol(),
													new XSymbol()	};
	private int symPos = 0;
	
	// A reference to the table we are doing the plot of.
	private final Float64Table table;

	// A reference to the content pane.
	JPanel cp = new JPanel(new BorderLayout());
	
	// A reference to the panel containing the plot.
	private PlotPanel plotPanel;
	
	// A reference to the panel containing the current high dimension breakpoints.
	private JPanel bpPanel;
	
	// List of indexes into the n-dimensional table.
	private final List<Integer64> indep;
	
	// The number format used to format numbers.
	private final DecimalFormat nf = (DecimalFormat)NumberFormat.getInstance();
	

	/**
	*  Creates a new instance of a table plot window using the
	*  specified table.
	*
	*  @param  app       Reference to the main application class.
	*  @param  theTable  The table we want to do a plot of.
	*/
    @SuppressWarnings("OverridableMethodCallInConstructor")
	public TablePlotWindow( Float64Table theTable ) throws NoSuchMethodException {
		super( "Plot of " + theTable.getName() );

		setResizable( true );
		setSize(600, 450);
		
		// Set up some class variables.
		this.table = theTable;
		nf.setMinimumFractionDigits(3);
		
		// Set up this window's user interface.
		getContentPane().add(cp);
		
		//	Create navigation buttons.
		JButton fullLeftBtn = new JButton("<<");
		fullLeftBtn.setHorizontalAlignment(SwingConstants.CENTER);
		fullLeftBtn.addActionListener(new ActionListener() {
            @Override
			public void actionPerformed(ActionEvent evt) {
				doFullLeftButton();
			}
		} );
		
		LEFT_BTN.setEnabled(false);
		LEFT_BTN.setHorizontalAlignment(SwingConstants.CENTER);
		LEFT_BTN.addActionListener(new ActionListener() {
            @Override
			public void actionPerformed(ActionEvent evt) {
				doLeftButton();
			}
		} );
		
		RIGHT_BTN.setEnabled(table.getDimensions() > 2);
		RIGHT_BTN.setHorizontalAlignment(SwingConstants.CENTER);
		RIGHT_BTN.addActionListener(new ActionListener() {
            @Override
			public void actionPerformed(ActionEvent evt) {
				doRightButton();
			}
		} );
			
		JButton fullRightBtn = new JButton(">>");
		fullRightBtn.setHorizontalAlignment(SwingConstants.CENTER);
		fullRightBtn.addActionListener(new ActionListener() {
            @Override
			public void actionPerformed(ActionEvent evt) {
				doFullRightButton();
			}
		} );
		
		JButton closeButton = new JButton("Close");
		this.getRootPane().setDefaultButton(closeButton);
		closeButton.addActionListener(new ActionListener() {
            @Override
			public void actionPerformed(ActionEvent evt) {
				handleClose(null);
			}
		} );
		
		//	Create a button panel and add all the buttons to it.
		Box box = new Box(BoxLayout.X_AXIS);
		box.setBackground( SystemColor.window );
		box.add(Box.createGlue());
		box.add(fullLeftBtn);
		box.add(LEFT_BTN);
		box.add(RIGHT_BTN);
		box.add(fullRightBtn);
		box.add(Box.createGlue());
		box.add(closeButton);
		box.add(Box.createHorizontalStrut(20));
		if (table.getDimensions() < 3) {
            fullLeftBtn.setEnabled(false);
            LEFT_BTN.setEnabled(false);
            RIGHT_BTN.setEnabled(false);
            fullRightBtn.setEnabled(false);
		}
		JPanel buttonPanel = new JPanel();
		buttonPanel.add(box);
		buttonPanel.setBorder(BorderFactory.createEmptyBorder(10,10,10,10));
		cp.add(buttonPanel, BorderLayout.SOUTH);
		
		
		//	Initialize the plot symbols.
		initPlotSymbols();
		
		//	Create some class variables that are used in redrawPlot().
        int numDim = table.getDimensions();
        indep = FastTable.newInstance();
        for (int i = 0; i < numDim; ++i)
            indep.add(Integer64.ZERO);
		redrawPlot();
		JPanel legendPanel = createLegendPanel();

		if (legendPanel != null) {
		    Box legendBox = new Box(BoxLayout.Y_AXIS);
		    legendBox.add(Box.createGlue());
		    legendBox.add(legendPanel);
		    legendBox.add(Box.createGlue());
			JPanel lpanel = new JPanel();
			lpanel.add(legendBox);
			lpanel.setBorder(BorderFactory.createEmptyBorder(0, 6, 0, 6));
			lpanel.setBackground( Color.white );
		    cp.add(lpanel, BorderLayout.EAST);
		}

		//  Create a menu bar for this window.
		JMenuBar menuBar = createMenuBar();
		this.setJMenuBar(menuBar);
	}

	/**
	*  Initializes the menus associated with this window.
	*/
	private JMenuBar createMenuBar() throws NoSuchMethodException {

		JMenuBar menuBar = new JMenuBar();
		
		// Set up the file menu.
		JMenu menu = AppUtilities.buildMenu( this, "File", FILEM_STRINGS );
		menuBar.add( menu );

        //  Add a Quit menu item (if it isn't automatically present).
        AeroLS guiApp = (AeroLS)MDIApplication.getInstance();
        if (!QuitJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(guiApp.getQuitJMenuItem());
        }
        
		// Set up the edit menu.
		menu = AppUtilities.buildMenu( this, "Edit", EDITM_STRINGS );
		menuBar.add( menu );

		//  Create a Windows menu.
		menu = MDIApplication.newWindowsMenu("Windows");
		menuBar.add(menu);
		
        //  Create an about menu item (if not automatically present).
        if (!AboutJMenuItem.isAutomaticallyPresent()) {
            //  Create Help menu.
            menu = new JMenu("Help");
            menuBar.add(menu);
            
            //  Add the "About" item to the Help menu.
            menu.add(guiApp.getAboutJMenuItem());
        }
        
		return menuBar;
	}


	/**
	*  Method that regenerates and redraws the plot to the window.
	*/
	private void redrawPlot() {
		
		//	Remove the old plot.
		if (plotPanel != null)
			cp.remove(plotPanel);
		
		//	Remove the high dimension breakpoint panel.
		if (bpPanel != null)
			cp.remove(bpPanel);

		//	Create the new plot and add it to the window.
		plotPanel = createPlotPanel();
		if (plotPanel != null)
			cp.add(plotPanel, BorderLayout.CENTER);
		
		//	Create a new breakpoint label at the top (if necissary).
		bpPanel = createBPPanel();
		if (bpPanel != null)
			cp.add(bpPanel, BorderLayout.NORTH);
		
		validate();
		repaint();
	}
	
	/**
	*  Create a high dimension breakpoint panel.
	*/
	private JPanel createBPPanel() {
		JPanel thePanel = null;
		
		int dim = table.getDimensions();
		if (dim > 2) {
			int numElements = dim - 2;
			thePanel = new JPanel(new GridLayout(numElements, 1));
			thePanel.setBackground( Color.white );
			
			for (int i=0; i < numElements; ++i) {
				Float64Vector breakpoints = table.getBreakpoints(i);
				String indepName = table.getIndepName(i).toString();
				String message = indepName + " = " + nf.format(breakpoints.get(indep.get(i).intValue()));
				JLabel label = new JLabel(message, SwingConstants.CENTER);
				thePanel.add(label);
			}
		}
		
		return thePanel;
	}
	
	/**
	*  Creates the plot of the table showing the 1st & 2nd independents at
	*  the current breakpoint for higher independents.
	*/
	private PlotPanel createPlotPanel() {
	
		//	Build up a run list.
		int dim = table.getDimensions();
		Float64Vector xArr = table.getBreakpoints(dim - 1);
		PlotRunList runList = new PlotRunList();
		symPos = 0;
		
		if (dim < 2) {
			//	Just a single run for 1D plots.
			runList.add(createRun(xArr, indep));
			
		} else {
			//	All higher dimension plots are represented by a plot with multiple runs and panes.
			Float64Vector d2Arr = table.getBreakpoints(dim - 2);
			int length = d2Arr.getDimension();
			for (int i=0; i < length; ++i) {
				indep.set(dim - 2, Integer64.valueOf(i));
				runList.add(createRun(xArr, indep));
			}
		}
		
		//	Create the plot.
		Plot2D aPlot = new SimplePlotXY(runList, null,
								table.getIndepName(dim-1).toString(), table.getName().toString(),
								nf, nf);
		
		//	Create an panel to hold our plot.
		PlotPanel panel = new PlotPanel(aPlot);
		panel.setBackground( Color.white );
		
		return panel;
	}
	
	/**
	*  Creates a single run from a table at the specified dimension.
	*/
	private PlotRun createRun(Float64Vector xArr, List<Integer64> indep) {
	
		int dim = table.getDimensions();
		PlotSymbol symbol = nextSymbol();
		PlotRun run = new PlotRun();
		int length = xArr.getDimension();
		for (int i=0; i < length; ++i) {
            indep.set(dim - 1, Integer64.valueOf(i)); // indep[dim - 1] = i;
			double value = table.get(indep).doubleValue();
			run.add(new PlotDatum(xArr.getValue(i), value, true, symbol));
		}
		
		return run;
	}
	
	/**
	*  Returns the next symbol in the symbol list.  When the end
	*  of the list is reached, the 1st symbol is returned.
	*/
	private PlotSymbol nextSymbol() {
		if (symPos >= SYMBOLS.length)
			symPos = 0;
		return SYMBOLS[symPos++];
	}
	
	/**
	*  Initialize the plot symbols.
	*/
	private void initPlotSymbols() {
		int length = SYMBOLS.length;
		for (int i=0; i < length; ++i)
			SYMBOLS[i].setSize(SYM_SIZE);
		
		int start = NUMPLAINSYM;
		int end = 2*NUMPLAINSYM;
		for (int i=start; i < end; ++i)
			SYMBOLS[i].setFillColor(Color.black);
			
	}
	
	/**
	*  Creates a legend panel showing the breakpoints for
	*  the 2nd dimension plotted (if there is one).
	*/
	private JPanel createLegendPanel() {
		JPanel thePanel = null;

		int dim = table.getDimensions();
		
		if (dim > 1) {
			String indepName = table.getIndepName(dim - 2).toString();
			Float64Vector d2Arr = table.getBreakpoints(dim - 2);
			int d2length = d2Arr.getDimension();
			symPos = 0;
			
			//	Create the panel.
			thePanel = new JPanel(new GridLayout(d2length + 1,1));
			thePanel.setBorder( BorderFactory.createLineBorder(Color.black) );
			thePanel.add(new JLabel(" " + indepName + " ", SwingConstants.CENTER));
			
			//	Add labels to the panel for each breakpoint.
			for (int i=0; i < d2length; ++i) {
				//	Create a panel to draw the plot symbols.
				SymbolPanel symbol = new SymbolPanel(nextSymbol());
				symbol.setBackground(Color.white);
				
				//	Create a panel for displaying the plot symbol + the value.
				JPanel label = new JPanel();
				label.setBackground(Color.white);
				label.add(symbol);
				
				//	Create a value label.
				String value = nf.format(d2Arr.getValue(i));
				label.add( new JLabel(" " + value + " ") );
				thePanel.add(label);
			}
			
		    //	Put a line around the legend panel.
		    Border emptyBorder = BorderFactory.createEmptyBorder(10,0,10,10);
		    Border lineBorder = BorderFactory.createLineBorder(Color.black);
		    thePanel.setBorder( BorderFactory.createCompoundBorder(emptyBorder, lineBorder ) );
			thePanel.setBackground(Color.white);
		}
		
		return thePanel;
	}
	
	/**
	*  Handles the user clicking on the full-left button.
	*/
	private void doFullLeftButton() {
		LEFT_BTN.setEnabled(false);
		RIGHT_BTN.setEnabled(true);
		
		int dim = table.getDimensions() - 3;
		for (int i=0; i <= dim; ++i) {
			indep.set(i, Integer64.ZERO);
		}

		redrawPlot();
	}

	/**
	*  Handles the user clicking on the left button.
	*/
	private void doLeftButton() {
		RIGHT_BTN.setEnabled(true);
		int dim = table.getDimensions() - 3;
		int old_dim = dim;
		
        indep.set(dim, indep.get(dim).minus(Integer64.ONE));        // --_index[dim];
        while (dim != 0 && indep.get(dim).intValue() < 0) {
            int bpIdx = table.getNumBreakpoints(dim) - 1;
            indep.set(dim, Integer64.valueOf(bpIdx));               // _index[dimOrder[dim]] = bpIdx;
            --dim;
            indep.set(dim, indep.get(dim).minus(Integer64.ONE));    // --_index[dimOrder[dim]];
        }
		
        // Have we reached the end of all the breakpoints of
        // all the dimensions of the table?
        dim = old_dim;
        while (dim != 0 && indep.get(dim).intValue() == 0) {
            --dim;
        }
        if (dim == 0 && indep.get(dim).intValue() == 0)
            LEFT_BTN.setEnabled(false);
		
		redrawPlot();
	}
	
	/**
	*  Handles the user clicking on the right button.
	*/
	private void doRightButton() {
		LEFT_BTN.setEnabled(true);
		int dim = table.getDimensions() - 3;
		int old_dim = dim;
		
        indep.set(dim, indep.get(dim).plus(Integer64.ONE)); // ++_index[dimOrder[dim]];
        while (dim != 0 && indep.get(dim).intValue() == table.getNumBreakpoints(dim)) {
            indep.set(dim, Integer64.ZERO); // _index[dimOrder[dim]] = 0;
            --dim;
            indep.set(dim, indep.get(dim).plus(Integer64.ONE)); // ++_index[dimOrder[dim]];
        }

        // Have we reached the end of all the breakpoints of
        // all the dimensions of the table?
        dim = old_dim;
        while (dim != 0 && indep.get(dim).intValue() == table.getNumBreakpoints(dim) - 1) {
            --dim;
        }
        if (dim == 0 && indep.get(dim).intValue() == table.getNumBreakpoints(dim) - 1)
            RIGHT_BTN.setEnabled(false);
		
		redrawPlot();
	}
	
	/**
	*  Handles the user clicking on the full-right button.
	*/
	private void doFullRightButton() {
		LEFT_BTN.setEnabled(true);
		RIGHT_BTN.setEnabled(false);
		
		int dim = table.getDimensions() - 3;
		for (int i=0; i <= dim; ++i) {
            int bpIdx = table.getNumBreakpoints(i) - 1;
            indep.set(i, Integer64.valueOf(bpIdx)); // _index[dimOrder[i]] = bpIdx;
		}

		redrawPlot();
	}

	/**
	*  Handle the user choosing "Close" from the File menu.  This implementation
	*  dispatches a "Window Closing" event to this window.
	*/
	public void handleClose( ActionEvent event ) {
		this.dispatchEvent( new WindowEvent( this, WindowEvent.WINDOW_CLOSING ) );
	}


	/**
	*  A panel for drawing plot symbols.
	*/
	private class SymbolPanel extends JPanel {
		
		private final PlotSymbol symbol;
		
		/**
		*  Constructor
		*/
		SymbolPanel(PlotSymbol theSymbol) {
			symbol = theSymbol;
			int width = theSymbol.getSize();
			Dimension dim = new Dimension(width + 8, width + 2);
			SymbolPanel.this.setMinimumSize(dim);
			SymbolPanel.this.setPreferredSize(dim);
		}
		
		/**
		*  Called automatically by system to paint this component.
		*/
        @Override
		public void paintComponent( Graphics gc ) {

			// Paint the background
			super.paintComponent( gc );
		
			if (symbol != null) {
				// Deal with a possible border.
				Insets insets = getInsets();
				int currentWidth = getWidth() - insets.left - insets.right;
				int currentHeight = getHeight() - insets.top - insets.bottom;
				
				int x = currentWidth/2 + insets.left;
				int y = currentHeight/2 + insets.top;
				
				//	Render the symbol image.
				symbol.draw(gc, x, y);
				
				//	Draw a line through the center.
				x -= currentWidth/2;
				gc.drawLine(x, y, x + currentWidth, y);
				
			}
		}

	}
}

