/*
 *  TableWindow  -- A window that lists out tables output from AeroLS.
 *
 *   Copyright (C) 2002-2024 by Joseph A. Huwaldt
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */

import jahuwaldt.io.ExtFilenameFilter;
import jahuwaldt.js.tables.*;
import jahuwaldt.swing.AboutJMenuItem;
import jahuwaldt.swing.AppUtilities;
import jahuwaldt.swing.MDIApplication;
import jahuwaldt.swing.QuitJMenuItem;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.List;
import javax.swing.*;
import javax.swing.border.EmptyBorder;

/**
 * A Window for listing out a set of tables and allowing the user to interact
 * with them.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: June 27, 2002
 * @version January 21, 2024
 */
public class TableWindow extends JFrame {

    // Definition of this window's menus.
	private static final 
		String[][] FILEM_STRINGS ={		{ "New", "N", null },
										{ "Open...", "O", null },
										{ null, null, null },
										{ "Close", "W", "handleClose" },
										{ "Save", "S", null },
										{ "Save To CSV...", null, "handleSaveToCSV" },
                                        {"Save To FDA...", null, "handleSaveToFDA"},
                                        {"Save To OTIS...", null, "handleSaveToOTIS"}
        };

	private static final 
		String[][] EDITM_STRINGS = {	{ "Undo", "Z", null },
										{ null, null, null },
										{ "Cut", "X", null },
										{ "Copy", "C", null},
										{ "Paste", "V", null }
        };

    // The buttons used by this application.
    private final JButton removeButton = new JButton("Remove");
    private final JButton viewButton = new JButton("View");
    private final JButton lookupButton = new JButton("Lookup");
    private final JButton plotButton = new JButton("Plot");

    private static final FTableReader CSV_READER = new CSVReader();
    private static final FTableReader FDA_READER = new FDAReader();
    private static final FTableReader OTIS_READER = new OTISTableReader();

    //  A file filter that recognizes MacOS file types.
    private final ExtFilenameFilter fnFilter = new ExtFilenameFilter();

    // The list of tables shown in the GUI.
    private JList tableList = null;

    // The data model for the tableList.
    private final DefaultListModel listModel;

    //  Database of all the tables that have been supplied to this window.
    private FTableList allTables = null;

    //  Reference to the last file that we referenced.
    private File lastFile;

    /**
     * Construct a new table window with the given name and database of tables.
     */
    @SuppressWarnings("OverridableMethodCallInConstructor")
    public TableWindow(String name, FTableList tdb) throws NoSuchMethodException {
        super(name);

        this.allTables = tdb;

        // Set up the file name filter.
        fnFilter.addExtension("fda");
        fnFilter.addExtension("csv");

        //  Have the window dispose of itself when it closes.
        this.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
        this.setResizable(true);

        // Set up buttons.
        ButtonL theBL = new ButtonL();
        removeButton.setAlignmentX(Component.CENTER_ALIGNMENT);
        removeButton.addActionListener(theBL);
        removeButton.setActionCommand("Remove");
        removeButton.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
        removeButton.setToolTipText("Removes the selected tables");
        plotButton.setAlignmentX(Component.CENTER_ALIGNMENT);
        plotButton.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
        plotButton.setToolTipText("Creates plots of the selected tables");
        plotButton.addActionListener(theBL);
        plotButton.setActionCommand("Plot");
        lookupButton.setAlignmentX(Component.CENTER_ALIGNMENT);
        lookupButton.addActionListener(theBL);
        lookupButton.setActionCommand("Lookup");
        lookupButton.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
        lookupButton.setToolTipText("Lookup values in the selected tables");
        viewButton.setAlignmentX(Component.CENTER_ALIGNMENT);
        viewButton.addActionListener(theBL);
        viewButton.setActionCommand("View");
        viewButton.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
        viewButton.setToolTipText("View the data in tabular form");

        Box buttonBox = Box.createVerticalBox();
        buttonBox.add(Box.createVerticalStrut(6));
        buttonBox.add(viewButton);
        buttonBox.add(Box.createVerticalStrut(6));
        buttonBox.add(lookupButton);
        buttonBox.add(Box.createVerticalStrut(6));
        buttonBox.add(plotButton);
        buttonBox.add(Box.createGlue());
        buttonBox.add(removeButton);
        buttonBox.add(Box.createVerticalStrut(20));

        // Set up the table list.
        listModel = new DefaultListModel();

        tableList = new JList(listModel);
        tableList.setPrototypeCellValue("A Table Entry Size");
        tableList.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
        JScrollPane listScrollPane = new JScrollPane(tableList);
        listScrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);

        // Add the input tables to the list model.
        for (Float64Table newTable : tdb) {
            listModel.addElement(newTable.getName());
        }

        //  Lay out the window.
        JPanel contentPanel = new JPanel(new BorderLayout());
        getContentPane().add(contentPanel);

        JLabel label = new JLabel("Select a table from the list:");
        JPanel panel = new JPanel();
        panel.add(label);
        panel.setBorder(new EmptyBorder(10, 10, 10, 10));
        contentPanel.add(panel, BorderLayout.NORTH);
        contentPanel.add(listScrollPane, BorderLayout.CENTER);
        contentPanel.add(buttonBox, BorderLayout.EAST);

        //  Create a menu bar for this window.
        JMenuBar menuBar = createMenuBar();
        this.setJMenuBar(menuBar);
    }

    /**
     * Initializes the menus associated with this window.
     */
    private JMenuBar createMenuBar() throws NoSuchMethodException {

        JMenuBar menuBar = new JMenuBar();

        // Set up the file menu.
        JMenu menu = AppUtilities.buildMenu(this, "File", FILEM_STRINGS);
        menuBar.add(menu);

        //  Add a Quit menu item (if it isn't automatically present).
        MDIApplication guiApp = MDIApplication.getInstance();
        if (!QuitJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(guiApp.getQuitJMenuItem());
        }

        // Set up the edit menu.
        menu = AppUtilities.buildMenu(this, "Edit", EDITM_STRINGS);
        menuBar.add(menu);

        //  Create a Window's menu.
        menu = MDIApplication.newWindowsMenu("Windows");
        menuBar.add(menu);

        //  Create an about menu item (if not automatically present).
        if (!AboutJMenuItem.isAutomaticallyPresent()) {
            //  Create Help menu.
            menu = new JMenu("Help");
            menuBar.add(menu);
            
            //  Add the "About" item to the Help menu.
            menu.add(guiApp.getAboutJMenuItem());
        }

        return menuBar;
    }

    /**
     * Have the user select a file for saving the data to.
     * 
     * @param extension The desired extension for the file.
     * @return The file the user has selected for saving or null if none was selected.
     */
    private File doSelectFile4Save(String extension) {
        
		// Does this data already have a file name/path associated with it?
        String fileName = "Untitled";
        String dir = null;
        if (lastFile != null) {
            dir = lastFile.getParent();
            fileName = lastFile.getName();
        }
        int idx = fileName.lastIndexOf(".");
        if (idx > 0)
            fileName = fileName.substring(0, idx);
        fileName += extension;
        
        //	Ask the user to select a file for saving.
        ExtFilenameFilter filter = new ExtFilenameFilter(extension, "Comma-Separated-Value table file");
        String fileExistsMsgFmt = "<html>A file with the name \"{0}\" already exists.<br>Do you want to replace it?</html>";
        File chosenFile = AppUtilities.selectFile4Save(TableWindow.this,
                "Choose a name for this data table file", dir, fileName, filter,
                extension, fileExistsMsgFmt, "Warning");
        
        if (chosenFile != null) {
            if (chosenFile.exists() && !chosenFile.canWrite()) {
                String canNotWrite2File = "<html>Can not write to the selected file.<br>Make sure it is not locked or in use by another program.</html>";
                JOptionPane.showMessageDialog(TableWindow.this,
                        canNotWrite2File, "I/O Error", JOptionPane.ERROR_MESSAGE);
            }
        }
        
        return chosenFile;
    }
    
    /**
     * Handle the user choosing "Save To CSV" from the File menu. Saves the
     * selected tables to a CSV formatted file.
     */
    public void handleSaveToCSV(ActionEvent event) {

        if (allTables == null)
            return;

        // Get a list of the items the user has selected.
        List<String> selection = tableList.getSelectedValuesList();
        if (selection.isEmpty())
            return;

        // Ask the user to select a file for saving the data to.
        File theFile = doSelectFile4Save(".csv");
        
        if (theFile != null) {
            // Export the tables to the selected file.
            exportFile(theFile, selection, CSV_READER);

            lastFile = theFile;
        }

    }

    /**
     * Handle the user choosing "Save To FDA" from the File menu. Saves the
     * selected tables to a FDA formatted file.
     */
    public void handleSaveToFDA(ActionEvent event) {

        if (allTables == null)
            return;

        // Get a list of the items the user has selected.
        List selection = tableList.getSelectedValuesList();
        if (selection.isEmpty())
            return;

        // Ask the user to select a file for saving the data to.
        File theFile = doSelectFile4Save(".fda");
        
        if (theFile != null) {

            // Export the tables to the selected file.
            exportFile(theFile, selection, FDA_READER);

            lastFile = theFile;
        }

    }

    /**
     * Handle the user choosing "Save To OTIS" from the File menu. Saves the
     * selected tables to a OTIS formatted table file.
     */
    public void handleSaveToOTIS(ActionEvent event) {

        if (allTables == null)
            return;

        // Get a list of the items the user has selected.
        List selection = tableList.getSelectedValuesList();
        if (selection.isEmpty())
            return;

        // Ask the user to select a file for saving the data to.
        File theFile = doSelectFile4Save(".itd");
        
        if (theFile != null) {

            // Export the tables to the selected file.
            exportFile(theFile, selection, OTIS_READER);

            lastFile = theFile;
        }

    }

    /**
     * Method that actually exports the selected tables to a disk file.
     */
    private void exportFile(File theFile, List<String> selection, FTableReader treader) {

        // Create a new table database containing only the selected tables.
        FTableList theTables = FTableList.newInstance();
        int length = selection.size();
        for (int i = 0; i < length; ++i)
            theTables.add(allTables.get(selection.get(i)));

        // Copy over the notes.
        theTables.addAllNotes(allTables.getAllNotes());

        FileOutputStream os = null;
        try {
            // Write out the selected tables to formatted table file.
            os = new FileOutputStream(theFile);
            treader.write(os, theTables);
            os.close();

        } catch (IOException e) {
            if (os != null)
                try {
                    os.close();
                } catch (IOException ex) { }
            theFile.delete();
            JOptionPane.showMessageDialog(this, e.getMessage(), "I/O Error",
                    JOptionPane.ERROR_MESSAGE);
        }
    }

    /**
     * Handle the user clicking on the "Remove" button. Removes the selected
     * tables from the list of currently open tables.
     */
    private void handleRemove() {

        if (!tableList.isSelectionEmpty()) {

            // Get a list of the items the user has selected.
            List<String> selection = tableList.getSelectedValuesList();
            int length = selection.size();
            StringBuilder buffer = new StringBuilder("Are you sure you want to remove these tables?\n    ");
            for (int i = 0; i < length; ++i) {
                buffer.append(selection.get(i));
                if (i < length - 1)
                    buffer.append(", ");
            }

            // Ask the user for confirmation.
            int result = JOptionPane.showConfirmDialog(this, buffer.toString(), "Warning",
                    JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE);
            if (result == JOptionPane.YES_OPTION) {

                // The user has confirmed, so remove the selected tables from database and list.
                for (int i = 0; i < length; ++i) {
                    String table = selection.get(i);
                    allTables.remove(table);
                    listModel.removeElement(table);
                }

                tableList.clearSelection();

                // Update the display.
                this.repaint();
            }
        }

    }

    /**
     * Handle the user clicking on the "View" button. Opens a window to allow
     * the user to view the contents of the table in a tabular fashion.
     */
    private void handleView() {
        if (!tableList.isSelectionEmpty()) {

            try {
                // Get a list of the items the user has selected.
                List<String> selection = tableList.getSelectedValuesList();
                int length = selection.size();

                for (int i = 0; i < length; ++i) {
                    // Create an instance of a view window for each selected table.
                    Float64Table theTable = allTables.get(selection.get(i));
                    TableViewWindow viewWindow = new TableViewWindow(theTable, lastFile);
                    viewWindow.pack();
                    AppUtilities.positionWindow(viewWindow, viewWindow.getWidth(), viewWindow.getHeight());
                    viewWindow.setVisible(true);
                    MDIApplication.addWindow(viewWindow);
                }

            } catch (NoSuchMethodException e) {
                System.err.println("Menu action listener method not found!");
                e.printStackTrace();
            }
        }

    }

    /**
     * Handle the user clicking on the "Lookup" button. Opens a lookup window
     * and lets the user interactively lookup values in the table.
     */
    private void handleLookup() {
        if (!tableList.isSelectionEmpty()) {

            try {

                // Get a list of the items the user has selected.
                List<String> selection = tableList.getSelectedValuesList();
                int length = selection.size();

                for (int i = 0; i < length; ++i) {
                    // Create an instance of a lookup window for each selected table.
                    Float64Table theTable = allTables.get(selection.get(i));
                    LookupWindow lookup = new LookupWindow(theTable);
                    lookup.setVisible(true);
                    MDIApplication.addWindow(lookup);
                }

            } catch (NoSuchMethodException e) {
                System.err.println("Menu action listener method not found!");
                e.printStackTrace();
            }

        }

    }

    /**
     * Handle the user clicking on the "Plot" button. Opens a plot window and
     * displays the currently selected table's data.
     */
    private void handlePlot() {
        if (!tableList.isSelectionEmpty()) {

            try {

                // Get a list of the items the user has selected.
                List<String> selection = tableList.getSelectedValuesList();
                int length = selection.size();

                for (int i = 0; i < length; ++i) {
                    // Create an instance of a plot window for each selected table.
                    Float64Table theTable = allTables.get(selection.get(i));
                    TablePlotWindow plot = new TablePlotWindow(theTable);
                    AppUtilities.positionWindow(plot, plot.getWidth(), plot.getHeight());
                    plot.setVisible(true);
                    MDIApplication.addWindow(plot);
                }

            } catch (NoSuchMethodException e) {
                System.err.println("Menu action listener method not found!");
                e.printStackTrace();
            }
        }

    }

    /**
     * Handle the user choosing "Close" from the File menu. This implementation
     * dispatches a "Window Closing" event to this window.
     */
    public void handleClose(ActionEvent event) {
        this.dispatchEvent(new WindowEvent(this, WindowEvent.WINDOW_CLOSING));
    }

    /**
     * Defines a button listener for our buttons.
     */
    private class ButtonL implements ActionListener {

        @Override
        public void actionPerformed(ActionEvent evt) {
            String cmd = evt.getActionCommand();
            
            switch (cmd) {
                case "Remove":
                    handleRemove();
                    break;
                case "Lookup":
                    handleLookup();
                    break;
                case "View":
                    handleView();
                    break;
                case "Plot":
                    handlePlot();
                    break;
                default:
                    break;
            }
           
        }

    }

}
