/*
*  VL0Reader  -- A class that enables reading of VL0 geometry files in AeroLS.
*
*   Copyright (C) 2002-2015 by Joseph A. Huwaldt
*   All rights reserved.
*   
*   This library is free software; you can redistribute it and/or
*   modify it under the terms of the GNU Lesser General Public
*   License as published by the Free Software Foundation; either
*   version 2.1 of the License, or (at your option) any later version.
*   
*   This library is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   Library General Public License for more details.
*
*  You should have received a copy of the GNU Lesser General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*  Or visit:  http://www.gnu.org/licenses/lgpl.html
*/

import java.io.*;
import java.text.*;
import javax.swing.JOptionPane;

import jahuwaldt.js.ParamTools.*;
import jahuwaldt.aero.panelgeom.*;
import javax.measure.converter.UnitConverter;
import javax.measure.quantity.Area;
import javax.measure.quantity.Length;
import javax.measure.unit.SI;
import javax.measure.unit.Unit;


/**
*  A class that reads in geometry data from a VL0 "GMETRY.DAT"
*  file.  This is a file format that defines reference
*  quantities and trapezoidal panel geometry.
*
*  <p>  Modified by:  Joseph A. Huwaldt   </p>
*
*  @author  Joseph A. Huwaldt   Date: June 24, 2002
*  @version June 20, 2015
*/
public class VL0Reader implements ParamNames {

    //  Debug flag.
    private static final boolean DEBUG = false;

    
	/**
	*  Method that parses in a VL0 DAT formatted text stream
	*  and creates a new model from it.
	*
	*  @param  inStream  The input stream to VL0 GMETRY.DAT file that we are reading from.
	*  @param  pdb       The parameter database containing the parameters for this project.
	*  @return A new ProjectData object with the data from the file filled in.
	*/
	public static ProjectData read(InputStream inStream, ParamDatabase pdb) throws IOException, Exception {

	    //  Create a new project data container.
	    ProjectData pdata = new ProjectData(pdb);
	    
        //  Create a component to hold the networks from the VL0 file.
	    GeomComponent comp = new GeomComponent("VL0 Geometry");
	    pdata.geometry.add(comp);
	    
		// Create a reader to access the ASCII file.
		LineNumberReader reader = new LineNumberReader(new InputStreamReader(inStream));


        //  Ask the user what units to use.
		Unit<Length> lengthUnit = askForUnits();
		Unit<Area> areaUnit = getAreaUnits(lengthUnit);
		
        try {
		    //  Begin parsing lines.

		    //  1st Line
	        DecimalFormat nf = (DecimalFormat)NumberFormat.getInstance();
	    	String data = reader.readLine();
		    ParsePosition position = new ParsePosition(0);

		    //  Parse out the reference area.
            double value = parseDouble(data, position, nf);
            Param param = pdb.get(SREF);
            param.setUnit(areaUnit);
            param.setValue(value);
            
            //  Parse out the reference chord length.
            value = parseDouble(data, position, nf);
            param = pdb.get(CREF);
            param.setUnit(lengthUnit);
            param.setValue(value);
            
            //  Parse out (and discard) the X location of the leading edge of cref.
            parseDouble(data, position, nf);
            
            //  Parse out the reference span length.
            value = parseDouble(data, position, nf);
            param = pdb.get(BREF);
            param.setUnit(lengthUnit);
            param.setValue(value);
            
            //  Parse out the location of the panel control point.
            value = parseDouble(data, position, nf);
            param = pdb.get(CP);
            param.setValue(value);
            
            //  Parse out the number of trapezoids used.
            int ntrap = parseInt(data, position, nf);

            if (DEBUG)
                System.out.println("" + pdb.get(SREF).getValue() + " " + pdb.get(CREF).getValue() +
                    " " + pdb.get(BREF).getValue() + " " + pdb.get(CP).getValue() +
                    " " + ntrap);
            
            
            //  Loop over all the trapezoids to be read in.
            for (int i=0; i < ntrap; ++i) {
                //  Read in 1st line of trapezoid data.
	    	    data = reader.readLine();
		        position.setIndex(0);

		        //  Parse out the trapezoid span.
                double BT = parseDouble(data, position, nf);
            
		        //  Parse out the trapezoid root chord.
                double CR = parseDouble(data, position, nf);

		        //  Parse out the trapezoid tip chord.
                double CT = parseDouble(data, position, nf);

		        //  Parse out the trapezoid leading edge sweep.
                double sweepLE = parseDouble(data, position, nf);

		        //  Parse out the trapezoid dihedral.
                double dihed = parseDouble(data, position, nf);

		        //  Parse out the X location of the root chord LE.
                double x0 = parseDouble(data, position, nf);

		        //  Parse out the Y location of the root chord LE.
                double y0 = parseDouble(data, position, nf);

		        //  Parse out the Z location of the root chord LE.
                double z0 = parseDouble(data, position, nf);
                
                //  Convert all the geometry to reference units from the input units.
                UnitConverter conv = lengthUnit.getConverterTo(SI.METER);
                BT = conv.convert(BT);
                CR = conv.convert(CR);
                CT = conv.convert(CT);
                x0 = conv.convert(x0);
                y0 = conv.convert(y0);
                z0 = conv.convert(z0);
                
                //  Read in 2nd line of trapezoid data.
 	    	    data = reader.readLine();
		        position.setIndex(0);

		        //  Parse out the number of panels in the chordwise direction.
                int nc = parseInt(data, position, nf);

		        //  Parse out the number of panels in the spanwise direction.
                //  The number of chordwise strips
                int nb = parseInt(data, position, nf);

                if (DEBUG) {
                    System.out.println("" + BT + " " + CR + " " + CT + " " + sweepLE +
                            " " + dihed + " " + x0 + " " + y0 + " " + z0);
                    System.out.println("" + nc + " " + nb);
                }


                //  Allocate memory for line positions.
                double[] cdist = new double[nc+1];
                double[] bdist = new double[nb+1];
                
                //  Read in the % trapezoid chord of the lines separating the panels.
                data = reader.readLine();
                position.setIndex(0);
		        for (int j=0; j <= nc; ++j) {
                    value = parseDouble(data, position, nf);
                    cdist[j] = value;
                    if (DEBUG)  System.out.print("" + value + " ");
		        }
		        if (DEBUG)  System.out.println();

                //  Read in the % trapezoid span of the lines separating the panels.
                data = reader.readLine();
                position.setIndex(0);
		        for (int j=0; j <= nb; ++j) {
                    value = parseDouble(data, position, nf);
                    bdist[j] = value;
                    if (DEBUG)  System.out.print("" + value + " ");
		        }
		        if (DEBUG)  System.out.println();


		        //  Create the network geometry and add it to the component.
		        GeomNetwork net = createNetwork(BT, CR, CT, sweepLE, dihed, x0, y0, z0,
		                                                cdist, bdist, i+1);
		        comp.add(net);
            }
            
            
        } catch (ParseException e) {
            System.out.println(e.getMessage());
            int lineNumber = reader.getLineNumber();
            reader.close();
            throw new IOException("Parsing error on line #" + lineNumber + ".");
        }

        return pdata;
	}


    /**
    *  Method that parses a double from the specified string at the specified position.
    */
    private static double parseDouble(String data, ParsePosition position, NumberFormat nf)
                                throws ParseException {
        int oldPos = position.getIndex();
        Number num = nf.parse(data, position);
        if (num == null) throw new ParseException("Null value returned for number in parseDouble().", oldPos);

        double value = num.doubleValue();

        //  Skip the trailing space.
        position.setIndex(position.getIndex()+1);

        return value;
    }
    
    /**
    *  Method that parses an integer from the specified string at the specified position.
    */
    private static int parseInt(String data, ParsePosition position, NumberFormat nf)
                                throws ParseException {
        int oldPos = position.getIndex();
        Number num = nf.parse(data, position);
        if (num == null) throw new ParseException("Null value returned for number in parseInt().", oldPos);

        int value = num.intValue();

        //  Skip the trailing space.
        position.setIndex(position.getIndex()+1);

        return value;
    }


    /**
    *  Method that takes the input geometry parameters and creates a paneled
    *  network geometry object.
    */
    private static GeomNetwork createNetwork(double b, double cr, double ct, double sweepLE,
                                    double dihed, double x0, double y0, double z0,
                                    double[] cdist, double[] bdist, int idx) {
 
        //  How many rows and columns of points are there?
        int rows = cdist.length;
        int cols = bdist.length;
        
        //  Allocate memory for arrays.
        double[][] xpnts = new double[cols][rows];
        double[][] ypnts = new double[cols][rows];
        double[][] zpnts = new double[cols][rows];

        //  Determine location of leading edge of tip chord.
        double x1 = x0 + b*Math.tan(Math.toRadians(sweepLE));
        double y1 = y0 + b;
        double z1 = z0 + b*Math.tan(Math.toRadians(dihed));
        
        //  Generate the points from inboard to outboard.
        for (int i=0; i < cols; ++i) {
            double dist = bdist[i]/100.0;
            double y = y0 + dist*(y1 - y0);
            double chord = cr + dist*(ct - cr);
            double xle = x0 + dist*(x1 - x0);
            double z = z0 + dist*(z1 - z0);
            
			//	Generate points from leading edge to trailing edge.
            for (int j=0; j < rows; ++j) {
                dist = cdist[j]/100.0;
                double x = xle + dist*chord;

                xpnts[i][j] = x;
                ypnts[i][j] = y;
                zpnts[i][j] = z;
            }
        }

        //  Create the network object.
        GeomNetwork net = new GeomNetwork(xpnts, ypnts, zpnts, "Network #" + idx);
        
        return net;
    }


    /**
    *  Ask the user to select from a list of unit types.
    */
    private static Unit<Length> askForUnits() throws InterruptedException {

        //  Extract the names of the length units from the list used by this program.
        int numUnits = LENGTH_UNITS.length;
        String[] names = new String[numUnits];
        for (int i=0; i < numUnits; ++i)
            names[i] = LENGTH_UNITS[i].toString();

        //  Ask the user to choose one.
        String selectedValue = (String)JOptionPane.showInputDialog(null,
                "Choose the units used in the file", "Input Units",
                JOptionPane.INFORMATION_MESSAGE, null,
                names, names[0]);

        //  Check for "cancel".
        if (selectedValue == null)
            throw new InterruptedException("User canceled.");
            
        //  Figure out which one the user chose.
        int idx = 0;
        for (int i=0; i < numUnits; ++i)
            if (selectedValue.equals(names[i])) {
                idx = i;
                break;
            }
        
        return LENGTH_UNITS[idx];
    }

    /**
    *  Method that matches up the appropriate area units to go
    *  with a set of length units.
    */
    private static Unit<Area> getAreaUnits(Unit<Length> lengthUnit) {
        int numUnits = LENGTH_UNITS.length;

        //  Find the length units in the list of length units used by this program.
        int idx = 0;
        for (int i=0; i < numUnits; ++i)
            if (lengthUnit == LENGTH_UNITS[i]) {
                idx = i;
                break;
            }

        //  The area units should be at the same index in the area unit list.
        return AREA_UNITS[idx];
    }
    
}

