/*
*  VLAnalysis  -- Class that actually implements the vortex lifting surface algorithm used in AeroLS.
*
*   Copyright (C) 2002-2024 by Joseph A. Huwaldt
*   All rights reserved.
*   
*   This library is free software; you can redistribute it and/or
*   modify it under the terms of the GNU Lesser General Public
*   License as published by the Free Software Foundation; either
*   version 2.1 of the License, or (at your option) any later version.
*   
*   This library is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   Library General Public License for more details.
*
*  You should have received a copy of the GNU Lesser General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*  Or visit:  http://www.gnu.org/licenses/lgpl.html
*/

import jahuwaldt.aero.AeroCoefXAft;
import jahuwaldt.aero.AeroCoefficients;
import jahuwaldt.aero.AeroReference;
import jahuwaldt.js.tables.FTableList;
import jahuwaldt.js.tables.Float64Table;
import jahuwaldt.swing.ProgressBarHandler;
import jahuwaldt.tools.math.LinAlg;
import java.util.ArrayList;
import java.util.List;
import javolution.util.FastTable;
import org.jscience.mathematics.number.Integer64;


/**
*  This class uses a vortex lifting surface algorithm
*  to determine the aerodynamic characteristics of the
*  specified lifting surface geometry.
*
*  <p>  Modified by:  Joseph A. Huwaldt   </p>
*
*  @author  Joseph A. Huwaldt,   Date: June 26, 2002
*  @version January 21, 2024
*/
public class VLAnalysis {

    //  Debugging flag.
    private static final boolean DEBUG = false;

    //  Constructor inputs held by this class.
    private final VLGeometry geom;
    private final AeroReference ref;
    private final double[] machBP;
    private final double[] alphaBP;
    private final double beta;
    private final double tol;
    private final double wakeTE;

    //  A reference to the progress bar used to display progress.
    private ProgressBarHandler progBar;
        
    
    /**
    *  Constructs an object to analyze the specified geometry.
    *
	*  @param  geom    A reference to the panel geometry for this model.
	*  @param  ref     Aerodynamic reference quantities for non-dimensionalization.
	*  @param  machBP  Mach number breakpoints for analysis.
	*  @param  alphaBP Angle of attack breakpoints for analysis in degrees.
	*  @param  beta    Angle of sideslip for analysis in degrees.
	*  @param  tol     Vortex singularity tolerance.  Tolerance for determining if
	*                  the extension of a bound vortex is to close to a control point.
	*  @param  wakeTE  X location of the trailing edges of the trailing wakes.
    */
    public VLAnalysis(VLGeometry geom, AeroReference ref, double[] machBP, double[] alphaBP,
	                    double beta, double tol, double wakeTE) {

        this.geom = geom;
        this.ref = ref;
        this.machBP = machBP;
        this.alphaBP = alphaBP;
        this.beta = beta;
        this.tol = tol;
        this.wakeTE = wakeTE;
        
    }

    
    /**
    *  Method that is called by the analysis thread to run the
    *  analysis.
    *
    *  @param  progBar Reference to a progress bar handler that displays progress for
    *                  long tasks.
	*  @param  pscale  Progress scaling factor between 0 and 1.  Indicates how long
	*                  this task should take relative to others for updating progress
	*                  bar.  Set to 1.0 if this task should take up the entire progress
	*                  bar.
    */
    public FTableList run(ProgressBarHandler progBar, float pscale) throws InterruptedException {

        this.progBar = progBar;

        //  Create a table database to hold our data.
        List<Float64Table> tdb = new ArrayList();
        
        //  Compute forces & moments.
		calcFM(geom, ref, machBP, alphaBP, beta, tol, wakeTE, pscale*0.5F, tdb);

        //  Compute stability & control derivatives at alpha=beta=0.
        calcStabDerivatives(geom, ref, machBP, tol, wakeTE, pscale*0.5F, tdb);

        FTableList tList = FTableList.valueOf(tdb);
        tList.addNote( "References:  " + ref );
        return tList;
	}


    /**
    *  Method that calculates stability and control derivatives at alpha=beta=0
    *  and returns a table database containing them.
    *
	*  @param  geom    A reference to the panel geometry for this model.
	*  @param  ref     Aerodynamic reference quantities for non-dimensionalization.
	*  @param  machBP  Mach number breakpoints to calculate derivatives at.
	*  @param  tol     Tolerance for determining if the extension of a bound vortex is
	*                  to close to a control point.
	*  @param  wakeTE  X location of the trailing edges of the trailing wakes.
	*  @param  pscale  Progress scaling factor between 0 and 1.  Indicates how long
	*                  this task should take relative to others for updating progress
	*                  bar.
	*  @param  tblLst  Reference to a list for holding data tables.
	*/
	private void calcStabDerivatives(VLGeometry geom, AeroReference ref, double[] machBP,
	                    double tol, double wakeTE, float pscale, List<Float64Table> tblLst) throws InterruptedException {

	    int symFactor = 1;
	    int symmetry = VLDownwash.NO_SYMM;


        //  Create blank tables that we will fill in with data.
		int numMBP = machBP.length;
	    Float64Table tableCLa = Float64Table.valueOf("CLalpha", "Mach", machBP, new double[numMBP]);
	    Float64Table tableCMa = Float64Table.valueOf("CMalpha", "Mach", machBP, new double[numMBP]);
	    Float64Table tableXac = Float64Table.valueOf("Xac", "Mach", machBP, new double[numMBP]);
	    Float64Table tableCLq = Float64Table.valueOf("CLq", "Mach", machBP, new double[numMBP]);
	    Float64Table tableCMq = Float64Table.valueOf("CMq", "Mach", machBP, new double[numMBP]);
	    Float64Table tableCYb = Float64Table.valueOf("CYbeta", "Mach", machBP, new double[numMBP]);
	    Float64Table tableCnb = Float64Table.valueOf("Cnbeta", "Mach", machBP, new double[numMBP]);
	    Float64Table tableClb = Float64Table.valueOf("Clbeta", "Mach", machBP, new double[numMBP]);
	    Float64Table tableCYr = Float64Table.valueOf("CYr", "Mach", machBP, new double[numMBP]);
	    Float64Table tableCnr = Float64Table.valueOf("Cnr", "Mach", machBP, new double[numMBP]);
	    Float64Table tableClr = Float64Table.valueOf("Clr", "Mach", machBP, new double[numMBP]);
	    Float64Table tableCYp = Float64Table.valueOf("CYp", "Mach", machBP, new double[numMBP]);
	    Float64Table tableCnp = Float64Table.valueOf("Cnp", "Mach", machBP, new double[numMBP]);
	    Float64Table tableClp = Float64Table.valueOf("Clp", "Mach", machBP, new double[numMBP]);

	    
	    //  Determine if we have a half or full model.
	    boolean hasSymmetry = geom.isMirrored();

	    //  Take advantage of the symmetry if we have a half-model.
	    if (hasSymmetry) {
	        geom.setMirrored(false);    //  Use half-model only.
	        symFactor = 2;
	        symmetry = VLDownwash.SYMMETRIC;
	    }
	    if (DEBUG)
	        System.out.println("hasSymmetry = " + hasSymmetry + ", symFactor = " +
	                            symFactor + ", symmetry = " + symmetry +
	                            ", numPanels = " + geom.getNumberOfPanels());

	    
		//	Allocate memory for arrays that we need.
		double[] daNdx = new double[geom.getNumberOfPanels()];
		double[] dloadNdx = new double[geom.getNumberOfPanels()];
        List<Integer64> pos = FastTable.newInstance();
        pos.add(Integer64.ZERO);
		
		//	Loop over the Mach number breakpoints.
		for (int i=0; i < numMBP; ++i) {
			pos.set(0, Integer64.valueOf(i));
		
			//	Generate a (possibly symmetric) downwash matrix for this Mach number.
			VLDownwash dw = new VLDownwash(symmetry, machBP[i], geom, tol, wakeTE);


			//  **** Alpha Derivatives ****
			
		    //  Calculate the slope of the angle of attack normal to the plane of each panel
			//  with respect to free stream angle of attack: d(alphaN)/d(alpha).
			daNdx = calcdAlphaNdAlpha(0, 0, geom, daNdx);
			
			//	Calculate the load coefficient slope normal to each panel.  This destroys the contents of daNdx.
			dloadNdx = calcLoadNCoef(daNdx, dw, geom, ref.getRefAreaRU(), dloadNdx);
			
			//	Calculate the normal force (and lift force: alpha=0) coefficient slope.
			double CNx = symFactor*calcCN(dloadNdx, geom);
			tableCLa.set(pos, CNx);

            //  Calculate the pitching moment coefficient slope.
			double CMx = symFactor*calcCM(dloadNdx, geom, ref);
			tableCMa.set(pos, CMx);

            //  Calculate the aerodynamic center X location.
			double Xac = ref.getMomentRefXRU() - ref.getRefLengthRU()*CMx/CNx;
			tableXac.set(pos, Xac);

			if (DEBUG) {
			    double r2di = Math.PI/180;
			    System.out.println("mach = " + machBP[i] + ", CNa = " + CNx*r2di +
			        ", CMa = " + CMx*r2di + ", Xac = " + Xac);
			}

			
            //  **** Pitch Rate Derivatives ****
			
		    //  Calculate the slope of the angle of attack normal to the plane of each panel
			//  with respect to pitch rate angle: d(alphaN)/d(qhat).
			daNdx = calcdAlphaNdqhat(geom, ref, daNdx);
			
			//	Calculate the load coefficient slope normal to each panel.  This destroys the contents of daNdx.
			dloadNdx = calcLoadNCoef(daNdx, dw, geom, ref.getRefAreaRU(), dloadNdx);
			
			//	Calculate the normal force (and lift force: alpha=0) coefficient slope.
			CNx = symFactor*calcCN(dloadNdx, geom);
			tableCLq.set(pos, CNx);

            //  Calculate the pitching moment coefficient slope.
			CMx = symFactor*calcCM(dloadNdx, geom, ref);
			tableCMq.set(pos, CMx);

			if (DEBUG) {
			    double r2di = Math.PI/180;
			    System.out.println("mach = " + machBP[i] + ", CNq = " + CNx*r2di +
			        ", CMq = " + CMx*r2di);
			}

			
			//  ****  Sideslip Derivatives ****

			//	Generate an anti-symmetric downwash matrix for this Mach number
			//  if the previous downwash matrix was symmetric.
			if (hasSymmetry)
			    dw = new VLDownwash(VLDownwash.ANTI_SYMM, machBP[i], geom, tol, wakeTE);
			
		    //  Calculate the slope of the angle of attack normal to the plane of each panel
			//  with respect to free stream sideslip angle: d(alphaN)/d(beta).
			daNdx = calcdAlphaNdBeta(0, 0, geom, daNdx);
			
			//	Calculate the load coefficient slope normal to each panel.  This destroys the contents of daNdx.
			dloadNdx = calcLoadNCoef(daNdx, dw, geom, ref.getRefAreaRU(), dloadNdx);
			
			//	Calculate the side force coefficient slope.
			double CYx = symFactor*calcCY(dloadNdx, geom);
			tableCYb.set(pos, CYx);

			//	Calculate the yawing moment coefficient slope.
			double Cnx = symFactor*calcCn(dloadNdx, geom, ref);
			tableCnb.set(pos, Cnx);

			//	Calculate the rolling moment coefficient slope.
			double Clx = symFactor*calcCl(dloadNdx, geom, ref);
			tableClb.set(pos, Clx);

			if (DEBUG) {
			    double r2di = Math.PI/180;
			    System.out.println("        CYb = " + CYx*r2di + ", Cnb = " + Cnx*r2di +
			        ", Clb = " + Clx*r2di);
			}

			
			//  ****  Yaw Rate Derivatives ****

		    //  Calculate the slope of the angle of attack normal to the plane of each panel
			//  with respect to yaw rate angle: d(alphaN)/d(rhat).
			daNdx = calcdAlphaNdrhat(geom, ref, daNdx);
			
			//	Calculate the load coefficient slope normal to each panel.  This destroys the contents of daNdx.
			dloadNdx = calcLoadNCoef(daNdx, dw, geom, ref.getRefAreaRU(), dloadNdx);
			
			//	Calculate the side force coefficient slope.
			CYx = symFactor*calcCY(dloadNdx, geom);
			tableCYr.set(pos, CYx);

			//	Calculate the yawing moment coefficient slope.
			Cnx = symFactor*calcCn(dloadNdx, geom, ref);
			tableCnr.set(pos, Cnx);

			//	Calculate the rolling moment coefficient slope.
			Clx = symFactor*calcCl(dloadNdx, geom, ref);
			tableClr.set(pos, Clx);

			if (DEBUG) {
			    double r2di = Math.PI/180;
			    System.out.println("        CYr = " + CYx*r2di + ", Cnr = " + Cnx*r2di +
			        ", Clr = " + Clx*r2di);
			}

			
			//  ****  Roll Rate Derivatives ****

		    //  Calculate the slope of the angle of attack normal to the plane of each panel
			//  with respect to roll rate angle: d(alphaN)/d(phat).
			daNdx = calcdAlphaNdphat(geom, ref, daNdx);
			
			//	Calculate the load coefficient slope normal to each panel.  This destroys the contents of daNdx.
			dloadNdx = calcLoadNCoef(daNdx, dw, geom, ref.getRefAreaRU(), dloadNdx);
			
			//	Calculate the side force coefficient slope.
			CYx = symFactor*calcCY(dloadNdx, geom);
			tableCYp.set(pos, CYx);

			//	Calculate the yawing moment coefficient slope.
			Cnx = symFactor*calcCn(dloadNdx, geom, ref);
			tableCnp.set(pos, Cnx);

			//	Calculate the rolling moment coefficient slope.
			Clx = symFactor*calcCl(dloadNdx, geom, ref);
			tableClp.set(pos, Clx);

			if (DEBUG) {
			    double r2di = Math.PI/180;
			    System.out.println("        CYp = " + CYx*r2di + ", Cnp = " + Cnx*r2di +
			        ", Clp = " + Clx*r2di);
			}

			
			//	Update progress bar.
			incrementProgress(pscale/numMBP);
		}


		//  Add the filled in tables to the table database.
		tblLst.add(tableCLa);
		tblLst.add(tableCMa);
		tblLst.add(tableXac);
		tblLst.add(tableCYb);
		tblLst.add(tableCnb);
		tblLst.add(tableClb);
		tblLst.add(tableCLq);
		tblLst.add(tableCMq);
		tblLst.add(tableCYr);
		tblLst.add(tableCnr);
		tblLst.add(tableClr);
		tblLst.add(tableCYp);
		tblLst.add(tableCnp);
		tblLst.add(tableClp);

		//  Be kind, rewind.
		if (hasSymmetry)  geom.setMirrored(true);
		
	}


    /**
    *  Method that calculates configuration forces and moments and returns a table database
    *  containing them.
    *
	*  @param  geom    A reference to the panel geometry for this model.
	*  @param  ref     Aerodynamic reference quantities for non-dimensionalization.
	*  @param  machBP  Mach number breakpoints to calculate forces & moments at.
	*  @param  alphaBP Angle of attack breakpoints for analysis in degrees.
	*  @param  beta    Angle of sideslip for analysis in degrees.
	*  @param  tol     Tolerance for determining if the extension of a bound vortex is
	*                  to close to a control point.
	*  @param  wakeTE  X location of the trailing edges of the trailing wakes.
	*  @param  pscale  Progress scaling factor between 0 and 1.  Indicates how long
	*                  this task should take relative to others for updating progress
	*                  bar.
	*  @param  tblLst  Reference to a database for holding data tables.
	*/
	private void calcFM(VLGeometry geom, AeroReference ref, double[] machBP, double[] alphaBP, double beta,
	                    double tol, double wakeTE, float pscale, List<Float64Table> tblLst) throws InterruptedException {
        
	    int symFactor = 1;
	    int symmetry = VLDownwash.NO_SYMM;

        //  Create blank tables that we will fill in with data.
		int numMBP = machBP.length;
	    int numAOABP = alphaBP.length;
	    String[] indepNames = {"Mach", "Alpha"};
	    Float64Table tableCN = Float64Table.valueOf("CN", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableCA = Float64Table.valueOf("CA", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableCYb = Float64Table.valueOf("CYb", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableCMb = Float64Table.valueOf("CMb", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableCnb = Float64Table.valueOf("Cnb", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableClb = Float64Table.valueOf("Clb", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableCL = Float64Table.valueOf("CL", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableCD = Float64Table.valueOf("CD", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableCY = Float64Table.valueOf("CY", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableCM = Float64Table.valueOf("CM", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableCn = Float64Table.valueOf("Cn", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableCl = Float64Table.valueOf("Cl", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableCLSA = Float64Table.valueOf("CLSA", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableCDSA = Float64Table.valueOf("CDSA", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableCYSA = Float64Table.valueOf("CYSA", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableCMSA = Float64Table.valueOf("CMSA", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableCnSA = Float64Table.valueOf("CnSA", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);
	    Float64Table tableClSA = Float64Table.valueOf("ClSA", indepNames, machBP, alphaBP, new double[numMBP][numAOABP]);


        //  Create an aero coefficients object to use for converting between
	    //  body and wind axes.
	    AeroCoefficients coef = new AeroCoefXAft(ref);

	    
	    //  Determine if we have a symmetric or asymmetric model.
	    boolean hasSymmetry = geom.isMirrored() && beta == 0;

	    //  Take advantage of the symmetry if we have a half-model and no sideslip.
	    if (hasSymmetry) {
	        geom.setMirrored(false);    //  Use half-model only.
	        symFactor = 2;
	        symmetry = VLDownwash.SYMMETRIC;
	    }
	    if (DEBUG)
	        System.out.println("hasSymmetry = " + hasSymmetry + ", symFactor = " +
	                            symFactor + ", symmetry = " + symmetry +
	                            ", numPanels = " + geom.getNumberOfPanels());
	    

		//	Allocate memory for arrays that we need.
		double[] alphaN = new double[geom.getNumberOfPanels()];
		double[] loadNCoef = new double[geom.getNumberOfPanels()];
        List<Integer64> pos = FastTable.newInstance();
        pos.add(Integer64.ZERO);
        pos.add(Integer64.ZERO);

		//	Loop over the Mach number breakpoints.
		for (int i=0; i < numMBP; ++i) {
		    pos.set(0, Integer64.valueOf(i));
		    
			//	Generate the (possibly symmetric) downwash matrix for this Mach number.
			VLDownwash dw = new VLDownwash(symmetry, machBP[i], geom, tol, wakeTE);
			
			//	Update progress bar.
			incrementProgress(0.2F*pscale/numMBP);
			

			//  Loop over the angle of attack breakpoints.
			for (int j=0; j < numAOABP; ++j) {
                pos.set(1, Integer64.valueOf(j));
                
			    //  Calculate the angle of attack normal to the plane of each panel.
				alphaN = calcAlphaN(alphaBP[j], beta, geom, alphaN);
				
				//	Calculate the load coefficient normal to each panel.  This destroys the contents of alphaN.
				loadNCoef = calcLoadNCoef(alphaN, dw, geom, ref.getRefAreaRU(), loadNCoef);
				
				//	Calculate the normal force coefficient.
				double CN = symFactor*calcCN(loadNCoef, geom);
				tableCN.set(pos, CN);

				//	Calculate the axial force coefficient.
				double CA = symFactor*calcCA(loadNCoef, geom);
				tableCA.set(pos, CA);

				//	Calculate the side force coefficient.
				double CYb = 0;
				if (!hasSymmetry)
				    CYb = symFactor*calcCY(loadNCoef, geom);
				tableCYb.set(pos, CYb);

				//  Convert body axes to wind axes.
				coef.setBodyAxesForces(CN, CA, CYb);
				coef.setAngles(Math.toRadians(alphaBP[j]), Math.toRadians(beta), 0);
				double CL = coef.getCL();
				double CD = coef.getCD();
				double CY = coef.getCY();
				tableCL.set(pos, CL);
				tableCD.set(pos, CD);
				tableCY.set(pos, CY);
				
				//  Convert body axes to stability axes.
				CL = coef.getCLSA();
				CD = coef.getCDSA();
				CY = coef.getCYSA();
				tableCLSA.set(pos, CL);
				tableCDSA.set(pos, CD);
				tableCYSA.set(pos, CY);
				
                //  Calculate the pitching moment coefficient.
			    double CMb = symFactor*calcCM(loadNCoef, geom, ref);
			    tableCMb.set(pos, CMb);
			    

                //  Calculate the yawing moment coefficient.
			    double Cnb = 0;
				if (!hasSymmetry)
			        Cnb = symFactor*calcCn(loadNCoef, geom, ref);
			    tableCnb.set(pos, Cnb);
			    
                //  Calculate the rolling moment coefficient.
			    double Clb = 0;
				if (!hasSymmetry)
			        Clb = symFactor*calcCl(loadNCoef, geom, ref);
			    tableClb.set(pos, Clb);
			    
				//  Convert body axes to wind axes.
				coef.setBodyAxesMoments(CMb, Cnb, Clb);
				double CM = coef.getCM();
				double Cn = coef.getCn();
				double Cl = coef.getCl();
				tableCM.set(pos, CM);
				tableCn.set(pos, Cn);
				tableCl.set(pos, Cl);

				//  Convert body axes to stability axes.
				CM = coef.getCMSA();
				Cn = coef.getCnSA();
				Cl = coef.getClSA();
				tableCMSA.set(pos, CM);
				tableCnSA.set(pos, Cn);
				tableClSA.set(pos, Cl);
				
				
				if (DEBUG) {
				    System.out.println("mach = " + machBP[i] + ", alpha = " + alphaBP[j] + ", beta = " + beta);
				    System.out.println("       CN =  " + CN  + ", CA = "  + CA  + ", CMb = " + CMb);
				    System.out.println("       CYb = " + CYb + ", Cnb = " + Cnb + ", Clb = " + Clb);
				}
				
				//	Update progress bar.
				incrementProgress(0.8F*pscale/numMBP/numAOABP);
				
			}   //  Next j
			
		}   //  Next i

		//  Add the filled in tables to the table database.
		tblLst.add(tableCN);
		tblLst.add(tableCA);
		tblLst.add(tableCYb);
		tblLst.add(tableCMb);
		tblLst.add(tableCnb);
		tblLst.add(tableClb);
		tblLst.add(tableCL);
		tblLst.add(tableCD);
		tblLst.add(tableCY);
		tblLst.add(tableCM);
		tblLst.add(tableCn);
		tblLst.add(tableCl);
		tblLst.add(tableCLSA);
		tblLst.add(tableCDSA);
		tblLst.add(tableCYSA);
		tblLst.add(tableCMSA);
		tblLst.add(tableCnSA);
		tblLst.add(tableClSA);
        
		//  Be kind, rewind.
		if (hasSymmetry)  geom.setMirrored(true);
	}


	/**
	*  Method that calculates the "angle of attack" normal to the plane of each panel.
	*  The "normal angle of attack" is the direction of the free stream flow dotted with the
	*  local normal vector of each panel:  alphaN[n] = (Vinf/|Vinf|) dot n[n].
	*
	*  @param  alpha  The angle of attack in degrees.
	*  @param  beta   The angle of sideslip in degrees.
	*  @param  geom   A reference to the panel geometry for this model.
	*  @param  alphaN A reference to an existing array of size geom.getNumberOfPanels() that
	*                 will be filled in with the local alphaN for each panel.
	*  @return A reference to the input array alphaN is returned.
	*/
	private static double[] calcAlphaN(double alpha, double beta, VLGeometry geom, double[] alphaN) {
	
		//	Convert degrees to radians.
        alpha = Math.toRadians(alpha);
		beta = Math.toRadians(beta);
		
		//	Compute direction cosines of the free stream velocity vector:
		//  Vinf/|Vinf| = (Ui + Vj + Wk)/|Vinf|.
		double calpha = Math.cos(alpha);
		double UoVinf = calpha*Math.cos(beta);
		double VoVinf = -calpha*Math.sin(beta);
		double WoVinf = Math.sin(alpha);
		
		//	Loop over all the panels in the model.
		int numPanels = alphaN.length;
		for (int i=0; i < numPanels; ++i) {
		
			//	Get the normal vector of this panel.
			double nx = geom.getNormalX(i);
			double ny = geom.getNormalY(i);
			double nz = geom.getNormalZ(i);
			
			//	Compute (Vinf/|Vinf|) dot n[i].
			alphaN[i] = UoVinf*nx + VoVinf*ny + WoVinf*nz;
		}
		
		return alphaN;
	}
	
	
	/**
	*  Method that calculates the slope of the "angle of attack" normal to the plane of each panel
	*  with respect to the free stream angle of attack:  d(alphaN)/d(alpha)
	*
	*  @param  alpha  The angle of attack in degrees.
	*  @param  beta   The angle of sideslip in degrees.
	*  @param  geom   A reference to the panel geometry for this model.
	*  @param  daNda  A reference to an existing array of size geom.getNumberOfPanels() that
	*                 will be filled in with the local d(alphaN)/d(alpha) for each panel.
	*  @return A reference to the input array daNda is returned.
	*/
	private static double[] calcdAlphaNdAlpha(double alpha, double beta, VLGeometry geom, double[] daNda) {
	
		//	Convert degrees to radians.
        alpha = Math.toRadians(alpha);
		beta = Math.toRadians(beta);
		
		//	Compute direction cosines of the free stream velocity vector derivative:
		//  d(Vinf/|Vinf|)/da = (dU/dai + dV/daj + dW/dak)/|Vinf|.
		double salpha = Math.sin(alpha);
		double dUdaoVinf = -salpha*Math.cos(beta);
		double dVdaoVinf = salpha*Math.sin(beta);
		double dWdaoVinf = Math.cos(alpha);
		
		//	Loop over all the panels in the model.
		int numPanels = daNda.length;
		for (int i=0; i < numPanels; ++i) {
		
			//	Get the normal vector of this panel.
			double nx = geom.getNormalX(i);
			double ny = geom.getNormalY(i);
			double nz = geom.getNormalZ(i);
			
			//	Compute d(Vinf/|Vinf|)/da dot n[i].
			daNda[i] = dUdaoVinf*nx + dVdaoVinf*ny + dWdaoVinf*nz;
		}
		
		return daNda;
	}
	
	/**
	*  Method that calculates the slope of the "angle of attack" normal to the plane of each panel
	*  with respect to the free stream angle of sideslip:  d(alphaN)/d(beta)
	*
	*  @param  alpha  The angle of attack in degrees.
	*  @param  beta   The angle of sideslip in degrees.
	*  @param  geom   A reference to the panel geometry for this model.
	*  @param  daNdb  A reference to an existing array of size geom.getNumberOfPanels() that
	*                 will be filled in with the local d(alphaN)/d(alpha) for each panel.
	*  @return A reference to the input array daNda is returned.
	*/
	private static double[] calcdAlphaNdBeta(double alpha, double beta, VLGeometry geom, double[] daNdb) {
	
		//	Convert degrees to radians.
        alpha = Math.toRadians(alpha);
		beta = Math.toRadians(beta);
		
		//	Compute direction cosines of the free stream velocity vector derivative:
		//  d(Vinf/|Vinf|)/db = (dU/dbi + dV/dbj + dW/dbk)/|Vinf|.
		double calpha = Math.cos(alpha);
		double dUdboVinf = -calpha*Math.sin(beta);
		double dVdboVinf = -calpha*Math.cos(beta);
//		double dWdboVinf = 0;
		
		//	Loop over all the panels in the model.
		int numPanels = daNdb.length;
		for (int i=0; i < numPanels; ++i) {
		
			//	Get the normal vector of this panel.
			double nx = geom.getNormalX(i);
			double ny = geom.getNormalY(i);
//			double nz = geom.getNormalZ(i);
			
			//	Compute d(Vinf/|Vinf|)/db dot n[i].
			daNdb[i] = dUdboVinf*nx + dVdboVinf*ny + 0;
		}
		
		return daNdb;
	}
	
		
	/**
	*  Method that calculates the slope of the "angle of attack" normal to the plane of each panel
	*  with respect to the pitch rate angle:  d(alphaN)/d(qhat)
	*
	*  @param  geom   A reference to the panel geometry for this model.
	*  @param  ref    Aerodynamic reference quantities for non-dimensionalization.
	*  @param  daNdq  A reference to an existing array of size geom.getNumberOfPanels() that
	*                 will be filled in with the local d(alphaN)/d(qhat) for each panel.
	*  @return A reference to the input array daNdq is returned.
	*/
	private static double[] calcdAlphaNdqhat(VLGeometry geom, AeroReference ref, double[] daNdq) {
	
	    double xcg = ref.getMomentRefXRU();
	    double cref = ref.getRefLengthRU();

		
		//	Loop over all the panels in the model.
		int numPanels = daNdq.length;
		for (int i=0; i < numPanels; ++i) {

		    //  Get the X location of the 1/4 MGC for this panel.
			double xmgc4 = geom.getMGCQuarterChordX(i);
			
		    //	Compute direction cosines of the free stream velocity vector derivative:
		    //  d(Vinf/|Vinf|)/dq = (dU/dqi + dV/dqj + dW/dqk)/|Vinf|.
//		    double dUdqoVinf = 0;
//		    double dVdqoVinf = 0;
		    double dWdqoVinf = 2.0/cref*(xmgc4 - xcg);
		    
			//	Get the Z component of the normal vector of this panel.
			double nz = geom.getNormalZ(i);
			
			//	Compute d(Vinf/|Vinf|)/dq dot n[i].
			daNdq[i] = 0 + 0 + dWdqoVinf*nz;
		}
		
		return daNdq;
	}
	
		
	/**
	*  Method that calculates the slope of the "angle of attack" normal to the plane of each panel
	*  with respect to the yaw rate angle:  d(alphaN)/d(rhat)
	*
	*  @param  geom   A reference to the panel geometry for this model.
	*  @param  ref    Aerodynamic reference quantities for non-dimensionalization.
	*  @param  daNdr  A reference to an existing array of size geom.getNumberOfPanels() that
	*                 will be filled in with the local d(alphaN)/d(rhat) for each panel.
	*  @return A reference to the input array daNdr is returned.
	*/
	private static double[] calcdAlphaNdrhat(VLGeometry geom, AeroReference ref, double[] daNdr) {
	
	    double xcg = ref.getMomentRefXRU();
	    double ycg = ref.getMomentRefYRU();
	    double bref = ref.getRefSpanRU();

		
		//	Loop over all the panels in the model.
		int numPanels = daNdr.length;
		for (int i=0; i < numPanels; ++i) {

		    //  Get the location of the 1/4 MGC for this panel.
			double xmgc4 = geom.getMGCQuarterChordX(i);
			double ymgc4 = geom.getMGCQuarterChordY(i);
			
		    //	Compute direction cosines of the free stream velocity vector derivative:
		    //  d(Vinf/|Vinf|)/dr = (dU/dri + dV/drj + dW/drk)/|Vinf|.
		    double dUdroVinf = -2/bref*(ymgc4 - ycg);
		    double dVdroVinf = -2/bref*(xmgc4 - xcg);
//		    double dWdroVinf = 0;
		    
			//	Get the normal vector of this panel.
			double nx = geom.getNormalX(i);
			double ny = geom.getNormalY(i);
			
			//	Compute d(Vinf/|Vinf|)/dr dot n[i].
			daNdr[i] = dUdroVinf*nx + dVdroVinf*ny + 0;
		}
		
		return daNdr;
	}
	
		
	/**
	*  Method that calculates the slope of the "angle of attack" normal to the plane of each panel
	*  with respect to the roll rate angle:  d(alphaN)/d(phat)
	*
	*  @param  geom   A reference to the panel geometry for this model.
	*  @param  ref    Aerodynamic reference quantities for non-dimensionalization.
	*  @param  daNdp  A reference to an existing array of size geom.getNumberOfPanels() that
	*                 will be filled in with the local d(alphaN)/d(phat) for each panel.
	*  @return A reference to the input array daNdp is returned.
	*/
	private static double[] calcdAlphaNdphat(VLGeometry geom, AeroReference ref, double[] daNdp) {
	
	    double ycg = ref.getMomentRefYRU();
	    double zcg = ref.getMomentRefZRU();
	    double bref = ref.getRefSpanRU();

		
		//	Loop over all the panels in the model.
		int numPanels = daNdp.length;
		for (int i=0; i < numPanels; ++i) {

		    //  Get the location of the 1/4 MGC for this panel.
			double ymgc4 = geom.getMGCQuarterChordY(i);
			double zmgc4 = geom.getMGCQuarterChordZ(i);
			
		    //	Compute direction cosines of the free stream velocity vector derivative:
		    //  d(Vinf/|Vinf|)/dp = (dU/dpi + dV/dpj + dW/dpk)/|Vinf|.
//		    double dUdpoVinf = 0;
		    double dVdpoVinf = -2/bref*(zmgc4 - zcg);
		    double dWdpoVinf = 2/bref*(ymgc4 - ycg);
		    
			//	Get the the normal vector of this panel.
			double ny = geom.getNormalY(i);
			double nz = geom.getNormalZ(i);
			
			//	Compute d(Vinf/|Vinf|)/dr dot n[i].
			daNdp[i] = 0 + dVdpoVinf*ny + dWdpoVinf*nz;
		}
		
		return daNdp;
	}
	
		
	/**
	*  Calculate the load coefficient normal to each panel surface:  CLN[n] = LN[i]/(qbar*Sref).
	*
	*  @param  alphaN  Array containing the local angle of attack normal to each panel: alphaN[numPanels].
	*                  The contents of this array are destroyed by this method.
	*  @param  dw      A reference to the downwash matrix object.
	*  @param  geom    A reference to the panel geometry for this model.
	*  @param  Sref    The reference area used to non-dimensionalize this model in reference units (square Meters).
	*  @param  loadN   A reference to an existing array of size geom.getNumberOfPanels() that
	*                  will be filled in with the local normal load coefficient for each panel.
	*  @return A reference to the input array loadN is returned.
	*/
	private static double[] calcLoadNCoef(double[] alphaN, VLDownwash dw, VLGeometry geom, double Sref, double[] loadN) {
	
		//	Extract a copy of the downwash matrix.
		double[][] W = dw.getMatrix();
		
		//	Calculate the right hand side of the matrix equation W*x = B.
		double[] B = alphaN;		//	B points to or is aliased to the same array (same memory location) as alphaN.
		int numPanels = alphaN.length;
		for (int i=0; i < numPanels; ++i)
			B[i] = 2.0/Sref*geom.getSpan(i)*alphaN[i];

        //  Solve the system of equations:	W*x = B.
        int[] ipvt = new int[numPanels];
        LinAlg.gefa(W, numPanels, numPanels, ipvt);
		LinAlg.gesl(W, numPanels, numPanels, ipvt, B, 0);
		
		//	Copy results into the output array.
		System.arraycopy(B,0, loadN,0, numPanels);
		
		return loadN;
	}
	
	
	/**
	*  Method that calculates the total body axis normal force coefficient given the distributed
	*  normal load coefficient.
	*
	*  @param  loadN  An array containing the local panel normal force coefficient (normal to panel).
	*  @param  geom   A reference to the panel geometry for this model.
	*/
	private static double calcCN(double[] loadN, VLGeometry geom) {
		
		double CN = 0;
		int numPanels = loadN.length;
		for (int i=0; i < numPanels; ++i) {
			double nz = geom.getNormalZ(i);
			CN += loadN[i]*nz;
		}
		
		return CN;
	}
	

	/**
	*  Method that calculates the total body axis axial force coefficient given the distributed
	*  normal load coefficient.
	*
	*  @param  loadN  An array containing the local panel normal force coefficient (normal to panel).
	*  @param  geom   A reference to the panel geometry for this model.
	*/
	private static double calcCA(double[] loadN, VLGeometry geom) {
		
		double CA = 0;
		int numPanels = loadN.length;
		for (int i=0; i < numPanels; ++i) {
			double nx = geom.getNormalX(i);
			CA += loadN[i]*nx;
		}
		
		return CA;
	}
	

	/**
	*  Method that calculates the total body axis side force coefficient given the distributed
	*  normal load coefficient.
	*
	*  @param  loadN  An array containing the local panel normal force coefficient (normal to panel).
	*  @param  geom   A reference to the panel geometry for this model.
	*/
	private static double calcCY(double[] loadN, VLGeometry geom) {
		
		double CY = 0;
		int numPanels = loadN.length;
		for (int i=0; i < numPanels; ++i) {
			double ny = geom.getNormalY(i);
			CY += loadN[i]*ny;
		}
		
		return CY;
	}
	

	/**
	*  Method that calculates the total pitching moment coefficient given the distributed
	*  normal load coefficient.
	*
	*  @param  loadN  An array containing the local panel normal force coefficient (normal to panel).
	*  @param  geom   A reference to the panel geometry for this model.
	*  @param  ref    Aerodynamic reference quantities for non-dimensionalization.
	*/
	private static double calcCM(double[] loadN, VLGeometry geom, AeroReference ref) {

	    double xcg = ref.getMomentRefXRU();
	    
		double CM = 0;
		int numPanels = loadN.length;
		for (int i=0; i < numPanels; ++i) {
			double nz = geom.getNormalZ(i);
			double xmgc4 = geom.getMGCQuarterChordX(i);
			CM -= loadN[i]*nz*(xmgc4 - xcg);
		}
		
	    double cref = ref.getRefLengthRU();
		return CM/cref;
	}
	
	
	/**
	*  Method that calculates the total yawing moment coefficient given the distributed
	*  normal load coefficient.
	*
	*  @param  loadN  An array containing the local panel normal force coefficient (normal to panel).
	*  @param  geom   A reference to the panel geometry for this model.
	*  @param  ref    Aerodynamic reference quantities for non-dimensionalization.
	*/
	private static double calcCn(double[] loadN, VLGeometry geom, AeroReference ref) {

	    double xcg = ref.getMomentRefXRU();
	    
		double Cn = 0;
		int numPanels = loadN.length;
		for (int i=0; i < numPanels; ++i) {
			double ny = geom.getNormalY(i);
			double xmgc4 = geom.getMGCQuarterChordX(i);
			Cn -= loadN[i]*ny*(xmgc4 - xcg);
		}
		
	    double bref = ref.getRefSpanRU();
		return Cn/bref;
	}
	
	/**
	*  Method that calculates the total rolling moment coefficient given the distributed
	*  normal load coefficient.
	*
	*  @param  loadN  An array containing the local panel normal force coefficient (normal to panel).
	*  @param  geom   A reference to the panel geometry for this model.
	*  @param  ref    Aerodynamic reference quantities for non-dimensionalization.
	*/
	private static double calcCl(double[] loadN, VLGeometry geom, AeroReference ref) {

        double ycg = ref.getMomentRefYRU();
	    double zcg = ref.getMomentRefZRU();
	    
		double Cl = 0;
		int numPanels = loadN.length;
		for (int i=0; i < numPanels; ++i) {
			double ny = geom.getNormalY(i);
			double nz = geom.getNormalZ(i);
			double ymgc4 = geom.getMGCQuarterChordY(i);
			double zmgc4 = geom.getMGCQuarterChordZ(i);
			Cl += ( ny*(zmgc4 - zcg) - nz*(ymgc4 - ycg) )*loadN[i];
		}
		
	    double bref = ref.getRefSpanRU();
		return Cl/bref;
	}
	

    /**
    *  Increment the display of the progress bar by the
    *  specified amount.
    */
    private void incrementProgress(float inc) throws InterruptedException {
        if (progBar != null) {
	        if (progBar.isCanceled()) throw new InterruptedException("User canceled.");
	    
            progBar.setProgress(progBar.getProgress() + inc);
        }
    }
    
}


