/*
*  VLDownwash  -- Object that computes and stores the downwash matrix used in the AeroLS program.
*
*   Copyright (C) 2002-2015 by Joseph A. Huwaldt
*   All rights reserved.
*   
*   This library is free software; you can redistribute it and/or
*   modify it under the terms of the GNU Lesser General Public
*   License as published by the Free Software Foundation; either
*   version 2.1 of the License, or (at your option) any later version.
*   
*   This library is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   Library General Public License for more details.
*
*  You should have received a copy of the GNU Lesser General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*  Or visit:  http://www.gnu.org/licenses/lgpl.html
*/

import java.io.PrintStream;
import java.text.NumberFormat;
import jahuwaldt.util.ExponentialFormat;


/**
*  A class that computes and contains the downwash matrix
*  used in a vortex lattice analysis.
*
*  <p>  Modified by:  Joseph A. Huwaldt   </p>
*
*  @author  Joseph A. Huwaldt   Date: June 22, 2002
*  @version June 23, 2015
*/
public class VLDownwash {

    /**
    *  Symmetric configuration flag.  Use this for a clean symmetric half-model
    *  or for a half-model with symmetrically deflected flaps.
    */
    public static final int SYMMETRIC = 1;

    /**
    *  Anti-symmetric configuration flag.  Use this for a half-model with
    *  asymmetrically deflected flaps (such as ailerons).
    */
    public static final int ANTI_SYMM = -1;

    /**
    *  No symmetry configuration flag. Use this for a full (left/right) model
    *  where no symmetry is desired.
    */
    public static final int NO_SYMM = 0;

    
    //  Indicies into 3D vectors.
    private static final int X = 0;
    private static final int Y = 1;
    private static final int Z = 2;

    //  The downwash matrix this object creates/contains.
    private final double[][] W;

    
    /**
    *  Construct a downwash matrix using the specified conditions
    *  and geometry.
    *
    *  @param  symmetry  Symmetry flag.  Can have any value.  Normally, the constants
    *                    defined by this class are used (Symmetric=1, Anti-Symmetric=-1
    *                    and No Symmetry = 0).
    *  @param  mach      The Mach number for the cases being run.  The geometry is
    *                    scaled by Mach number to get an "equivalent" M=0 solution.
	*  @param   geometry Reference to the processed geometry parameters for this model.
	*  @param  tol       Tolerance for determining if the extension of a bound vortex is
	*                    to close to a control point.
	*  @param  wakeTE    X location of the trailing edges of the trailing wakes.
    */
    public VLDownwash(int symmetry, double mach, VLGeometry geometry, double tol, double wakeTE) {
        int k;

		//	Don't compute left hand side if there is no symmetry.
		int lastk = 1;
		if (symmetry == NO_SYMM)
			lastk = 0;
		
        //  Allocate memory for the downwash matrix.
		int numPanels = geometry.getNumberOfPanels();
        W = new double[numPanels][numPanels];
        
		
        //  Allocate memory for local vectors.
        double[] r0 = new double[3];
        double[] r1 = new double[3];
        double[] r2 = new double[3];
        double[] v = new double[3];
		double[] w = new double[3];
		double[] n = new double[3];
        
		
        //  Generate the downwash matrix.
        
        //  Loop over all the panel control points.
        for (int i=0; i < numPanels; ++i) {

			//	Get the unit normal for panel "i".
			n[X] = geometry.getNormalX(i);
			n[Y] = geometry.getNormalY(i);
			n[Z] = geometry.getNormalZ(i);

            //  Loop over all the panels that are influencing the C.P. of panel "i".
            for (int j=0; j < numPanels; ++j) {
		
				//	One pass for the right hand side and one for the left
                //  (if symmetry != 0).
                for (k=0; k <= lastk; ++k) {
				
					//	Compute the total induced velocity at the control point of panel "i"
					//	due to unit circulation of all the vorticies associated with panel "j".
					v = totalInducedVelocity(i, j, mach, wakeTE, tol, k, geometry,
					                                        r0, r1, r2, v);
					
                    //  Find the total induced flow due to both right and left hand sides of model.
                    if (k==0) {
                        w[X] = v[X];
                        w[Y] = v[Y];
                        w[Z] = v[Z];
						
                    } else {
                        w[X] += symmetry*v[X];
                        w[Y] += symmetry*v[Y];
                        w[Z] += symmetry*v[Z];
					}
					
					if (k == lastk)
                        //  Place the total induced flow normal to panel "i" in the downwash matrix.
                        W[j][i] = -dot(w, n);
					
				}
				
			}	//	Next j
		}	//	Next i
		
    }


    /**
    *  Returns a copy of the actual downwash matrix computed
    *  by this object.  A copy is returned, not the original.
    */
    public double[][] getMatrix() {
		int size = W.length;
		double[][] matrix = new double[size][size];
		for (int i=0; i < size; ++i)
			System.arraycopy(W[i],0, matrix[i],0, size);
			
        return matrix;
    }

    /**
    *  Method that prints out the downwash matrix.
    */
    public void print(PrintStream out) {
        ExponentialFormat nf = new ExponentialFormat();
        nf.setMaximumFractionDigits(6);
		nf.setMinimumFractionDigits(6);
		nf.setMaximumIntegerDigits(1);
		nf.setMinimumExponentDigits(2);

        int numPanels = W.length;
		out.println(numPanels);

		for (int i=0; i < numPanels; ++i) {
		    for (int j=0; j < numPanels; ++j)
                out.print(formatNumber(14, nf, W[j][i]));
		    out.println();
		}
        
    }
    

    /**
    *  Method that computes the total induced velocity at a control point
    *  of panel "n" (in 3 components) due to a unit circulation of all the vortices
	*  associated with panel "m".
    *
	*  @param  n   The index of the panel we are finding the induced velocity on.
	*  @param  m   The index of the panel containing the vortices we are want the influence from.
    *  @param  mach    The Mach number for the cases being run.  The geometry is
    *                  scaled by Mach number to get an "equivalent" M=0 solution.
	*  @param  wakeTE  The X location of the trailing edge of any trailing wakes.
	*  @param  tol     Tolerance for determining if the extension of a bound vortex is
	*                  to close to a control point.
	*  @param  k       Equal to 0 if we are calculating the influence of the right hand side
	*                  (or full configuration) panels and equal to 1 if we are calculating
	*                  the influence of the left hand side panels due to symmetry.
	*  @param  geom    A reference to the geometry for this model.
    *  @param  r0  Three element array used internally by this method but allocated externally (for performance).
    *  @param  r1  Three element array used internally by this method but allocated externally (for performance).
    *  @param  r2  Three element array used internally by this method but allocated externally (for performance).
    *  @param  v   Three element array that will contain the resulting total induced 
    *              velocity components on exit.  Any existing data in this array will be
    *              destroyed.
    *  @return A reference to the input array "v" is returned.  Will contain components of total
	*          induced velocity on C.P. of panel "n" due to unit strength vortices of panel "m".
    */
    private static double[] totalInducedVelocity(int n, int m, double mach, double wakeTE, double tol, int k, VLGeometry geom,
													double[] r0, double[] r1, double[] r2, double[] v) {
		
		//	Get the Mach correction factor.
		double mbeta = prandtlGlauert(mach);
        
		//	Get the geometry we need.
		double Xp = geom.getXp(n)*mbeta;
		double Yp = geom.getYp(n);
		double Zp = geom.getZp(n);
		double Xbo = geom.getXbo(m)*mbeta;
		double Ybo = geom.getYbo(m);
		double Zbo = geom.getZbo(m);
		double Xbi = geom.getXbi(m)*mbeta;
		double Ybi = geom.getYbi(m);
		double Zbi = geom.getZbi(m);
		double XTEi = geom.getXTEi(m)*mbeta;
		double YTEi = geom.getYTEi(m);
		double ZTEi = geom.getZTEi(m);
		double XTEo = geom.getXTEo(m)*mbeta;
		double YTEo = geom.getYTEo(m);
		double ZTEo = geom.getZTEo(m);
		
		if (k == 1) {
			//  If left hand side due to symmetry, flop the horseshoe vortex to the other side.
			double temp = Ybo;
			Ybo = -Ybi;
			Ybi = -temp;
			temp = Zbo;
			Zbo = Zbi;
			Zbi = temp;
			temp = Xbo;
			Xbo = Xbi;
			Xbi = temp;

			temp = XTEo;
			XTEo = XTEi;
			XTEi = temp;
			temp = YTEo;
			YTEo = -YTEi;
			YTEi = -temp;
			temp = ZTEo;
			ZTEo = ZTEi;
			ZTEi = temp;
		}

		//  Compute geometry of the bound vortex.
		r0[X] = Xbo - Xbi;
		r0[Y] = Ybo - Ybi;
		r0[Z] = Zbo - Zbi;
		r1[X] = Xp - Xbi;
		r1[Y] = Yp - Ybi;
		r1[Z] = Zp - Zbi;
		r2[X] = Xp - Xbo;
		r2[Y] = Yp - Ybo;
		r2[Z] = Zp - Zbo;

        //  Avoid a possible singularity.
		//  If we are far away from the vortex line, calculate the induced velocity.
		//  If we are close to the line, only calculate the induced velocity if
		//      we are also directly behind (or in front) the vortex segment itself.
		//  Note:  cos(theta1) = dotProduct(r0,r1)/mr0/mr1.
		//         cos(theta2) = dotProduct(r0,r2)/mr0/mr2.
		//         rp = magnitude(crossProduct(r1,r2,v))/mr0.
        double mr0 = norm(r0);
        if (norm(cross(r1,r2,v))/mr0 > tol ||
                (dot(r0,r1)/mr0/norm(r1) > 0 &&
                        dot(r0,r2)/mr0/norm(r2) < 0) )
                        
            //  Induced velocity at C.P. due to unit ciruclation of bound vortex.
            v = inducedVelocity(r0, r1, r2, v);

        else {
            v[X] = 0;
            v[Y] = 0;
            v[Z] = 0;
        }
        double wx = v[X];
        double wy = v[Y];
        double wz = v[Z];


        //  Compute geometry of the inboard side vortex.		
        r0[X] = Xbi - XTEi;
        r0[Y] = Ybi - YTEi;
        r0[Z] = Zbi - ZTEi;
		r2[X] = r1[X];
		r2[Y] = r1[Y];
		r2[Z] = r1[Z];
        r1[X] = Xp - XTEi;
        r1[Y] = Yp - YTEi;
        r1[Z] = Zp - ZTEi;
        
		//  Induced velocity at C.P. due to unit ciruclation of inboard side vortex.
		v = inducedVelocity(r0, r1, r2, v);
		wx += v[X];
		wy += v[Y];
		wz += v[Z];
        
        
		//  Compute geometry of the outboard side vortex.
		r0[X] = XTEo - Xbo;
		r0[Y] = YTEo - Ybo;
		r0[Z] = ZTEo - Zbo;
		r1[X] = Xp - Xbo;
		r1[Y] = Yp - Ybo;
		r1[Z] = Zp - Zbo;
		r2[X] = Xp - XTEo;
		r2[Y] = Yp - YTEo;
		r2[Z] = Zp - ZTEo;

		//  Induced velocity at C.P. due to unit ciruclation of outboard side vortex.
		v = inducedVelocity(r0, r1, r2, v);
		wx += v[X];
		wy += v[Y];
		wz += v[Z];


        //  Calculate induced velocity due to trailing side-edge vorticies.
        int next = geom.getNextAft(m);
		while (next > -1) {
		    double XLEi = geom.getXLEi(next)*mbeta;
		    double YLEi = geom.getYLEi(next);
		    double ZLEi = geom.getZLEi(next);
		    XTEi = geom.getXTEi(next)*mbeta;
		    YTEi = geom.getYTEi(next);
		    ZTEi = geom.getZTEi(next);
			
		    double XLEo = geom.getXLEo(next)*mbeta;
		    double YLEo = geom.getYLEo(next);
		    double ZLEo = geom.getZLEo(next);
		    XTEo = geom.getXTEo(next)*mbeta;
		    YTEo = geom.getYTEo(next);
		    ZTEo = geom.getZTEo(next);
			
			if (k == 1) {
				//  If left hand side due to symmetry, flop the horseshoe vortex to the other side.
				double temp = XLEo;
				XLEo = XLEi;
				XLEi = temp;
				temp = YLEo;
				YLEo = -YLEi;
				YLEi = -temp;
				temp = ZLEo;
				ZLEo = ZLEi;
				ZLEi = temp;
				
				temp = XTEo;
				XTEo = XTEi;
				XTEi = temp;
				temp = YTEo;
				YTEo = -YTEi;
				YTEi = -temp;
				temp = ZTEo;
				ZTEo = ZTEi;
				ZTEi = temp;
			}

            //  Compute geometry of the inboard side.
		    r0[X] = XLEi - XTEi;
		    r0[Y] = YLEi - YTEi;
		    r0[Z] = ZLEi - ZTEi;
		    r1[X] = Xp - XTEi;
		    r1[Y] = Yp - YTEi;
		    r1[Z] = Zp - ZTEi;
		    r2[X] = Xp - XLEi;
		    r2[Y] = Yp - YLEi;
		    r2[Z] = Zp - ZLEi;
		    
		    //  Induced velocity at C.P. due to unit ciruclation of inboard side vortex.
		    v = inducedVelocity(r0, r1, r2, v);
		    wx += v[X];
		    wy += v[Y];
		    wz += v[Z];

            //  Compute geometry of the outboard side.
		    r0[X] = XTEo - XLEo;
		    r0[Y] = YTEo - YLEo;
		    r0[Z] = ZTEo - ZLEo;
		    r1[X] = Xp - XLEo;
		    r1[Y] = Yp - YLEo;
		    r1[Z] = Zp - ZLEo;
		    r2[X] = Xp - XTEo;
		    r2[Y] = Yp - YTEo;
		    r2[Z] = Zp - ZTEo;
		    
		    //  Induced velocity at C.P. due to unit ciruclation of inboard side vortex.
		    v = inducedVelocity(r0, r1, r2, v);
		    wx += v[X];
		    wy += v[Y];
		    wz += v[Z];

		    next = geom.getNextAft(next);
		}


		//  Compute geometry of the inboard trailing vortex.
		r0[X] = XTEi - wakeTE;
		r0[Y] = 0;
		r0[Z] = 0;
		r1[X] = Xp - wakeTE;
		r1[Y] = Yp - YTEi;
		r1[Z] = Zp - ZTEi;
		r2[X] = Xp - XTEi;
		r2[Y] = r1[Y];
		r2[Z] = r1[Z];

		//  Induced velocity at C.P. due to unit ciruclation of inboard tailing vortex.
		v = inducedVelocity(r0, r1, r2, v);
		wx += v[X];
		wy += v[Y];
		wz += v[Z];
		
		//  Compute geometry of the outboard trailing vortex.
		r0[X] = wakeTE - XTEo;
		r0[Y] = 0;
		r0[Z] = 0;
		r2[X] = r1[X];
		r2[Y] = Yp - YTEo;
		r2[Z] = Zp - ZTEo;
		r1[X] = Xp - XTEo;
		r1[Y] = r2[Y];
		r1[Z] = r2[Z];

		//  Induced velocity at C.P. due to unit ciruclation of outboard trailing vortex.
		v = inducedVelocity(r0, r1, r2, v);
		
		//	Compute total induced velocity for output.
		v[X] += wx;
		v[Y] += wy;
		v[Z] += wz;
		
		return v;
	}
	

    /**
    *  Method that computes the induced velocity at an arbitrary control point
    *  (in 3 components) due to a unit circulation about a vortex segment
    *  defined by the vector r0.  This method solves Eqn. 6.37a in Bertin & Smith
    *  for a unit circulation (Gamma = 1.0).
    *
    *  @param  r0  A 3D vector defining the location of the vortex segment in space.
    *  @param  r1  A 3D vector from the start of the vortex segment to the C.P.
    *  @param  r2  A 3D vector from the end of the vortex segment to the C.P.
    *  @param  v   An existing 3D vector that will contain the resulting induced 
    *              velocity components.  Any existing data in this array will be
    *              destroyed.
    *  @return A reference to the input array "v" is returned.
    */
    private static double[] inducedVelocity(double[] r0, double[] r1, double[] r2, double[] v) {

		if (norm(r0) > 0.00001) {
		
			double mr1i = 1/norm(r1);
			double mr2i = 1/norm(r2);
	
			//  Compute Fac2ab.
			v[X] = r1[X]*mr1i;
			v[Y] = r1[Y]*mr1i;
			v[Z] = r1[Z]*mr1i;
			double dp1 = dot(r0, v);
	
			v[X] = r2[X]*mr2i;
			v[Y] = r2[Y]*mr2i;
			v[Z] = r2[Z]*mr2i;
			double dp2 = dot(r0, v);
	
			double fac2ab = dp1 - dp2;
	
			//  Compute r1 X r2.
			v = cross(r1, r2, v);
			double mr1xr22 = norm(v);
			mr1xr22 *= mr1xr22;
	
			//  Compute induced velocity.
			double factor = fac2ab/mr1xr22/4/Math.PI;
			v[X] *= factor;
			v[Y] *= factor;
			v[Z] *= factor;
        
		} else {
		    //  Avoids divide-by-zero caused when r1 and r2 are coincident
		    //  (which happens when r0 has zero length).
			v[X] = v[Y] = v[Z] = 0;
		}
		
        return v;
    }

	/**
	*  Returns the length or magnitude of a 3D vector.
	*
	*  @param  vector  A 3D (3 element) vector.
	*  @return The length of the input vector.
	*/
	private static double norm(double[] vector) {
		double dx = vector[X];
		double dy = vector[Y];
		double dz = vector[Z];
		return Math.sqrt(dx*dx + dy*dy + dz*dz);
	}

	/**
	*  Returns the dot product of two 3D vectors.
	*
	*  @param a  The 1st 3D (3 element) vector.
	*  @param b  The 2nd 3D (3 element) vector.
	*  @return The dot product of the two vectors.
	*/
	private static double dot(double[] a, double[] b) {
		return (a[X]*b[X] + a[Y]*b[Y] + a[Z]*b[Z]);
	}

	/**
	*  Returns the right-handed cross product of 3D vectors A and B in AcrossB.
	*
	*  @param A  The 1st 3D (3 element) vector.
	*  @param B  The 2nd 3D (3 element) vector.
	*  @param AcrossB  A 3D (3 element) vector to be filled in with the cross
	*                  product of A and B.  This vector will be overwritten
	*                  by this method.
	*  @return A reference to the input vector AcrossB.
	*/
	private static double[] cross(double[] A, double[] B, double[] AcrossB) {
		AcrossB[X] = A[Y] * B[Z] - A[Z] * B[Y];
		AcrossB[Y] = A[Z] * B[X] - A[X] * B[Z];
		AcrossB[Z] = A[X] * B[Y] - A[Y] * B[X];
		return AcrossB;
	}

	/**
	*  Prandtl-Glauert compressibility correction.
	*  A compressibility correction that relates incompressible
	*  flow over a given 2D profile to subsonic
	*  compressible flow over the same profile.  It approximately
	*  corrects incompressible data for compressibility effects.
	*
	*  @param  M     The subsonic Mach number you want to scale
	*                the coef0 to.
	*  @return The correction factor that will correct for compressibility effects.
	*  @throws IllegalArgumentException if the Mach number is not
	*          subsonic.
	*/
	private static double prandtlGlauert(double M) {
        if (M >= 1.0)
            throw new IllegalArgumentException("Mach must be subsonic.");
		return 1.0/Math.sqrt(1.0 - M*M);
	}

	/**
	*  Format a number for output into a fixed width field.
	*  This method formats the given number using the supplied
	*  NumberFormat object.  It then adds spaces to the start
	*  of the formatted number until the string reaches the
	*  specified length.
	*
	*  @param  size   The overall length of the formatted number including
	*                 the decimal point, minus sign, "E" notation, etc.
	*  @param  nf     The NumberFormat to use when formatting this number.
	*  @param  number The number to be formatted.
	*/
	private static String formatNumber( int size, NumberFormat nf, double number ) {
		StringBuilder buffer = new StringBuilder( nf.format( number ) );
		addSpaces( size, buffer );
		return buffer.toString();
	}


	/**
	*  Method that adds spaces to the start of the formatted number
	*  (or any other string actually) until the string reaches the
	*  specified length.
	*
	*  @param  size   The overall length of the resulting string.  Spaces
	*                 will be added to the input string until it reaches
	*                 this length.
	*  @param  input  The input string buffer to have the spaces added onto.
	*  @returns The input string buffer is modified, but a reference to it
	*           is also returned.
	*/
	private static StringBuilder addSpaces( int size, StringBuilder input ) {
		int length = input.length();
		while ( length < size ) {
			input.insert( 0, " " );
			length += 1;
		}
		return input;
	}

}


