/*
*  VLGeometry  -- Class that computes geometry specifically required by the AeroLS program.
*
*   Copyright (C) 2002-2015 by Joseph A. Huwaldt
*   All rights reserved.
*   
*   This library is free software; you can redistribute it and/or
*   modify it under the terms of the GNU Lesser General Public
*   License as published by the Free Software Foundation; either
*   version 2.1 of the License, or (at your option) any later version.
*   
*   This library is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   Library General Public License for more details.
*
*  You should have received a copy of the GNU Lesser General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*  Or visit:  http://www.gnu.org/licenses/lgpl.html
*/

import jahuwaldt.aero.panelgeom.*;


/**
*  An object that contains all the geometry data required to
*  conduct a vortex lattice analysis.  This class serves as
*  an interface between the generic paneled geometry contained
*  in GeomVehicle and the geometry requirements of a vortex
*  lattice or lifting surface theory analysis.  The input geometry
*  may be either a half-model (set the mirrored flag to true) or a
*  full configuration geometry (set the mirrored flag to false).
*
*  <p>  Modified by:  Joseph A. Huwaldt   </p>
*
*  @author  Joseph A. Huwaldt,   Date: June 22, 2002
*  @version June 20, 2015
*/
public class VLGeometry {

    //  Flag indicating that we are mirroring the input geometry to the opposite side.
    private boolean mirror = false;

    //  The number of panels stored internally.
    private final int numPanels;
    
    //  Reference to the network that each panel came from.
    private final GeomNetwork[] nets;

    //  Indexes to the panel in the network it came from.
    private final int[] string, point;
    
	//	Index plus 1 of the next panel aft of the panel i.  A value of 0 means panel i is at the TE.
	private final int[] nextAft;
	
    //  X, Y, Z location of the control points for each panel.
    private final double[] px, py, pz;
	
    //  Location of the inboard edge of the 1/4 chord of each panel.
    private final double[] Xbi, Ybi, Zbi;
	
	//	Location of the outboard edge of the 1/4 chord of each panel.
	private final double[] Xbo, Ybo, Zbo;

	//	Average span of each panel.
	private final double[] b;
	
	//	Location of the 1/4 mean geometric chord (MGC) of each panel.
	private final double[] Xmgc4, Ymgc4, Zmgc4;
	
	
    /**
    *  Construct a VLGeometry object using the specified input data.
    *
    *  @param  vehicle  A list of geometry components that make up the vehicle to be analyzed.
	*                   Must contain at least one component with at least one network with at
	*                   least one panel to create a VLGeometry object.
    *  @param  CP       Location of the control point on the aerodynamic panels as
    *                   fraction of panel mean geometric chord.  Use 0.75 unless
    *                   experimental data are available to indicate otherwise.
    *  @param  mirrored Set to true to indicate that the geometry represents a half-model
    *                   that should be mirrored to the opposite side for analysis.  Set
    *                   to false to analyze the geometry as input.
    */
    public VLGeometry(GeomVehicle vehicle, double CP, boolean mirrored) {

        mirror = mirrored;
        
		int totPanels = vehicle.getNumberOfPanels();
		this.numPanels = totPanels;
		
		//	Allocate memory for the various arrays that we need.
		nets = new GeomNetwork[totPanels];
		string = new int[totPanels];
		point = new int[totPanels];
		nextAft = new int[totPanels];
		px = new double[totPanels];
		py = new double[totPanels];
		pz = new double[totPanels];
		Xbi = new double[totPanels];
		Ybi = new double[totPanels];
		Zbi = new double[totPanels];
		Xbo = new double[totPanels];
		Ybo = new double[totPanels];
		Zbo = new double[totPanels];
		b = new double[totPanels];
		Xmgc4 = new double[totPanels];
		Ymgc4 = new double[totPanels];
		Zmgc4 = new double[totPanels];
		
		
		//	Loop over all the panels in all the networks in all the components in the specified vehicle.
		//	Collect together all the data from these panels into the arrays just created.
		int pos=0;
		int numComponents = vehicle.size();
		for (int c=0; c < numComponents; ++c) {
			GeomComponent comp = (GeomComponent)vehicle.get(c);
			int numNetworks = comp.size();
			
			for (int n=0; n < numNetworks; ++n) {
				GeomNetwork net = (GeomNetwork)comp.get(n);
				int numCols = net.getNumStrings()-1;
				int numRows = net.getNumPointsPerString()-1;
				double[][] pntsX = net.getAllX();
				double[][] pntsY = net.getAllY();
				double[][] pntsZ = net.getAllZ();
				
				//	Loop over all the panels in this network.
				for (int i=0; i < numCols; ++i) {
					for (int j=0; j < numRows; ++j) {

					    //  Store off reference information.
					    nets[pos] = net;
					    string[pos] = i;
					    point[pos] = j;

                        //  Extract the inboard LE location.
					    double x0 = pntsX[i][j];
					    double y0 = pntsY[i][j];
					    double z0 = pntsZ[i][j];
					    
						//	Extract the inboard TE location.
						int jp1 = j + 1;
						double x1 = pntsX[i][jp1];
						double y1 = pntsY[i][jp1];
						double z1 = pntsZ[i][jp1];
						
						//	Calculate the location of the inboard end of the bound vortex.
						Xbi[pos] = pointOnLine(x0, x1, 0.25);
						Ybi[pos] = pointOnLine(y0, y1, 0.25);
						Zbi[pos] = pointOnLine(z0, z1, 0.25);

						//	Extract the outboard TE location.
						int ip1 = i + 1;
						double x2 = pntsX[ip1][jp1];
						double y2 = pntsY[ip1][jp1];
						double z2 = pntsZ[ip1][jp1];

						//  Extract the outboard LE location.
						double x3 = pntsX[ip1][j];
						double y3 = pntsY[ip1][j];
						double z3 = pntsZ[ip1][j];
						
						//	Calculate the location of the outboard edge of the bound vortex.
						Xbo[pos] = pointOnLine(x3, x2, 0.25);
						Ybo[pos] = pointOnLine(y3, y2, 0.25);
						Zbo[pos] = pointOnLine(z3, z2, 0.25);
						
						//	Calculate the control point location.
						//	First on the inboard and outboard edges.
						double xt1 = pointOnLine(x0, x1, CP);
						double yt1 = pointOnLine(y0, y1, CP);
						double zt1 = pointOnLine(z0, z1, CP);
						double xt2 = pointOnLine(x3, x2, CP);
						double yt2 = pointOnLine(y3, y2, CP);
						double zt2 = pointOnLine(z3, z2, CP);
						
						//	Then in the middle.
						px[pos] = pointOnLine(xt1, xt2, 0.5);
						py[pos] = pointOnLine(yt1, yt2, 0.5);
						pz[pos] = pointOnLine(zt1, zt2, 0.5);

						//  Calculate the location of the 1/4 chord of the MGC of the panel.
						//  Find point along 1/4 chord line (bound vortex line) that corresponds
						//  with the MGC span location.
						xt1 = x1 - x0;
						yt1 = y1 - y0;
						zt1 = z1 - z0;
						double cr = Math.sqrt(xt1*xt1 + yt1*yt1 + zt1*zt1);
						xt1 = x2 - x3;
						yt1 = y2 - y3;
						zt1 = z2 - z3;
						double ct = Math.sqrt(xt1*xt1 + yt1*yt1 + zt1*zt1);
                        if (cr >= ct) {
                            double taper = ct/cr;
                            double eta = (1 + 2*taper)/(1 + taper)/3;
						    Xmgc4[pos] = pointOnLine(Xbi[pos], Xbo[pos], eta);
						    Ymgc4[pos] = pointOnLine(Ybi[pos], Ybo[pos], eta);
						    Zmgc4[pos] = pointOnLine(Zbi[pos], Zbo[pos], eta);

                        } else {
                            //  Avoiding divide by zero in ct/cr when cr = 0.
                            //  This can occure at the tip of a port side
                            //  delta wing for instance.
                            double taper = cr/ct;
                            double eta = (1 + 2*taper)/(1 + taper)/3;
						    Xmgc4[pos] = pointOnLine(Xbo[pos], Xbi[pos], eta);
						    Ymgc4[pos] = pointOnLine(Ybo[pos], Ybi[pos], eta);
						    Zmgc4[pos] = pointOnLine(Zbo[pos], Zbi[pos], eta);
                        }
						
						//	Calculate the average true span of the panel (not projected).
                        if (cr >= ct) {
                            //  Span of outboard leading edge point.
                            xt1 = distPoint2Line(x0,y0,z0, x1,y1,z1, x3,y3,z3);

                            //  Span of outboard trailing edge point.
                            xt2 = distPoint2Line(x0,y0,z0, x1,y1,z1, x2,y2,z2);

                            //  Average span.
                            b[pos] = (xt1 + xt2)/2;

                        } else {
                            //  Avoiding divide by zero in distPoint2Line() when cr = 0.
                            //  This can occure at the tip of a port side
                            //  delta wing for instance.
                            xt1 = distPoint2Line(x3,y3,z3, x2,y2,z2, x0,y0,z0);
                            xt2 = distPoint2Line(x3,y3,z3, x2,y2,z2, x1,y1,z1);
                            b[pos] = (xt1 + xt2)/2;
                        }
						
						//	Record the index (plus 1) of the next panel aft from the one we are working on.
						if (j < numRows-1)
							nextAft[pos] = pos + 1 + 1;
						
						//	Increment our overall panel index.
						++pos;
					}
				}
				
				
			}	//	Next n
			
		}	//	Next c
		
    }


	/**
	*  Method that returns a value that is some fraction of the distance
	*  between the input values.
	*
	*  @param a  The 1st value.
	*  @param b  The 2nd value.
	*  @param fraction  The fraction of the distance between the two input
	*                   values.  For example, 0.25F returns a value that is
	*                   1/4 of the way between a and b.
	*/
	private static double pointOnLine(double a, double b, double fraction) {
		return ((b - a)*fraction + a);
	}
	

    /**
    *  Method that calculates the shortest distance between a point
    *  and a line in 3D space.  Make sure the line does not have
    *  zero length (points a and b being coincident) before calling this
    *  method or a divide-by-zero exception will be thrown!
    *
    *  @param  xa, ya, za  The coordinates of a point on the line.
    *  @param  xb, yb, zb  The coordinates of another point on the line.
    *  @param  xc, yc, zc  The coordinates of the point we are finding the
    *                      shortest distance to.
    */
    private double distPoint2Line(double xa, double ya, double za, double xb, double yb, double zb,
                                        double xc, double yc, double zc) {
        double r0x = xb - xa;
        double r0y = yb - ya;
        double r0z = zb - za;
        
        double r1x = xc - xa;
        double r1y = yc - ya;
        double r1z = zc - za;

        double r1dotr0 = r1x*r0x + r1y*r0y + r1z*r0z;
        double r0dotr0 = r0x*r0x + r0y*r0y + r0z*r0z;
        double b1 = r1dotr0/r0dotr0;

        double dx = r1x - b1*r0x;
        double dy = r1y - b1*r0y;
        double dz = r1z - b1*r0z;

        double rp = Math.sqrt(dx*dx + dy*dy + dz*dz);

        return rp;
    }
                                    
	/**
	*  Method that sets the mirroring flag indicating if the input geometry
	*  should be mirrored to the opposite side.
	*/
	public void setMirrored(boolean flag) {
        mirror = flag;
	}

	/**
	*  Method that returns true if the input geometry is being mirrored to
	*  the opposite side.
	*/
	public boolean isMirrored() {
        return mirror;
	}

	
	/**
	*  Returns the number of panels in this model.
	*/
	public final int getNumberOfPanels() {
	    int nP = this.numPanels;
	    if (mirror) nP *= 2;
		return nP;
	}
	
	
	/**
	*  Returns surface area of the specified panel.
	*
	*  @param idx  Index of the panel to return the surface area for.
	*  @return The surface area of the specified panel.
	*/
	public final double getArea(int idx) {
	    if (idx >= numPanels && mirror)
	        idx -= numPanels;
		return nets[idx].getArea(string[idx], point[idx]);
	}
	
	/**
	*  Returns the average span of the specified panel.
	*
	*  @param idx  Index of the panel to return the average span for.
	*  @return The average span of the specified panel.
	*/
	public final double getSpan(int idx) {
	    if (idx >= numPanels && mirror)
	        idx -= numPanels;
		return b[idx];
	}
	
	/**
	*  Returns the X location of the 1/4 chord of the MGC of the specified panel.
	*
	*  @param idx  Index of the panel to return the average span for.
	*  @return The X location of the MGC 1/4 chord of the specified panel.
	*/
	public final double getMGCQuarterChordX(int idx) {
	    if (idx >= numPanels && mirror)
	        idx -= numPanels;
		return Xmgc4[idx];
	}
	
	/**
	*  Returns the Y location of the 1/4 chord of the MGC of the specified panel.
	*
	*  @param idx  Index of the panel to return the average span for.
	*  @return The Y location of the MGC 1/4 chord of the specified panel.
	*/
	public final double getMGCQuarterChordY(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
	        return -Ymgc4[idx];
	    }
		return Ymgc4[idx];
	}
	
	/**
	*  Returns the Z location of the 1/4 chord of the MGC of the specified panel.
	*
	*  @param idx  Index of the panel to return the average span for.
	*  @return The Z location of the MGC 1/4 chord of the specified panel.
	*/
	public final double getMGCQuarterChordZ(int idx) {
	    if (idx >= numPanels && mirror)
	        idx -= numPanels;
		return Zmgc4[idx];
	}
	
	/**
	*  Returns the X component of the normal vector of the specified panel.
	*
	*  @param idx  Index of the panel to return the normal vector for.
	*  @return The X component of the normal vector of the specified panel.
	*/
	public final double getNormalX(int idx) {
	    if (idx >= numPanels && mirror)
	        idx -= numPanels;
		return -nets[idx].getNormalX(string[idx], point[idx]);
	}
	
	/**
	*  Returns the Y component of the normal vector of the specified panel.
	*
	*  @param idx  Index of the panel to return the normal vector for.
	*  @return The Y component of the normal vector of the specified panel.
	*/
	public final double getNormalY(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
	        return nets[idx].getNormalY(string[idx], point[idx]);
	    }
		return -nets[idx].getNormalY(string[idx], point[idx]);
	}
	
	/**
	*  Returns the Z component of the normal vector of the specified panel.
	*
	*  @param idx  Index of the panel to return the normal vector for.
	*  @return The Z component of the normal vector of the specified panel.
	*/
	public final double getNormalZ(int idx) {
	    if (idx >= numPanels && mirror)
	        idx -= numPanels;
		return -nets[idx].getNormalZ(string[idx], point[idx]);
	}
	
	/**
	*  Returns the X coordinate of the inboard end of the swept bound vortex
	*  for the specified panel in the configuration in model coordinates.  This is
	*  X location of the 1/4 chord of the inboard (root) panel chord.
	*/
    public final double getXbi(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
	        return Xbo[idx];
	    }
        return Xbi[idx];
    }

	/**
	*  Returns the X coordinate of the outboard end of the swept bound vortex
	*  for the specified panel in the configuration in model coordinates.  This is the
	*  X location of the 1/4 chord of the outboard (tip) panel chord.
	*/
    public final double getXbo(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
	        return Xbi[idx];
        }
        return Xbo[idx];
    }

	/**
	*  Returns the Y coordinate of the inboard end of the swept bound vortex
	*  for the specified panel in the configuration in model coordinates.  This is the
	*  Y location of the inboard (root) edge of each panel.
	*/
    public final double getYbi(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
	        return -Ybo[idx];
	    }
        return Ybi[idx];
    }

	/**
	*  Returns the Y coordinate of the outboard end of the swept bound vortex
	*  for the specified panel in the configuration in model coordinates.  This is the
	*  Y location of the outboard (tip) edge of each panel.
	*/
    public final double getYbo(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
	        return -Ybi[idx];
	    }
        return Ybo[idx];
    }

	/**
	*  Returns the Z coordinate of the inboard end of the swept bound vortex
	*  for the specified panel in the configuration in model coordinates.  This is the
	*  Z location of the inboard (root) panel chord leading edge.
	*/
    public final double getZbi(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
	        return Zbo[idx];
	    }
        return Zbi[idx];
    }

	/**
	*  Returns the Z coordinate of the outboard end of the swept bound vortex
	*  for the specified panel in the configuration in model coordinates.  This is the
	*  Z location of the outboard (tip) panel chord leading edge.
	*/
    public final double getZbo(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
	        return Zbi[idx];
	    }
        return Zbo[idx];
    }

	/**
	*  Returns the X coordinate of the control point for the specified panel
	*  in the model in model coordinates.
	*/
    public final double getXp(int idx) {
	    if (idx >= numPanels && mirror)
	        idx -= numPanels;
        return px[idx];
    }

	/**
	*  Returns the Y coordinate of the control point for the specified panel
	*  in the model in model coordinates.
	*/
    public final double getYp(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
	        return -py[idx];
	    }
        return py[idx];
    }

	/**
	*  Returns the Z coordinate of the control point for the specified panel
	*  in the model in model coordinates.
	*/
    public final double getZp(int idx) {
	    if (idx >= numPanels && mirror)
	        idx -= numPanels;
        return pz[idx];
    }

	/**
	*  Returns the X coordinate of the inboard leading edge of the specified
	*  panel.
	*/
    public final double getXLEi(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
            double[][] pntsX = nets[idx].getAllX();
            return pntsX[string[idx]+1][point[idx]];
	    }
        double[][] pntsX = nets[idx].getAllX();
        return pntsX[string[idx]][point[idx]];
    }

	/**
	*  Returns the Y coordinate of the inboard leading edge of the specified
	*  panel.
	*/
    public final double getYLEi(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
            double[][] pntsY = nets[idx].getAllY();
            return -pntsY[string[idx]+1][point[idx]];
	    }
        double[][] pntsY = nets[idx].getAllY();
        return pntsY[string[idx]][point[idx]];
    }

	/**
	*  Returns the Z coordinate of the inboard leading edge of the specified
	*  panel.
	*/
    public final double getZLEi(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
            double[][] pntsZ = nets[idx].getAllZ();
            return pntsZ[string[idx]+1][point[idx]];
	    }
        double[][] pntsZ = nets[idx].getAllZ();
        return pntsZ[string[idx]][point[idx]];
    }

	/**
	*  Returns the X coordinate of the inboard trailing edge of the specified
	*  panel.
	*/
    public final double getXTEi(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
            double[][] pntsX = nets[idx].getAllX();
            return pntsX[string[idx]+1][point[idx]+1];
	    }
        double[][] pntsX = nets[idx].getAllX();
        return pntsX[string[idx]][point[idx]+1];
    }

	/**
	*  Returns the Y coordinate of the inboard trailing edge of the specified
	*  panel.
	*/
    public final double getYTEi(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
            double[][] pntsY = nets[idx].getAllY();
            return -pntsY[string[idx]+1][point[idx]+1];
	    }
        double[][] pntsY = nets[idx].getAllY();
        return pntsY[string[idx]][point[idx]+1];
    }

	/**
	*  Returns the Z coordinate of the inboard trailing edge of the specified
	*  panel.
	*/
    public final double getZTEi(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
            double[][] pntsZ = nets[idx].getAllZ();
            return pntsZ[string[idx]+1][point[idx]+1];
	    }
        double[][] pntsZ = nets[idx].getAllZ();
        return pntsZ[string[idx]][point[idx]+1];
    }

	/**
	*  Returns the X coordinate of the outboard leading edge of the specified
	*  panel.
	*/
    public final double getXLEo(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
            double[][] pntsX = nets[idx].getAllX();
            return pntsX[string[idx]][point[idx]];
	    }
        double[][] pntsX = nets[idx].getAllX();
        return pntsX[string[idx]+1][point[idx]];
    }

	/**
	*  Returns the Y coordinate of the outboard leading edge of the specified
	*  panel.
	*/
    public final double getYLEo(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
            double[][] pntsY = nets[idx].getAllY();
            return -pntsY[string[idx]][point[idx]];
	    }
        double[][] pntsY = nets[idx].getAllY();
        return pntsY[string[idx]+1][point[idx]];
    }

	/**
	*  Returns the Z coordinate of the outboard leading edge of the specified
	*  panel.
	*/
    public final double getZLEo(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
            double[][] pntsZ = nets[idx].getAllZ();
            return pntsZ[string[idx]][point[idx]];
	    }
        double[][] pntsZ = nets[idx].getAllZ();
        return pntsZ[string[idx]+1][point[idx]];
    }

	/**
	*  Returns the X coordinate of the outboard trailing edge of the specified
	*  panel.
	*/
    public final double getXTEo(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
            double[][] pntsX = nets[idx].getAllX();
            return pntsX[string[idx]][point[idx]+1];
	    }
        double[][] pntsX = nets[idx].getAllX();
        return pntsX[string[idx]+1][point[idx]+1];
    }

	/**
	*  Returns the Y coordinate of the outboard trailing edge of the specified
	*  panel.
	*/
    public final double getYTEo(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
            double[][] pntsY = nets[idx].getAllY();
            return -pntsY[string[idx]][point[idx]+1];
	    }
        double[][] pntsY = nets[idx].getAllY();
        return pntsY[string[idx]+1][point[idx]+1];
    }

	/**
	*  Returns the Z coordinate of the outboard trailing edge of the specified
	*  panel.
	*/
    public final double getZTEo(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
            double[][] pntsZ = nets[idx].getAllZ();
            return pntsZ[string[idx]][point[idx]+1];
	    }
        double[][] pntsZ = nets[idx].getAllZ();
        return pntsZ[string[idx]+1][point[idx]+1];
    }

	/**
	*  Returns the index of the next panel aft of the specified panel.
	*  If a value of -1 is returned, then there is no panel aft of the specified
	*  panel (it is on the trailing edge of a network).  Make sure you check
	*  for that case!
	**/
    public final int getNextAft(int idx) {
	    if (idx >= numPanels && mirror) {
	        idx -= numPanels;
			idx = nextAft[idx];
			if (idx > 0)
				idx += numPanels;
			return idx-1;
		}
        return nextAft[idx]-1;
    }

}


