/*
 *  XMLInterface  -- A class that enables reading and writing XML data for AeroLS.
 *
 *  Copyright (C) 2002-2024, Joseph A. Huwaldt
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/gpl.html
 */

import jahuwaldt.aero.panelgeom.GeomComponent;
import jahuwaldt.aero.panelgeom.GeomNetwork;
import jahuwaldt.aero.panelgeom.GeomVehicle;
import jahuwaldt.js.ParamTools.AutoParam;
import jahuwaldt.js.ParamTools.Param;
import jahuwaldt.js.ParamTools.ParamDatabase;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.text.ParsePosition;
import java.util.*;
import javax.measure.quantity.Dimensionless;
import javax.measure.unit.Unit;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;


/**
 * This class serves as an interface between the main AeroLS program and the
 * code for reading and writing data out to an XML file. This class uses the
 * JDOM package to actually do the reading and writing of XML files.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt,  Date: May 30, 2002
 * @version January 21, 2024
 */
public class XMLInterface {

    // Create a map of conversions from the old unit's package to the new one.
    private static final HashMap<String,String> UNIT_CONV = new HashMap();
    static {
        //  Define the unit conversions.
        UNIT_CONV.put("Feet", "ft");
        UNIT_CONV.put("Inches", "in");
        UNIT_CONV.put("Meters", "m");
        UNIT_CONV.put("Centimeters", "cm");
        
        UNIT_CONV.put("SqFeet", "ft^2");
        UNIT_CONV.put("SqInches", "in^2");
        UNIT_CONV.put("SqMeters", "m^2");
        UNIT_CONV.put("SqCentimeters", "cm^2");
        
        UNIT_CONV.put("Degrees", "deg");
        UNIT_CONV.put("Radians", "rad");
    }
    
	/**
	*  Method that converts this project's data to XML and writes
	*  that data out to the specified output stream (probably to a file).
	*/
	public static void outputToXML(OutputStream outStream, ProjectData data) throws IOException {

        //  Extract the data we are saving out from the project data object.
		String modelName = data.modelName;
		GeomVehicle geometry = data.geometry;
		
		//	Create XML root document element.
		Element rootElement = new Element("document");
		Document document = new Document(rootElement);
		
		//	Add the name of the project to the XML document.
		Element projNameElement = new Element("modelName");
		projNameElement.setText(modelName);
		rootElement.addContent(projNameElement);
		
		//	Add all the values and units of each parameter in the database to the XML document.
		Element paramsElement = new Element("parameters");
		addParamDatabase(paramsElement, data.pdb);
		rootElement.addContent(paramsElement);

		//  Add all the geometry to the XML document.
		if (!geometry.isEmpty()) {
	    	Element geomElement = new Element("geometry");
		    Element vehElement = new Element("vehicle");
		    vehElement.setAttribute("name", geometry.getName());
		    addComponents(vehElement, geometry);
		    geomElement.addContent(vehElement);
		    rootElement.addContent(geomElement);
		}

		//  Add the test conditions to the XML document.
		if (!data.testConditions.isEmpty()) {
            Element testCondElement = new Element("testConditions");
            addTestConditions(testCondElement, data.testConditions, data.beta);
            rootElement.addContent(testCondElement);
		}
		
		//	Write out the XML document.
		XMLOutputter outputter = new XMLOutputter("  ", true);
		outputter.output(document, outStream);
		
	}
	
	/**
	*  Method that parses this project's data from the specified XML input stream.
	*  If a parameter object is encountered in the input file that doesn't exist
	*  in the supplied parameter database, it is simply skipped over.
	*
	*  @param  inStream  The input stream to an XML file that we are reading from.
	*  @param  pdb       The parameter database containing the parameters for this
	*                    program.
	*  @return A new ProjectData object with the new data from the file filled in.
	*/
	public static ProjectData read(InputStream inStream, ParamDatabase pdb) throws IOException, Exception  {

	    //  Create a new project data container.
	    ProjectData data = new ProjectData(pdb);
	    
		//	Begin reading XML configuration file containing program info.
		SAXBuilder builder = new SAXBuilder();
		Document doc = builder.build(inStream);
		
		//	Extract the root of the XML document.
		Element root = doc.getRootElement();
		
		//	Extract the model name.
		Element projNameElement = root.getChild("modelName");
		if (projNameElement != null)
			data.modelName = projNameElement.getText();
		
        //  Read in parameter database.
	    inputParameters(root, data.pdb);

	    //  Read in the geometry.
	    inputGeometry(root, data.geometry);

	    //  Read in the test conditions.
	    inputTestConditions(root, data);
	    
		return data;
	}

	
	/**
	*  Method that loops through all the parameters in the parameter database and adds the 
	*  required data to the specified XML element.
	*/
	private static void addParamDatabase(Element parameters, ParamDatabase pdb) {
	
		for (Param param : pdb) {
            if (param == null) continue;
            
			//	Only save off parameters that have been manually input.
			boolean isManual = true;
			if (param instanceof AutoParam)
				isManual = ((AutoParam)param).isManuallyInput();
				
			if (isManual) {
			
				//	Get the value of this parameter in reference units.
				double value = param.getValue();
				
				//	Only output the value if it is an actual number.
				if (!Double.isNaN(value)) {
				
					//	Get the current units used by this parameter.
					Unit unit = param.getUnit();
			
					//	Get the name of the unit.
					String unitName = unit.toString();
			
					//	Create an XML element for this parameter.
					Element paramElement = new Element("param");
					paramElement.setAttribute("name", param.getLabel());
			
					//	Add the parameter value information.
					Element valueElement = new Element("value");
					valueElement.setText(Double.toString(value));
					paramElement.addContent(valueElement);
			
					//	Add the unit information.
					if (unitName != null && !unit.equals(Dimensionless.UNIT)) {
						Element unitElement = new Element("unit");
						unitElement.setText(unitName);
						paramElement.addContent(unitElement);
					}
				
					//	Finally, add this parameter's record to the root element.
					parameters.addContent(paramElement);
				}
			}
		}
		
	}
	
	
	/**
	*  Method that loops through all the components in the specified vehicle and adds that
	*  data to the specified XML element.
	*/
	private static void addComponents(Element parent, GeomVehicle vehicle) {
	
		for (Iterator i = vehicle.components(); i.hasNext();) {
		
			//	Extract component from vehicle.
			GeomComponent comp = (GeomComponent)i.next();
			
			//	Get the name of the component.
			String name = comp.getName();
			
			//	Create an XML element for this component.
			Element element = new Element("component");
			element.setAttribute("name", name);

			//  Add any networks contained in this component.
			addNetworks(element, comp);
			
			//	Finally, add this component's record to the parent element.
			parent.addContent(element);
		}
		
	}
	
	
	/**
	*  Method that loops through all the networks in the specified component and adds that
	*  data to the specified XML element.
	*/
	private static void addNetworks(Element parent, GeomComponent component) {
	
		for (Iterator i = component.networks(); i.hasNext();) {
		
			//	Extract network from component.
			GeomNetwork net = (GeomNetwork)i.next();
			int numStrings = net.getNumStrings();
			int pntsPerString = net.getNumPointsPerString();
			
			//	Get the name of the network.
			String name = net.getName();
			
			//	Create an XML element for this component.
			Element element = new Element("network");
			element.setAttribute("name", name);
			element.setAttribute("strings", Integer.toString(numStrings));
			element.setAttribute("points", Integer.toString(pntsPerString));

			//  Add arrays of data.
			addArray(element, net.getAllX(), "X");
			addArray(element, net.getAllY(), "Y");
			addArray(element, net.getAllZ(), "Z");
			
			//	Finally, add this network's record to the parent element.
			parent.addContent(element);
		}
		
	}

	
	/**
	*  Method that adds the specified 2D array to the specified XML element.
	*/
	private static void addArray(Element parent, double[][] array, String name) {

        if (array != null) {
		    //	Create an XML element for this component.
		    Element element = new Element("array");
		    element.setAttribute("name", name);

		    StringBuilder buffer = new StringBuilder();
		    int cols = array.length;
		    int rows = array[0].length;

		    for (int i=0; i < cols; ++i) {
                for (int j=0; j < rows; ++j) {
                    buffer.append(array[i][j]);
                    buffer.append(",");
                }
		    }

		    element.setText(buffer.toString());

			//	Finally, add this array's data to the parent element.
			parent.addContent(element);
        }
        
	}


	/**
	*  Method that outputs test condition breakpoints to the specified XML data element.
	*/
	private static void addTestConditions(Element parent, Hashtable tc, double beta) {

        Enumeration e = tc.keys();
        while (e.hasMoreElements()) {
            String key = (String)e.nextElement();
            double[] bp = (double[])tc.get(key);
	        int size = bp.length;

            Element element = new Element("table");
            element.setAttribute("name", key);
            element.setAttribute("length", String.valueOf(size));
            
	        StringBuilder buffer = new StringBuilder();
	        for (int i=0; i < size; ++i) {
	            buffer.append(bp[i]);
	            buffer.append(",");
	        }
	        element.setText(buffer.toString());

	        parent.addContent(element);
        }
		
		//	Add sideslip parameter.
		Element element = new Element("table");
		element.setAttribute("name", "Beta");
		element.setAttribute("length", "1");
		element.setText(String.valueOf(beta));
		parent.addContent(element);
        
	}

	
	/**
	*  Method that parses out parameter objects from the specified XML data element.
	*/
	private static void inputParameters(Element root, ParamDatabase pdb) throws Exception {

		//	Extract the list of parameters.
	    List paramList = null;
	    Element pElement = root.getChild("parameters");
	    if (pElement != null)
		    paramList = pElement.getChildren("param");
		
		if (paramList != null) {
			
			//	Loop over all the parameters parsing each in turn.
			int numParams = paramList.size();
			for (int i=0; i < numParams; ++i) {
				
				//	Get a reference to this parameter XML element.
				pElement = (Element)paramList.get(i);
			
				//	Extract the name of the parameter.
				String name = pElement.getAttribute("name").getValue();
				
				//	Extract a parameter with this name from the parameter database.
				Param param = pdb.get(name);

				//  If the parameter doesn't exist in the database, just skip it.
				if (param != null) {
				
    				//	Extract the new value for this parameter.
    				double value = Double.NaN;
    				Element element = pElement.getChild("value");
    				if (element != null) {
    					String valueText = element.getText();
    					if (!valueText.equals("NaN"))
    						value = Double.parseDouble(valueText);
    				}
				
    				//	Extract the new units for this parameter.
    				Unit unit = Dimensionless.UNIT;
    				element = pElement.getChild("unit");
    				if (element != null) {
    					String unitName = element.getText();
                        
                        //  Convert from the old unit package to the new one if necessary.
                        if (UNIT_CONV.containsKey(unitName))
                            unitName = UNIT_CONV.get(unitName);
                        
    					unit = Unit.valueOf(unitName);
    				}
				
    				//	Finally, change this parameter's value and unit to match those read in
    				//	from the XML file.
    				if (param instanceof AutoParam)
    				    ((AutoParam)param).makeManuallyInput();
    				param.setUnit(unit);
    				param.setValue(value);
    				if (param instanceof AutoParam)
    				    ((AutoParam)param).updateChildren();
				}
			}
			
		}

	}


	/**
	*  Method that parses out geometry from the specified XML data element.
	*/
	private static void inputGeometry(Element root, GeomVehicle vehicle) throws Exception {

		//	Extract the list of vehicle geometry objects.
	    List compList = null;
	    Element element = root.getChild("geometry");
	    if (element != null)
	        element = element.getChild("vehicle");
	    if (element != null) {
	        vehicle.setName(element.getAttribute("name").getValue());
		    compList = element.getChildren("component");
	    }
	    
		inputComponents(compList, vehicle);

	}


	/**
	*  Method that parses out component geometry from the specified XML data element.
	*/
	private static void inputComponents(List list, GeomVehicle vehicle) throws Exception {

		if (list != null) {

			//	Loop over all the components parsing each in turn.
			int numElements = list.size();
			for (int i=0; i < numElements; ++i) {

			    //	Get a reference to this component XML element.
			    Element element = (Element)list.get(i);

			    //  Get the name of the component.
			    String name = element.getAttribute("name").getValue();

			    //  Create a component object.
			    GeomComponent comp = new GeomComponent(name);

                //  Get a list of XML records for all the networks in this component.
			    List netList = element.getChildren("network");

                //  Read in all the networks that are a part of this component.
			    inputNetworks(netList, comp);

                //  Finally, add this component to the vehicle geometry list.
			    vehicle.add(comp);
			}
			
		}
		
	}
	

	/**
	*  Method that parses out network geometry from the specified XML data element.
	*/
	private static void inputNetworks(List list, GeomComponent component) throws Exception {

		if (list != null) {

			//	Loop over all the networks parsing each in turn.
			int numElements = list.size();
			for (int i=0; i < numElements; ++i) {

			    //	Get a reference to this networks XML element.
			    Element element = (Element)list.get(i);

			    //  Get the name of the networks.
			    String name = element.getAttribute("name").getValue();

                //  Get the number of strings in the arrays.
			    int numStrings =Integer.parseInt(element.getAttribute("strings").getValue());
			    
                //  Get the number of points per string in the arrays.
			    int pntsPerString =Integer.parseInt(element.getAttribute("points").getValue());

                //  Get a list of XML records for all the arrays in this network.
			    List arrList = element.getChildren("array");

                //  Read in arrays of data.
                double[][] xArray = inputArray(arrList, "X", numStrings, pntsPerString);
                double[][] yArray = inputArray(arrList, "Y", numStrings, pntsPerString);
                double[][] zArray = inputArray(arrList, "Z", numStrings, pntsPerString);
			    
			    //  Create a networks object.
                if (xArray != null && yArray != null && zArray != null) {
			        GeomNetwork net = new GeomNetwork(xArray, yArray, zArray, name);

                    //  Finally, add this network to the component geometry list.
			        component.add(net);
                }
			}
		}

	}
	
	/**
	*  Method that parses out network geometry from the specified XML data element.
	*/
	private static double[][] inputArray(List list, String id, int cols, int rows) throws Exception {

        double array[][] = null;
        
		if (list != null) {
             
			//	Loop over all the arrays and look for the one we want.
			int numElements = list.size();
			for (int i=0; i < numElements; ++i) {

			    //	Get a reference to this array XML element.
			    Element element = (Element)list.get(i);

			    //  Get the name of the array.
			    String name = element.getAttribute("name").getValue();
                if (name.equals(id)) {

                    //  Found the array we want.  Extract the data.
                    String data = element.getText();

                    //  Parse the array of numbers.
                    array = parseArrayData(data, cols, rows);

                    break;
                }
			}
		}

		return array;
	}


	/**
	*  Method that parses a single array of data out of a string of numbers
	*  separated by commas.
	*/
	private static double[][] parseArrayData(String data, int cols, int rows)  {

        double[][] array = null;

        if (data != null) {
            NumberFormat nf = (DecimalFormat)NumberFormat.getInstance();
            ParsePosition position = new ParsePosition(0);

            array = new double[cols][rows];
            for (int i=0; i < cols; ++i) {
                for (int j=0; j < rows; ++j) {
                    //  Parse out the number.
                    array[i][j] = nf.parse(data, position).doubleValue();

                    //  Increment the position to skip the commas.
                    position.setIndex(position.getIndex()+1);
                }
            }
        }
        
        return array;
	}
	
	/**
	*  Method that parses out test conditions from the specified XML data element.
	*/
	private static void inputTestConditions(Element root, ProjectData projData) throws Exception {

		Hashtable tc = projData.testConditions;
		
		//	Extract the list of test condition objects.
	    Element element = root.getChild("testConditions");
	    if (element != null) {
	        List tableList = element.getChildren("table");

	        if (tableList != null) {
	            int size = tableList.size();
	            for (int i=0; i < size; ++i) {
	            
	                element = (Element)tableList.get(i);
	                String key = element.getAttribute("name").getValue();
	                int length = Integer.parseInt(element.getAttribute("length").getValue());
                    String data = element.getText();
                    double[] bp = parseTable(data, length);

					if (key.equals("Beta"))
						projData.beta = bp[0];
					
					else
						//  Add this table of breakpoints to the test conditions table.
						tc.put(key, bp);
	            }
	        }
	    }
	    
	}
	
	/**
	*  Method that parses a single table of data out of a string of numbers
	*  separated by commas.
	*/
	private static double[] parseTable(String data, int length)  {

        double[] array = null;

        if (data != null) {
            NumberFormat nf = (DecimalFormat)NumberFormat.getInstance();
            ParsePosition position = new ParsePosition(0);

            array = new double[length];
            for (int i=0; i < length; ++i) {
                //  Parse out the number.
                array[i] = nf.parse(data, position).doubleValue();

                //  Increment the position to skip the commas.
                position.setIndex(position.getIndex()+1);
            }
        }
        
        return array;
	}

}
