/*
 *  bfn_yrle_ytle - Method for calculating wing span as a function of yrle & ytle.
 *
 *  Copyright (C) 1998 - 2015 by Joseph A. Huwaldt 
 *  All rights reserved.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */

import jahuwaldt.js.ParamTools.AbstractParamMethod;
import jahuwaldt.js.param.Parameter;
import javax.measure.quantity.Length;

/**
 * Calculates wing span as a function of the Y coordinate of the leading edge of
 * the root and tip chords.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: May 23, 2002
 * @version June 18, 2015
 */
public class bfn_yrle_ytle extends AbstractParamMethod {

    /**
     * Constructor for this method.
     *
     * @param yrle Label for the "Y coordinate of root chord leading edge" parameter.
     * @param ytle Label for the "Y coordinate of tip chord leading edge" parameter.
     */
    public bfn_yrle_ytle(String yrle, String ytle) {
        _inputLabels = new String[2];
        _inputLabels[0] = yrle;
        _inputLabels[1] = ytle;
    }

    /**
     * Calculates wing span as a function of the Y coordinate of the leading
     * edge of the root and tip chords. Method: b = ytle - yrle
     *
     * @param inputs Array of input parameters. [0=Y location of root chord LE (ytle),
     *                  1=Y location of tip chord LE (ytle)].
     * @return Y coordinate of the leading edge of the root chord in length units.
     */
    @Override
    public Parameter<Length> calc(Parameter[] inputs) {
        Parameter<Length> yrle = inputs[0];
        Parameter<Length> ytle = inputs[1];
        Parameter<Length> b = ytle.minus(yrle);     // b = ytle - yrle;
        return b;
    }
}
