/*
 *  itfn_ir_twist - Method for calculating tip chord incidence as a function of ir & twist.
 *
 *  Copyright (C) 2002 - 2015 by Joseph A. Huwaldt
 *  All rights reserved.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */

import jahuwaldt.js.ParamTools.AbstractParamMethod;
import jahuwaldt.js.param.Parameter;
import javax.measure.quantity.Angle;

/**
 * Calculates tip chord incidence as a function of the incidence angle of the
 * root chord and the twist angle.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: May 29, 2002
 * @version June 19, 2015
 */
public class itfn_ir_twist extends AbstractParamMethod<Angle> {

    /**
     * Constructor for this method.
     *
     * @param ir Label for the "incidence of the root chord" parameter.
     * @param twist Label for the "root to tip twist angle" parameter.
     */
    public itfn_ir_twist(String ir, String twist) {
        _inputLabels = new String[2];
        _inputLabels[0] = ir;
        _inputLabels[1] = twist;
    }

    /**
     * Calculates tip chord incidence as a function of the incidence angle of
     * the root chord and the twist angle. Method: it = ir - twist
     *
     * @param inputs Array of input parameters. [0=incidence of root chord (ir),
     * 1=tip to root twist angle (twist)].
     * @return Tip chord plane incidence angle.
     */
    @Override
    public Parameter<Angle> calc(Parameter[] inputs) {
        Parameter<Angle> ir = inputs[0];
        Parameter<Angle> twist = inputs[1];
        Parameter<Angle> it = ir.minus(twist);
        return it;
    }

}
