/*
 *   Param - Base class for all parameter value objects.
 *
 *   Copyright (C) 1997-2016, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.ParamTools;

import jahuwaldt.js.param.Parameter;
import java.text.MessageFormat;
import java.text.NumberFormat;
import java.text.ParseException;
import static java.util.Objects.requireNonNull;
import java.util.ResourceBundle;
import javax.measure.converter.ConversionException;
import javax.measure.quantity.*;
import javax.measure.unit.Unit;

/**
 * A generic parameter class for use in my engineering programs. The Param object provides
 * a package that contains an extended amount of information on a single parameter or
 * value such as the units and an identifying label.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: November 23, 1997
 * @version March 28, 2016
 * @param <Q> The Quantity (unit type) for this parameter.
 */
public class Param<Q extends Quantity> implements java.io.Serializable {

    private static final long serialVersionUID = 8612771552684695545L;

    /**
     * The resource bundle for this package.
     */
    static final ResourceBundle RESOURCES
            = ResourceBundle.getBundle("jahuwaldt.js.ParamTools.ParamToolsResources", java.util.Locale.getDefault());

    /**
     * The value of this parameter.
     */
    private Parameter<Q> _theValue;

    /**
     * Is this parameter defined?
     */
    private boolean _defined;

    /**
     * String containing a label for this parameter.
     */
    private String _labelStr;

    /**
     * Array of available units for this parameter.
     */
    protected Unit<Q>[] _availUnits;

    /**
     * The number format used when displaying this parameter.
     */
    private NumberFormat _numFormat = null;

    //-----------------------------------------------------------------------------
    /**
     * Constructs a parameter object using a label string and a <code>Parameter</code>
     * value for the parameter.
     *
     * @param label The label to be attached to this object. May not be null.
     * @param value The value for the parameter to hold. May not be null.
     */
    public Param(String label, Parameter<Q> value) {
        _labelStr = requireNonNull(label);
        _theValue = requireNonNull(value);
        _defined = true;
    }

    /**
     * Constructs a parameter object using a label string and the units for the parameter
     * while setting the parameter's initial value to <code>Double.NaN</code> (undefined).
     *
     * @param label    The label to be attached to this object. May not be null.
     * @param theUnits The units associated with this parameter. May not be null.
     */
    public Param(String label, Unit<Q> theUnits) {
        _labelStr = requireNonNull(label);
        _theValue = Parameter.valueOf(Double.NaN, requireNonNull(theUnits));
        _defined = false;
    }

    /**
     * Constructs a parameter object using a label string a <code>double</code> value in
     * the specified units.
     *
     * @param label    The label to be attached to this object. May not be null.
     * @param value    The value for the parameter to hold. May not be null.
     * @param theUnits The units associated with this parameter.
     */
    public Param(String label, double value, Unit<Q> theUnits) {
        _labelStr = requireNonNull(label);
        _theValue = Parameter.valueOf(value, requireNonNull(theUnits));
        _defined = true;
    }

    //-----------------------------------------------------------------------------
    /**
     * Method to return the current value of this parameter.
     *
     * @return Return the current value of this parameter.
     */
    public Parameter<Q> get() {
        return _theValue;
    }

    /**
     * Method to return the current value of this parameter in the current units.
     *
     * @return The value of this parameter in the current units
     */
    public double getValue() {
        return _theValue.getValue();
    }

    /**
     * Method to set the value of this parameter. Makes the parameter "defined".
     *
     * @param value The value that the parameter object should be set to.
     */
    public void set(Parameter<Q> value) {

        if (!value.isNaN()) {
            _theValue = value.to(_theValue.getUnit());
            _defined = true;
        } else
            makeUndefined();
    }

    /**
     * Method to set the value of this parameter in the current units. Makes the parameter
     * "defined".
     *
     * @param value The value that the parameter object should be set to in the current
     *              units of this parameter object.
     */
    public void setValue(double value) {

        if (!Double.isNaN(value)) {
            if (value != _theValue.getValue())
                _theValue = Parameter.valueOf(value, _theValue.getUnit());
            _defined = true;
        } else
            makeUndefined();
    }

    /**
     * Method to set the value of this parameter in the current units using string input.
     * Makes this parameter "defined".
     *
     * @param value The value that the parameter object should be set to. If null or an
     *              empty string, then the parameter will be made undefined.
     * @throws NumberFormatException if value string can't be converted to a double.
     * @throws ParseException if value string can't be converted to a double.
     */
    public void setValue(String value) throws NumberFormatException, ParseException {
        if (value == null || value.equals(""))
            makeUndefined();

        else {
            double dValue;
            if (_numFormat == null)
                dValue = Double.valueOf(value);
            else
                dValue = _numFormat.parse(value).doubleValue();

            setValue(dValue);
        }
    }

    /**
     * Method to return the string representation of the label of this parameter.
     *
     * @return Returns the string representation of the parameter label.
     */
    public String getLabel() {
        return _labelStr;
    }

    /**
     * Method to change the label that is associated with this parameter.
     *
     * @param label The label to be attached to this object.
     */
    public void setLabel(String label) {
        _labelStr = label;
    }

    /**
     * Method to return the current units of this parameter.
     *
     * @return Returns the units of this parameter.
     */
    public Unit<Q> getUnit() {
        return _theValue.getUnit();
    }

    /**
     * Method to change the units of this parameter <strong>without converting</strong>
     * the value to the new units (value is left unchanged).
     *
     * @param newUnits The new units to be assigned to this parameter. May not be null.
     * @exception ConversionException Unable to convert between the specified units.
     */
    public void setUnit(Unit<Q> newUnits) throws ConversionException {
        //  Make sure the new units are compatible with the old units.
        if (!_theValue.getUnit().isCompatible(requireNonNull(newUnits)))
            throw new ConversionException(
                    MessageFormat.format(RESOURCES.getString("incompatibleUnits"),
                            "new", "the old units"));

        double value = _theValue.getValue();
        _theValue = Parameter.valueOf(value, newUnits);
    }

    /**
     * Method to convert the value of this parameter from the current units to the input
     * units (value is changed).
     *
     * @param newUnits The new units to convert this parameter to. May not be null.
     * @exception ConversionException Unable to convert between the specified units.
     */
    public void changeUnits(Unit<Q> newUnits) throws ConversionException {
        //  Convert the value and change the units.
        _theValue = _theValue.to(requireNonNull(newUnits));
    }

    /**
     * Sets the list of unit types that are available for use by this parameter. If this
     * parameter does not already have units assigned, then the 1st unit of the unit list
     * will become the current unit.
     *
     * @param newUnitArr The array of Units to be made into the available unit list.
     * @throws ConversionException if unable to convert between the specified units.
     * @throws IllegalArgumentException if input unit array contains a null element.
     */
    public void setAvailUnits(Unit<Q>[] newUnitArr) throws ConversionException, IllegalArgumentException {

        // First check to see if the user is zeroing out the list of units.
        if (newUnitArr == null) {
            _availUnits = null;
            return;
        }

        // Make sure none of the units are blank.
        for (int i = 0; i < newUnitArr.length; ++i) {
            if (newUnitArr[i] == null) {
                String idStr = "Param(" + getLabel() + ").setAvailUnits newUnitArr[" + i + "]";
                throw new IllegalArgumentException(
                        MessageFormat.format(RESOURCES.getString("paramNullErr"), idStr));
            }
        }

        //  Test to see if all the input units are compatible with each other.
        Unit<Q> testUnit = _theValue.getUnit();
        for (int i = 1; i < newUnitArr.length; ++i) {
            if (!testUnit.isCompatible(newUnitArr[i])) {
                String idStr = "Param(" + getLabel() + ").setAvailUnits newUnitArr[" + i + "]";
                throw new ConversionException(
                        MessageFormat.format(RESOURCES.getString("incompatibleUnits"),
                                idStr, "other newUnitArr values"));
            }
        }

        //  If we have gotten this far without an exception, then the types are compatible.
        _availUnits = newUnitArr;
    }

    /**
     * Gets the list of unit types that are available for use by this parameter. If this
     * parameter does not have a list of available units, <code>null</code> is returned.
     *
     * @return An array of unit types available with this parameter or <code>null</code>
     *         if there is no available unit list.
     */
    public Unit<Q>[] getAvailUnits() {
        return _availUnits;
    }

    /**
     * Set the number format to use when displaying this parameter. If no number format is
     * set, then a default will be used.
     *
     * @param nf The number format to use.
     */
    public void setNumberFormat(NumberFormat nf) {
        _numFormat = nf;
    }

    /**
     * Get the number format used when displaying this parameter. If no number format has
     * been set, null is returned.
     *
     * @return The number format used.
     */
    public NumberFormat getNumberFormat() {
        return _numFormat;
    }

    /**
     * Returns true if parameter is defined, false otherwise.
     *
     * @return <code>true</code> if this parameter is defined, <code>false</code> if it is
     *         undefined.
     */
    public boolean isDefined() {
        return _defined;
    }

    /**
     * Method to make this parameter undefined. The value is set to
     * <code>Double.NaN</code>.
     */
    public void makeUndefined() {
        _defined = false;
        _theValue = Parameter.valueOf(Double.NaN, _theValue.getUnit());
    }

    /**
     * Method to return the value of this parameter in current units as a string.
     *
     * @return Returns a string representation of the value of this parameter in the
     *         current units.
     */
    @Override
    public String toString() {
        String s;
        if (_labelStr == null)
            s = _theValue.toString();

        else {
            StringBuilder buf = new StringBuilder(_labelStr);
            buf.append(" = ");
            buf.append(_theValue.toString());
            s = buf.toString();
        }

        return s;
    }

}
