/*
 *   ParamInput - Panel used for entering Param values using a text field.
 *
 *   Copyright (C) 1997-2016, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.ParamTools;

import jahuwaldt.js.param.Parameter;
import jahuwaldt.swing.DecimalField;
import java.awt.BorderLayout;
import java.awt.Insets;
import java.awt.event.*;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.text.ParseException;
import static java.util.Objects.requireNonNull;
import java.util.ResourceBundle;
import javax.measure.converter.ConversionException;
import javax.measure.quantity.Quantity;
import javax.measure.unit.Unit;
import javax.swing.*;

/**
 * Defines a panel that is used for entering {@link Param} type parameter values (using a
 * text field). ParamInput provides a user interface for interacting with parameter
 * objects.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: November 23, 1997
 * @version March 28, 2016
 * @param <Q> The Quantity (unit type) for the parameter this input panel displays.
 */
public class ParamInput<Q extends Quantity> extends JPanel {

    private static final long serialVersionUID = 3835734949707038858L;

    /**
     * The resource bundle for this package.
     */
    protected static final ResourceBundle RESOURCES = Param.RESOURCES;

    /**
     * References to the focus listener used by this panel's text field.
     */
    protected FocusListener _tfFocusListener;

    /**
     * References to the action listener used by this panel's text field.
     */
    protected ActionListener _tfActionListener;

    /**
     * Value param had before becoming the input focus.
     */
    protected double _oldValue;

    /**
     * TextField for inputting a value to this parameter.
     */
    protected DecimalField _theTField;

    /**
     * The combo-box displaying the current units (when there are more than one). This
     * variable has a value of null if there are no units or only 1 unit.
     */
    protected JComboBox _unitList = null;

    /**
     * The label used to display the units of this parameter when there is only one unit.
     * If this variable is null, then there are either no units at all or a list of units
     * (more than one).
     */
    protected JLabel _unitLabel = null;

    /**
     * The number format object used to format numbers
     */
    protected NumberFormat _nf;

    /**
     * The separator between the label and the text field.
     */
    protected JLabel _separator = new JLabel(" =", JLabel.LEFT);

    /**
     * The parameter that this input panel refers to.
     */
    protected Param<Q> _theParam;

    //-----------------------------------------------------------------------------
    /**
     * Don't allow the user to access the default constructor.
     */
    protected ParamInput() {
    }

    /**
     * Constructs a ParamInput class that accepts a Param object for input.
     *
     * @param param The Param object that this input field should be associated with.
     */
    public ParamInput(Param<Q> param) {
        _theParam = requireNonNull(param);

        //  Create layout
        setLayout(new BorderLayout());

        //  Create the label for this parameter.
        Box hbox = Box.createHorizontalBox();
        add(hbox, BorderLayout.WEST);
        JLabel theLabel = new JLabel(param.getLabel(), JLabel.RIGHT);
        hbox.add(theLabel);

        //  Add the separator.
        hbox.add(_separator);

        //  Create text input field for this parameter.
        _nf = param.getNumberFormat();
        if (_nf == null)
            _nf = (DecimalFormat)NumberFormat.getInstance();

        if (param.isDefined())
            _theTField = new DecimalField(param.getValue(), 10, _nf);
        else
            _theTField = new DecimalField(10, _nf);

        _tfActionListener = new TextFieldListener();
        _tfFocusListener = new TextFieldFocusL();
        _theTField.addActionListener(_tfActionListener);
        _theTField.addFocusListener(_tfFocusListener);
        theLabel.setLabelFor(_theTField);
        add(_theTField, BorderLayout.CENTER);

        //  Create the units label or choice menu for this parameter.
        if (param._availUnits != null && param._availUnits.length > 1) {

            //  Many units are available, create a menu of them.
            _unitList = new JComboBox(param.getAvailUnits());
            _unitList.addActionListener(new UnitMenuListener());
            _unitList.setToolTipText(RESOURCES.getString("availUnitsToolTip"));

            //  Select the current units.
            _unitList.setSelectedItem(param.getUnit());
            add(_unitList, BorderLayout.EAST);

        } else {
            //  Only 1 unit available, display as static string.
            String unitStr = "";
            Unit<Q> theUnits = param.getUnit();
            if (theUnits != null) {
                unitStr += theUnits.toString();
            }
            unitStr += " ";
            _unitLabel = new JLabel(unitStr, JLabel.LEFT);
            add(_unitLabel, BorderLayout.EAST);
        }

    }

    //-----------------------------------------------------------------------------
    /**
     * We are adding an inset to our ParamInput panel. This is called by the container to
     * determine our insets.
     *
     * @return An Insets object is returned.
     */
    @Override
    public Insets getInsets() {
        return new Insets(0, 2, 0, 2);
    }

    /**
     * Reads the value input in the text field and puts it into the corresponding
     * parameter. If the TextEdit field is empty, the parameter is made undefined.
     */
    public void getInput() {
        double newValue = _theTField.getValue();

        if (Double.isNaN(newValue))
            _theParam.makeUndefined();

        else {
            //  If the value has changed, reset the parameter.
            if (newValue != _oldValue)
                //  Value changed by user, so reset it.
                _theParam.setValue(newValue);
        }

    }

    /**
     * Method to return the current value of the parameter associated with this ParamInput
     * object.
     *
     * @return The value of the parameter associated with this ParamInput object.
     */
    public Parameter<Q> get() {
        //  First retrieve any user inputs.
        this.getInput();

        //  Now return the parameter value.
        return _theParam.get();
    }

    /**
     * Method to return the current value of the parameter associated with this ParamInput
     * object. The value is returned in the current units.
     *
     * @return The value of the parameter associated with this ParamInput object in the
     *         current units is returned.
     */
    public double getValue() {
        //  First retrieve any user inputs.
        this.getInput();

        //  Now return the parameter value.
        return _theParam.getValue();
    }

    /**
     * Resets the text displayed in the edit text field to the current value of the
     * parameter attached to this object. This should be called if the value of the
     * parameter has changed without being input by the user (e.g.: by an internal
     * calculation).
     */
    public void resetText() {

        if (_theParam.isDefined())
            _theTField.setValue(_theParam.getValue());
        else
            _theTField.setText("");

        //  If there are a list of units, make sure the combo box is
        //  displaying the correct one.
        if (_unitList != null)
            _unitList.setSelectedItem(_theParam.getUnit());

        //  If there is only a single unit, make sure it is correct.
        else if (_unitLabel != null)
            _unitLabel.setText(_theParam.getUnit().toString());

        setOldValue();
    }

    /**
     * Sets the specified boolean to indicate whether or not this ParamInput should be
     * editable.
     *
     * @param editable Flag indicating if this ParamInput should allow editing or not.
     */
    public void setEditable(boolean editable) {
        _theTField.setEditable(editable);
    }

    /**
     * Returns the boolean indicating whether this ParamInput is editable or not.
     *
     * @return true if this ParamInput is editable.
     */
    public boolean isEditable() {
        return _theTField.isEditable();
    }

    /**
     * Set the old value of the parameter associated with this ParamInput object. This is
     * used internally to help determine if a new parameter value has been manually input
     * by the user, or set in some other way (e.g.: by calculation). The old value is in
     * the current units.
     */
    protected void setOldValue() {
        try {

            if (_theParam.isDefined()) {
                _oldValue = _theParam.getValue();

                // Round off the value to the current display format.
                if (_nf != null) {
                    String text = _nf.format(_oldValue);
                    _oldValue = _nf.parse(text).doubleValue();
                }

            } else
                _oldValue = Double.NaN;

        } catch (ParseException err) {
            // Shouldn't be possible.
            err.printStackTrace();
        }

    }

    /**
     * Set the String used to separate the label from the text field. The default
     * separator is " =".
     *
     * @param separator The separator to use. May not be null.
     */
    public void setSeparator(String separator) {
        this._separator.setText(requireNonNull(separator));
    }

    /**
     * Return the separator string used to separate the parameter label from the text
     * field. The default separator String is " =".
     *
     * @return The separator String.
     */
    public String getSeparator() {
        return _separator.getText();
    }

    /**
     * Method to return the value of this parameter as a string.
     *
     * @return Returns a string representation of the value of this parameter.
     */
    @Override
    public String toString() {
        if (_nf != null && _theParam != null)
            return _nf.format(_theParam.getValue());
        return super.toString();
    }

    /**
     * Defines a listener for our unit menu combo box.
     */
    protected class UnitMenuListener implements ActionListener {

        @Override
        public void actionPerformed(ActionEvent evt) {

            JComboBox unitList = (JComboBox)evt.getSource();
            @SuppressWarnings("unchecked")
            Unit<Q> selectedUnit = (Unit<Q>)unitList.getSelectedItem();

            if (!selectedUnit.equals(_theParam.getUnit())) {

                //  Change units to whatever the user selected.
                try {

                    _theParam.changeUnits(selectedUnit);
                    ParamInput.this.resetText();

                } catch (ConversionException e) {
                    /* Can't happen if availUnits are good! */
                }
            }

        }

    }

    /**
     * Defines a listener for our text field that is called whenever the user types in a
     * new number and presses enter.
     */
    protected class TextFieldListener implements ActionListener {

        @Override
        public void actionPerformed(ActionEvent e) {
            ParamInput.this.getInput();
        }
    }

    /**
     * Defines a listener for our text field that is called whenever the focus changes off
     * a particular text field.
     */
    protected class TextFieldFocusL extends FocusAdapter {

        @Override
        public void focusLost(FocusEvent e) {
            ParamInput.this.getInput();
        }

    }

}
