/*
 *   ARvfn_sweepLE_sweep4_taper - Method for calculating vert. tail aspect ratio.
 *
 *   Copyright (C) 1998-2015 by Joseph A. Huwaldt
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.pmethods.winggeom;

import jahuwaldt.js.ParamTools.AbstractParamMethod;
import jahuwaldt.js.param.Parameter;
import static jahuwaldt.js.param.Parameter.ONE;
import static jahuwaldt.js.param.Parameter.tan;
import javax.measure.quantity.Angle;
import javax.measure.quantity.Dimensionless;

/**
 * Calculates planform vertical tail aspect ratio based on leading edge sweep
 * angle, quarter chord sweep angle and vertical taper ratio.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt,  Date: October 21, 2001
 * @version June 18, 2015
 */
public class ARvfn_sweepLE_sweep4_taper extends AbstractParamMethod<Dimensionless> {

    /**
     * Constructor for this method.
     *
     * @param sweepLE Label for the "leading edge sweep" parameter.
     * @param sweep4 Label for the "1/4 chord sweep" parameter.
     * @param taper Label for the "taper ratio" parameter.
     */
    public ARvfn_sweepLE_sweep4_taper(String sweepLE, String sweep4, String taper) {
        _inputLabels = new String[3];
        _inputLabels[0] = sweepLE;
        _inputLabels[1] = sweep4;
        _inputLabels[2] = taper;
    }

    /**
     * Calculates planform vertical tail aspect ratio based on leading edge
     * sweep angle, quarter chord sweep angle and vertical taper ratio. Method:
     *      ARv = (1 - taper)/(1 + taper)/(tan(sweepLE) - tan(sweep4))/2
     *
     * @param inputs Array of input parameters. [0=leading edge sweep angle
     * (sweepLE), 1=quarter chord sweep angle (sweep4), 2=vertical taper ratio (taper)].
     * @return Vertical tail planform aspect ratio (AR).
     */
    @Override
    public Parameter<Dimensionless> calc(Parameter[] inputs) {
        Parameter<Angle> sweepLE = inputs[0];
        Parameter<Angle> sweep4 = inputs[1];
        double taper = inputs[2].getValue();
        Class nd = Dimensionless.class;

        // AR = (1f - taper) / (1f + taper) / (tan( sweepLE ) - tan( sweep4 ))/2;
        double t1 = (1 - taper) / (1 + taper);
        Parameter<Dimensionless> AR = ONE.times(t1 / 2).divide(tan(sweepLE).minus(tan(sweep4))).asType(nd);

        return AR;
    }

}
