/*
 *   Cbarfn_Cr_taper - Method for calculating Cbar as a function of Cr and taper.
 *
 *   Copyright (C) 1998-2016, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.pmethods.winggeom;

import jahuwaldt.js.ParamTools.*;
import jahuwaldt.js.param.Parameter;
import javax.measure.quantity.Length;
import javax.measure.unit.SI;

/**
 * Calculates wing mean aerodynamic chord length based on root chord length, and taper
 * ratio.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: February 6, 1999
 * @version March 26, 2016
 */
@SuppressWarnings("serial")
public class Cbarfn_Cr_taper extends AbstractParamMethod<Length> {

    /**
     * Constructor for this method.
     *
     * @param Cr    Label for the "root chord" parameter.
     * @param taper Label for the "taper ratio" parameter.
     */
    public Cbarfn_Cr_taper(String Cr, String taper) {
        _inputLabels = new String[2];
        _inputLabels[0] = Cr;
        _inputLabels[1] = taper;
    }

    /**
     * Calculates the mean aerodynamic chord length based on root chord length and taper
     * ratio.
     * <p>
     * Method: <code>Cbar = 2/3*Cr*(1 + taper + taper^2)/(1 + taper)</code>
     * </p>
     *
     * @param inputs Array of input parameters. [0=root chord length(Cr), 1=taper ratio
     *               (taper)].
     * @return Mean aerodynamic chord length (Cbar) in base units of length.
     */
    @Override
    public Parameter<Length> calc(Parameter[] inputs) {
        double Cr = inputs[0].doubleValue(SI.METER);
        double taper = inputs[1].getValue();
        double Cbar = 2. / 3. * Cr * (1 + taper + taper * taper) / (1 + taper);
        return Parameter.valueOf(Cbar, SI.METER).to(inputs[0].getUnit());
    }

}
