/*
 *   Crfn_Cbar_taper - Method for calculating Cr as a function of Cbar and taper.
 *
 *   Copyright (C) 1998-2016, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.pmethods.winggeom;

import jahuwaldt.js.ParamTools.*;
import jahuwaldt.js.param.Parameter;
import javax.measure.quantity.Length;
import javax.measure.unit.SI;

/**
 * Calculates wing root chord length based on mean aerodynamic chord, and wing taper
 * ratio.
 * 
 * <p> Modified by: Joseph A. Huwaldt  </p>
 * 
 * @author Joseph A. Huwaldt, Date: February 6, 1999
 * @version March 26, 2016
 */
@SuppressWarnings("serial")
public class Crfn_Cbar_taper extends AbstractParamMethod<Length> {

    /**
     * Constructor for this method.
     *
     * @param Cbar  Label for the "mean aerodynamic chord" parameter.
     * @param taper Label for the "taper ratio" parameter.
     */
    public Crfn_Cbar_taper(String Cbar, String taper) {
        _inputLabels = new String[2];
        _inputLabels[0] = Cbar;
        _inputLabels[1] = taper;
    }

    /**
     * Calculates wing root chord length based on mean aerodynamic chord, and wing taper
     * ratio.
     * <p>
     * Method:  <code>Cr = 3/2*Cbar*(1 + taper)/(1 + taper + taper^2)</code>
     * </p>
     *
     * @param inputs Array of input parameters. [0=mean aerodynamic chord (Cbar), 1=wing
     *               taper ratio (taper)].
     * @return Root chord length (Cr) in base units of length.
     */
    @Override
    public Parameter<Length> calc(Parameter[] inputs) {
        double Cbar = inputs[0].doubleValue(SI.METER);
        double taper = inputs[1].getValue();
        double Cr = 3. * 0.5 * Cbar * (1 + taper) / (1 + taper + taper * taper);
        return Parameter.valueOf(Cr, SI.METER).to(inputs[0].getUnit());
    }

}
