/*
*   bfn_AR_Cr_Ct - Method for calculating b as a function of AR, Cr and Ct.
*
*   Copyright (C) 1998-2009 by Joseph A. Huwaldt
*   All rights reserved.
*   
*   This library is free software; you can redistribute it and/or
*   modify it under the terms of the GNU Lesser General Public
*   License as published by the Free Software Foundation; either
*   version 2.1 of the License, or (at your option) any later version.
*   
*   This library is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   Library General Public License for more details.
*
*  You should have received a copy of the GNU Lesser General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*  Or visit:  http://www.gnu.org/licenses/lgpl.html
**/
package jahuwaldt.js.pmethods.winggeom;


import javax.measure.quantity.Length;
import javax.measure.unit.SI;

import jahuwaldt.js.param.Parameter;
import jahuwaldt.js.ParamTools.*;

/**
*  <p>  Calculates planform wing span based on aspect ratio,
*       wing root chord length, and wing tip chord length.
*  </p>
*
*  <p>  Modified by:  Joseph A. Huwaldt   </p>
*
*  @author   Joseph A. Huwaldt   Date:  February 6, 1999
*  @version  April 26, 2009
**/
public class bfn_AR_Cr_Ct extends AbstractParamMethod<Length> {

	/**
	*  Constructor for this method.
	*
	*  @param  AR  Label for the "aspect ratio" parameter.
	*  @param  Cr  Label for the "root chord" parameter.
	*  @param  Ct  Label for the "tip chord" parameter.
	**/
	public bfn_AR_Cr_Ct( String AR, String Cr, String Ct ) {
		_inputLabels = new String [ 3 ];
		_inputLabels[0] = AR;
		_inputLabels[1] = Cr;
		_inputLabels[2] = Ct;
	}

	/**
	*  Calculates planform wing span based on aspect ratio,
	*  wing root chord length, and wing tip chord length.
	*  Method:  b = 0.5*AR*(Cr + Ct)
	*
	*  @param  inputs  Array of input parameters.
	*                  [0=wing aspect ratio (AR), 1=wing root chord length (Cr),
	*                   2=wing tip chord length (Ct)].
	*  @return  Planform wing span (b) in base length units.
	**/
	public Parameter<Length> calc( Parameter[] inputs ) {
		Parameter AR = inputs[0];
		Parameter Cr = inputs[1];
		Parameter Ct = inputs[2];
		Parameter b = AR.times(Cr.plus(Ct)).divide(2);		//	b = 0.5 * AR * (Cr + Ct)
		return b.to(Cr.getUnit());
	}


}


