/*
 *   sweep4fn_sweepLE_AR_taper - Method for calculating 1/4c sweep as a function of sweepLE, AR and taper.
 *
 *   Copyright (C) 1998-2016, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.pmethods.winggeom;

import jahuwaldt.js.ParamTools.*;
import jahuwaldt.js.param.Parameter;
import javax.measure.quantity.Angle;
import javax.measure.unit.SI;

/**
 * Calculates quarter chord wing sweep angle based on the wing leading edge sweep angle,
 * wing aspect ratio, and wing taper ratio.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: February 7, 1999
 * @version March 26, 2016
 */
@SuppressWarnings("serial")
public class sweep4fn_sweepLE_AR_taper extends AbstractParamMethod<Angle> {

    /**
     * Constructor for this method.
     *
     * @param sweepLE Label for the "leading edge sweep" parameter.
     * @param AR      Label for the "aspect ratio" parameter.
     * @param taper   Label for the "taper ratio" parameter.
     */
    public sweep4fn_sweepLE_AR_taper(String sweepLE, String AR, String taper) {
        _inputLabels = new String[3];
        _inputLabels[0] = sweepLE;
        _inputLabels[1] = AR;
        _inputLabels[2] = taper;
    }

    /**
     * Calculates quarter chord wing sweep angle based on the wing leading edge sweep
     * angle, wing aspect ratio, and wing taper ratio.
     * <p>
     * Method:
     * <code>sweep4 = atan( tan(sweepLE) - 4/AR*(1/4-0)*(1 - taper)/(1 + taper) )</code>
     * </p>
     *
     * @param inputs Array of input parameters. [0=wing leading edge sweep angle
     *               (sweepLE), 1=wing aspect ratio (AR), 2=wing taper ratio (taper)].
     * @return Wing sweep angle at the quarter chord in base angular units (radians).
     */
    @Override
    public Parameter<Angle> calc(Parameter[] inputs) {
        double sweepLE = inputs[0].doubleValue(SI.RADIAN);
        double AR = inputs[1].getValue();
        double taper = inputs[2].getValue();
        double sweep4 = Math.atan(Math.tan(sweepLE) - (1 - taper) / (1 + taper) / AR);
        return Parameter.valueOf(sweep4, SI.RADIAN).to(inputs[0].getUnit());
    }

}
