/*
 *   sweepLEvfn_sweepTE_AR_taper - Method for calculating LE sweep of a vertical tail 
 *               as a function of sweepTE, AR and taper.
 *
 *   Copyright (C) 2001-2015 by Joseph A. Huwaldt
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.pmethods.winggeom;

import jahuwaldt.js.ParamTools.AbstractParamMethod;
import jahuwaldt.js.param.Parameter;
import javax.measure.quantity.Angle;
import javax.measure.unit.SI;

/**
 * <p>
 * Calculates vertical tail leading edge sweep angle based on the vertical
 * trailing sweep angle, vertical aspect ratio, and vertical taper ratio.
 * </p>
 *
 * <p>
 * Modified by: Joseph A. Huwaldt </p>
 * 
* @author Joseph A. Huwaldt Date: October 18, 2001
 * @version June 19, 2015
 */
public class sweepLEvfn_sweepTE_AR_taper extends AbstractParamMethod<Angle> {

    /**
     * Constructor for this method.
     *
     * @param sweepTE Label for the "sweep at the trailing edge" parameter.
     * @param AR Label for the "aspect ratio" parameter.
     * @param taper Label for the "taper ratio" parameter.
     */
    public sweepLEvfn_sweepTE_AR_taper(String sweepTE, String AR, String taper) {
        _inputLabels = new String[3];
        _inputLabels[0] = sweepTE;
        _inputLabels[1] = AR;
        _inputLabels[2] = taper;
    }

    /**
     * Calculates leading edge vertical tail sweep angle based on the vertical
     * trailing edge sweep angle, vertical aspect ratio, and vertical taper
     * ratio. Method:
     *      sweepLEv = atan( tan(sweepTE) - 4/ARv/2*(0-1)*(1 - taper)/(1 + taper) )
     *
     * @param inputs Array of input parameters. [0=vertical trailing edge sweep
     * angle (sweepTE), 1=vertical aspect ratio (AR), 2=vertical taper ratio
     * (taper)].
     * @return Vertical tail sweep angle at the leading edge in angular units.
     */
    @Override
    public Parameter<Angle> calc(Parameter[] inputs) {
        double sweepTE = inputs[0].doubleValue(SI.RADIAN);
        double AR = inputs[1].getValue();
        double taper = inputs[2].getValue();
        double sweepLE = Math.atan(Math.tan(sweepTE) + 4 / AR / 2 * (1 - taper) / (1 + taper));
        return Parameter.valueOf(sweepLE, SI.RADIAN).to(inputs[0].getUnit());
    }

}
