/*
 *   taperfn_Cr_Cbar - Method for calculating taper as a function of Cr and Cbar.
 *
 *   Copyright (C) 1998-2016, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.pmethods.winggeom;

import jahuwaldt.js.ParamTools.*;
import jahuwaldt.js.param.Parameter;
import javax.measure.quantity.Dimensionless;
import javax.measure.unit.SI;

/**
 * Calculates wing taper ratio based on wing root chord and mean aerodynamic chord length.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: February 6, 1999
 * @version March 26, 2016
 */
@SuppressWarnings("serial")
public class taperfn_Cr_Cbar extends AbstractParamMethod<Dimensionless> {

    /**
     * Constructor for this method.
     *
     * @param Cr   Label for the "root chord" parameter.
     * @param Cbar Label for the "mean aero chord" parameter.
     */
    public taperfn_Cr_Cbar(String Cr, String Cbar) {
        _inputLabels = new String[2];
        _inputLabels[0] = Cr;
        _inputLabels[1] = Cbar;
    }

    /**
     * Calculates wing taper ratio based on wing root chord and mean aerodynamic chord
     * length.
     * <p>
     * Method:
     * <code>taper = 1/(4*Cr)*(3Cbar - 2Cr + sqrt(3*(Cbar + 2*Cr)*(3*Cbar - 2*Cr)))</code>
     * </p>
     *
     * @param inputs Array of input parameters. [0=root chord length (Cr), 1=mean
     *               aerodynamic chord (Cbar)].
     * @return Wing taper ratio (taper).
     */
    @Override
    public Parameter<Dimensionless> calc(Parameter[] inputs) {
        double Cr = inputs[0].doubleValue(SI.METER);
        double Cbar = inputs[1].doubleValue(SI.METER);
        double temp = 2 * Cr;
        double taper = (3 * Cbar - 2 * Cr + Math.sqrt(3 * (Cbar + temp) * (3 * Cbar - temp))) * 0.25 / Cr;
        return Parameter.valueOf(taper, Dimensionless.UNIT);
    }

}
