/*
 *   FLABSavReader -- A class that can read and write FlightLab SAV multi-table files.
 *
 *   Copyright (C) 2000-2024, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.tables;

import static jahuwaldt.js.tables.FTableReader.RESOURCES;
import jahuwaldt.tools.math.MathTools;
import jahuwaldt.util.ExponentialFormat;
import java.io.*;
import java.text.DateFormat;
import java.text.MessageFormat;
import static java.text.MessageFormat.format;
import java.text.NumberFormat;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import static java.util.Objects.requireNonNull;
import javolution.text.Text;
import javolution.util.FastTable;
import org.jscience.mathematics.number.Integer64;
import org.jscience.mathematics.vector.Float64Vector;

/**
 * A class that provides methods for reading or writing an ART FlightLab SAV formatted
 * multi-table file. These files are used by the FlightLab simulation system to store
 * fuselage and airfoil aerodynamics data. The format is very restrictive and can only be
 * used to store the particular tables that FlightLab is looking for. See ART's FlightLab
 * documentation for more information.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: April 3, 2000
 * @version January 15, 2024
 * @see http://www.flightlab.com
 */
public class FLABSavReader implements FTableReader {

    //  The preferred file extension for files of this reader's type.
    private static final String EXTENSION = "sav";

    //  Different sub-types of FlightLab SAV files.
    private static final int UNKNOWN = 0;
    private static final int HIGH_LOW_AIRFOIL = 2;
    private static final int FUSELAGE2D_UNIFORM = 3;

    //  The airfoil table with low/high table names.
    private static final String[] AF_LOW_HIGH = {"CLL", "CDL", "CML", "CLH", "CDH", "CMH"};
    private static final int[] AF_LOW_HIGH_DIM = {2, 2, 2, 1, 1, 1};

    //  The airfoil low/high independent names.
    private static final String[] AF_LOW_HIGH_IND = {"AOATL", "AOATH", "MACHT"};

    //  The fuselage 2D uniform low/high table names.
    private static final String[] FUSELAGE2D_U = {
        "CDFAH", "CDFAL", "CDFBH", "CDFBL", "CLFAH", "CLFAL",
        "CLFBAH", "CLFBAL", "CMFAL", "CMFAH", "CMFBAH", "CMFBAL",
        "CNFBAH", "CNFBAL", "CRFBAH", "CRFBAL", "CYFBAH", "CYFBAL"};
    private static final int[] FUSELAGE2D_UDIM = {
        1, 1, 1, 1, 1, 1,
        2, 2, 1, 1, 2, 2,
        2, 2, 2, 2, 2, 2};

    //  The fuselage 2D uniform low/high independent names.
    private static final String[] FUSELAGE2D_UI = {
        "FAOAT1", "FAOAT2", "FAOST1", "FAOST2", "FAOAT3"};

    //  The fuselage 2D uniform low/high independent table names that go with each dependent.
    private static final String[] FUSELAGE2D_UTI = {
        "FAOAT2", "FAOAT1", "FAOST2", "FAOST1", "FAOAT2", "FAOAT1",
        "FAOST2", "FAOST1", "FAOAT1", "FAOAT2", "FAOST2", "FAOST1",
        "FAOST2", "FAOST1", "FAOST2", "FAOST1", "FAOST2", "FAOST1"};

    /**
     * Returns a string representation of the object. This will return a brief description
     * of the format read by this reader.
     */
    @Override
    public String toString() {
        return RESOURCES.getString("flabDescription");
    }

    /**
     * Returns the preferred file extension (not including the ".") for files of this
     * reader's type.
     *
     * @return The preferred file extension "sav".
     */
    @Override
    public String getExtension() {
        return EXTENSION;
    }

    /**
     * Method that determines if this reader can read data from the specified input
     * stream. Will return FTableReader.YES if the stream is recognized as one of the two
     * SAV format subtypes that this reader recognizes (airfoil or fuselage aero).
     *
     * @param name  The name of the file.
     * @param input An input stream containing the data to be read. Any methods that read
     *              from this stream must first set a mark and then reset back to that
     *              mark before the method returns (even if it returns with an exception).
     * @return FTableReader.NO if the file is not recognized at all or FTableReader.SAV if
     *         the stream is recognize as a FlightLab SAV file.
     */
    @Override
    public int canReadData(String name, BufferedInputStream input) throws IOException {
        requireNonNull(name, MessageFormat.format(RESOURCES.getString("paramNullErr"), "name"));
        requireNonNull(input, MessageFormat.format(RESOURCES.getString("paramNullErr"), "input"));
        
        //  Mark the current position in the input stream so that we can return to it.
        input.mark(10240);  //  10KB buffer

        //  FLAB Save files are required to be in ASCII with U.S. style number formatting.
        //  Get the default locale and save it.  Then change the default locale to US.
        Locale defLocale = Locale.getDefault();
        Locale.setDefault(Locale.US);

        int response = NO;
        try {
            //  Look for one of the specific SAV file sub-types that we can read.
            StreamTokenizer tokenizer = new StreamTokenizer(new InputStreamReader(input));

            // Set some tokenizer options.
            tokenizer.wordChars('!', 'z');
            tokenizer.commentChar('#');
            tokenizer.commentChar(';');
            tokenizer.parseNumbers();
            tokenizer.eolIsSignificant(false);

            // Determine SAV file type (kHighLowAirfoil or kFuselage2DUniform).
            int type = getType(tokenizer);
            if (type != UNKNOWN)
                response = YES;

        } catch (Throwable ignore) {
        } finally {
            //  Restore the default locale.
            Locale.setDefault(defLocale);

            input.reset();
        }

        return response;
    }

    /**
     * Returns true. This class can write data to a FlightLab SAV file.
     */
    @Override
    public boolean canWriteData() {
        return true;
    }

    /**
     * Compares this object with the specified object for order based on the
     * <code>toString().compareTo(o.toString())</code> method. Returns a negative integer,
     * zero, or a positive integer as this object is less than, equal to, or greater than
     * the specified object.
     */
    @Override
    public int compareTo(FTableReader o) {
        return this.toString().compareTo(o.toString());
    }

    /**
     * Method that reads in ART FlightLab SAV formatted data from the specified input
     * stream and returns that data as an FTableList object.
     *
     * @param input An input stream containing the SAV formatted data.
     * @return An FTableList object that contains the data read in from the specified
     *         stream.
     * @throws IOException If there is a problem reading the specified stream.
     */
    @Override
    public FTableList read(InputStream input) throws IOException {
        requireNonNull(input, MessageFormat.format(RESOURCES.getString("paramNullErr"), "input"));
        
        //  FLAB Save files are required to be in ASCII with U.S. style number formatting.
        //  Get the default locale and save it.  Then change the default locale to US.
        Locale defLocale = Locale.getDefault();
        Locale.setDefault(Locale.US);

        try {
            // Get a reference to the file and create a tokenizer.
            BufferedReader reader = new BufferedReader(new InputStreamReader(input));
            StreamTokenizer tokenizer = new StreamTokenizer(reader);

            FTableList db = readStream(tokenizer);

            return db;
            
        } finally {
            //  Restore the default locale.
            Locale.setDefault(defLocale);
        }
    }

    /**
     * Method for writing out all the data stored in the specified list of Float64Table
     * objects to the specified output stream in SAV format. The FlightLab SAV format
     * requires a specific number of tables and each table must have very specific names
     * and size constraints. See ART documentation for more information.
     *
     * @param output The output stream to which the data is to be written.
     * @param tables The list of Float64Table objects to be written out.
     * @throws IOException If there is a problem writing to the specified stream.
     * @see http://www.flightlab.com
     */
    @Override
    public void write(OutputStream output, FTableList tables) throws IOException {
        requireNonNull(output, MessageFormat.format(RESOURCES.getString("paramNullErr"), "output"));
        requireNonNull(tables, MessageFormat.format(RESOURCES.getString("paramNullErr"), "tables"));
        
        if (tables.isEmpty())
            return;

        //  FLAB Save files are required to be in ASCII with U.S. style number formatting.
        //  Get the default locale and save it.  Then change the default locale to US.
        Locale defLocale = Locale.getDefault();
        Locale.setDefault(Locale.US);

        try {
            BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(output));

            // Check for consistant tables and get the appropriate table type.
            int type = checkTables(tables);

            // Write out the tables.
            switch (type) {
                case HIGH_LOW_AIRFOIL:
                    writeHighLowAirfoil(writer, tables);
                    break;

                case FUSELAGE2D_UNIFORM:
                    writeFuselage2DUniform(writer, tables);
                    break;

                default:
                    throw new IOException(RESOURCES.getString("flabInvaldFormatMsg"));
            }

            //  Flush the output.
            writer.flush();

        } finally {
            //  Restore the default locale.
            Locale.setDefault(defLocale);
        }
    }

    /**
     * Read in an FlightLab SAV formatted multi-table stream and create a new table
     * database with the new files.
     *
     * @param tokenizer Tokenizer for the file being read in.
     * @return A table database containing the tables read in from the file.
     */
    private static FTableList readStream(StreamTokenizer tokenizer) throws IOException {

        // Set some tokenizer options.
        tokenizer.wordChars('!', 'z');
        tokenizer.commentChar('#');
        tokenizer.commentChar(';');
        tokenizer.parseNumbers();
        tokenizer.eolIsSignificant(false);

        // Determine SAV file type (kHighLowAirfoil or kFuselage2DUniform).
        int type = getType(tokenizer);

        //  Read in the table data based on SAV file sub-type.
        FTableList db = null;

        switch (type) {
            case HIGH_LOW_AIRFOIL:
                db = readHighLowAirfoil(tokenizer);
                break;

            case FUSELAGE2D_UNIFORM:
                db = readFuselage2DUniform(tokenizer);
                break;

            default:
                throw new IOException(RESOURCES.getString("flabUnkSubTypeMsg"));
        }

        return db;
    }

    /**
     * Return the FlightLab SAV file sub-type. SAV files are used in FlightLab to save a
     * variety of specific data sets. This checks the file to determine what specific data
     * set is stored in this SAV file. Returns one of the constants at the top of this
     * class.
     */
    private static int getType(StreamTokenizer tokenizer) throws IOException {

        int tokentype = tokenizer.nextToken();

        switch (tokentype) {
            case StreamTokenizer.TT_EOF:
                throw new IOException(format(RESOURCES.getString("EOFErrMsg"), tokenizer.lineno()));
            case StreamTokenizer.TT_NUMBER:
                throw new IOException(format(RESOURCES.getString("expWordErrMsg"), tokenizer.lineno()));
            case StreamTokenizer.TT_WORD:
                //  We found a word.  Should be the name of the 1st parameter.
                //  This should tell us the file sub-type.
                String word = tokenizer.sval;
                tokenizer.pushBack();
                if (word.equals("AOATL"))
                    //  We've got an airfoil table with low/high AOA and uniform increment.
                    return HIGH_LOW_AIRFOIL;
                
                else if (word.startsWith("FAOA") || word.startsWith("FAOS"))
                    //  We've got a fuselage 2D uniform table.
                    return FUSELAGE2D_UNIFORM;
                break;
            default:
                break;
        }

        return UNKNOWN;
    }

    /**
     * Read in an FlightLab SAV airfoil table with low/high and uniform increment
     * formatted multi-table stream and create a new table database.
     *
     * @param tokenizer Tokenizer for the file being read in.
     * @return A table database containing the tables read in from the file.
     */
    private static FTableList readHighLowAirfoil(StreamTokenizer tokenizer)
            throws IOException {

        // Start checking for End-Of-Line (EOL).
        tokenizer.eolIsSignificant(true);

        //  Read in all the independent tables.
        int numIndeps = AF_LOW_HIGH_IND.length;
        boolean[] found = new boolean[numIndeps];
        double[] indepMin = new double[numIndeps];
        double[] indepMax = new double[numIndeps];
        double[] indepDelta = new double[numIndeps];
        for (int i = 0; i < numIndeps; ++i) {

            //  Get the next token.  Should be one of the independent table names.
            String word = TRUtils.nextWord(tokenizer);

            //  Did we find a valid independent table name?
            int idx = idx2str(word, AF_LOW_HIGH_IND, found, RESOURCES.getString("flabUnexpIndepNameMsg"));

            // Skip over parameter count data as it should always be "3 1 0".
            TRUtils.nextLine(tokenizer);

            //  Retrieve the min, max and delta values for this independent.
            double min = TRUtils.nextNumber(tokenizer);
            double max = TRUtils.nextNumber(tokenizer);
            double delta = TRUtils.nextNumber(tokenizer);
            TRUtils.nextLine(tokenizer);
            TRUtils.nextLine(tokenizer);

            indepMin[idx] = min;
            indepMax[idx] = max;
            indepDelta[idx] = delta;
        }

        //  Create names for independent variables.
        String[] indepL = new String[2];
        indepL[0] = "MACHT";
        indepL[1] = "AOATL";
        String indepH = "AOATH";

        //  Extract range data.
        double aoatlmax = indepMax[0];
        double aoatlmin = indepMin[0];
        double aoatldelta = indepDelta[0];
        double aoathmax = indepMax[1];
        double aoathmin = indepMin[1];
        double aoathdelta = indepDelta[1];
        double machtmax = indepMax[2];
        double machtmin = indepMin[2];
        double machtdelta = indepDelta[2];

        //  Create breakpoint tables for independent variables.
        int numBPaoatl = (int)((aoatlmax - aoatlmin) / aoatldelta) + 1;
        int numBPaoath = (int)((aoathmax - aoathmin) / aoathdelta) + 1;
        int numBPmacht = (int)((machtmax - machtmin) / machtdelta) + 1;

        double[][] breakpointsL = new double[2][];
        breakpointsL[0] = new double[numBPmacht];
        breakpointsL[1] = new double[numBPaoatl];
        double[] breakpointsH = new double[numBPaoath];

        for (int i = 0; i < numBPmacht; ++i)
            breakpointsL[0][i] = machtmin + machtdelta * i;

        for (int i = 0; i < numBPaoatl; ++i)
            breakpointsL[1][i] = aoatlmin + aoatldelta * i;

        for (int i = 0; i < numBPaoath; ++i)
            breakpointsH[i] = aoathmin + aoathdelta * i;

        //  Create an array to contain the tables to be read in (6 tables).
        Float64Table[] tables = new Float64Table[6];

        //  Read in all the tables one after another.
        found = new boolean[6];
        for (int i = 0; i < 6; ++i) {

            //  Get the next token (should be a dependent table name).
            tokenizer.eolIsSignificant(false);
            String tblName = TRUtils.nextWord(tokenizer);

            //  Did we find a valid dependent table name?
            int idx = idx2str(tblName, AF_LOW_HIGH, found, RESOURCES.getString("flabUnexpDepNameMsg"));

            //  Deal with 1D and 2D tables a little differently.
            if (AF_LOW_HIGH_DIM[idx] == 1) {
                //  Create the table without any dependent data.
                Float64Table table = Float64Table.valueOf(tblName, indepH, breakpointsH, null);

                //  Put table into array of tables.
                tables[idx] = table;

                //  Read in the data for this table.
                read1DTableData(tokenizer, tblName, "AOATH", breakpointsH.length, table);

            } else {
                //  Must be a 2D table.

                Float64Table table = Float64Table.valueOf(tblName, indepL, breakpointsL, null);
                tables[idx] = table;
                read2DTableData(tokenizer, tblName, "AOATL", breakpointsL[1].length, breakpointsL[0].length, table);
            }

        }

        // Create a new table database.
        FTableList db = FTableList.valueOf(tables);

        return db;
    }

    /**
     * Read in an FlightLab SAV fuselage 2D uniform increment formatted multi-table stream
     * and create a new table database.
     *
     * @param tokenizer Tokenizer for the file being read in.
     * @return A table database containing the tables read in from the file.
     */
    private static FTableList readFuselage2DUniform(StreamTokenizer tokenizer)
            throws IOException {

        // Start checking for End-Of-Line (EOL).
        tokenizer.eolIsSignificant(true);

        //  Read in all the independent tables.
        int numIndeps = FUSELAGE2D_UI.length;
        boolean[] found = new boolean[numIndeps];
        double[] indepMin = new double[numIndeps];
        double[] indepMax = new double[numIndeps];
        double[] indepDelta = new double[numIndeps];
        for (int i = 0; i < numIndeps; ++i) {

            //  Get the next token.  Should be one of the independent table names.
            String word = TRUtils.nextWord(tokenizer);

            //  Did we find a valid independent table name?
            int idx = idx2str(word, FUSELAGE2D_UI, found, RESOURCES.getString("flabUnexpIndepNameMsg"));

            // Skip over parameter count data as it should always be "3 1 0".
            TRUtils.nextLine(tokenizer);

            //  Retrieve the min, max and delta values for this independent.
            double min = TRUtils.nextNumber(tokenizer);
            double max = TRUtils.nextNumber(tokenizer);
            double delta = TRUtils.nextNumber(tokenizer);
            TRUtils.nextLine(tokenizer);
            TRUtils.nextLine(tokenizer);

            indepMin[idx] = min;
            indepMax[idx] = max;
            indepDelta[idx] = delta;
        }

        //  Create names for independent variables.
        String[] indepLAS = new String[2];
        indepLAS[0] = FUSELAGE2D_UI[4];     //  FAOAT3
        indepLAS[1] = FUSELAGE2D_UI[2];     //  FAOST1
        String[] indepHAS = new String[2];
        indepHAS[0] = FUSELAGE2D_UI[4];     //  FAOAT3
        indepHAS[1] = FUSELAGE2D_UI[3];     //  FAOST2

        //  Extract range data.
        double faoat1max = indepMax[0];
        double faoat1min = indepMin[0];
        double faoat1delta = indepDelta[0];
        double faoat2max = indepMax[1];
        double faoat2min = indepMin[1];
        double faoat2delta = indepDelta[1];
        double faost1max = indepMax[2];
        double faost1min = indepMin[2];
        double faost1delta = indepDelta[2];
        double faost2max = indepMax[3];
        double faost2min = indepMin[3];
        double faost2delta = indepDelta[3];
        double faoat3max = indepMax[4];
        double faoat3min = indepMin[4];
        double faoat3delta = indepDelta[4];

        //  Create breakpoint tables for independent variables.
        int numBPfaoat1 = (int)((faoat1max - faoat1min) / faoat1delta) + 1;
        int numBPfaoat2 = (int)((faoat2max - faoat2min) / faoat2delta) + 1;
        int numBPfaoat3 = (int)((faoat3max - faoat3min) / faoat3delta) + 1;
        int numBPfaost1 = (int)((faost1max - faost1min) / faost1delta) + 1;
        int numBPfaost2 = (int)((faost2max - faost2min) / faost2delta) + 1;

        double[] bpLA = new double[numBPfaoat1];
        double[] bpHA = new double[numBPfaoat2];
        double[] bpLS = new double[numBPfaost1];
        double[] bpHS = new double[numBPfaost2];
        double[][] bpLAS = new double[2][];
        bpLAS[0] = new double[numBPfaoat3];
        bpLAS[1] = new double[numBPfaost1];
        double[][] bpHAS = new double[2][];
        bpHAS[0] = new double[numBPfaoat3];
        bpHAS[1] = new double[numBPfaost2];

        for (int i = 0; i < numBPfaoat3; ++i) {
            double value = faoat3min + faoat3delta * i;
            bpLAS[0][i] = value;
            bpHAS[0][i] = value;
        }
        for (int i = 0; i < numBPfaost1; ++i)
            bpLAS[1][i] = faost1min + faost1delta * i;
        for (int i = 0; i < numBPfaost2; ++i)
            bpHAS[1][i] = faost2min + faost2delta * i;

        for (int i = 0; i < numBPfaoat1; ++i)
            bpLA[i] = faoat1min + faoat1delta * i;

        for (int i = 0; i < numBPfaoat2; ++i)
            bpHA[i] = faoat2min + faoat2delta * i;

        for (int i = 0; i < numBPfaost1; ++i)
            bpLS[i] = faost1min + faost1delta * i;

        for (int i = 0; i < numBPfaost2; ++i)
            bpHS[i] = faost2min + faost2delta * i;

        //  Create an array to contain the tables to be read in (18 tables).
        Float64Table[] tables = new Float64Table[18];

        //  Read in all the tables one after another.
        found = new boolean[18];
        for (int i = 0; i < 18; ++i) {

            //  Get the next token (should be a dependent table name).
            tokenizer.eolIsSignificant(false);
            String tblName = TRUtils.nextWord(tokenizer);

            //  Did we find a valid dependent table name?
            int idx = idx2str(tblName, FUSELAGE2D_U, found, RESOURCES.getString("flabUnexpDepNameMsg"));

            //  Deal with 1D and 2D tables a little differently.
            if (FUSELAGE2D_UDIM[idx] == 1) {

                String indepName = FUSELAGE2D_UTI[idx];
                double indepBP[];
                if (indepName.endsWith("AT1"))
                    indepBP = bpLA;
                else if (indepName.endsWith("AT2"))
                    indepBP = bpHA;
                else if (indepName.endsWith("ST1"))
                    indepBP = bpLS;
                else // if (indepName.endsWith("ST2"))
                    indepBP = bpHS;

                //  Create the table without any dependent data.
                Float64Table table = Float64Table.valueOf(tblName, indepName, indepBP, null);

                //  Put table into array of tables.
                tables[idx] = table;

                //  Read in the data for this table.
                read1DTableData(tokenizer, tblName, indepName, indepBP.length, table);

            } else {
                //  Must be a 2D table.

                String indepName = FUSELAGE2D_UTI[idx];
                double indepBP[][] = null;
                String[] indepNames = null;
                if (indepName.endsWith("ST1")) {
                    indepBP = bpLAS;
                    indepNames = indepLAS;
                } else if (indepName.endsWith("ST2")) {
                    indepBP = bpHAS;
                    indepNames = indepHAS;
                }

                Float64Table table = Float64Table.valueOf(tblName, indepNames, indepBP, null);
                tables[idx] = table;
                read2DTableData(tokenizer, tblName, indepName, indepBP[1].length, indepBP[0].length, table);
            }
        }

        // Create a new table database.
        FTableList db = FTableList.valueOf(tables);

        return db;
    }

    /**
     * Method that searches through an array of strings and returns the index of the
     * string matching the input string.
     *
     * @param str   The string we are looking for.
     * @param list  The list of strings to search through.
     * @param found A list of booleans that indicate if the particular string has already
     *              been found (and so shouldn't be found again).
     * @param msg   The error message to place in the exception thrown if the string is
     *              not found.
     * @return The index in the list of the input string. If the string is not found, an
     *         exception is thrown.
     */
    private static int idx2str(String str, String[] list, boolean[] found, String msg) throws IOException {

        int length = list.length;
        int j;
        for (j = 0; j < length; ++j) {
            if (!found[j] && str.equals(list[j])) {
                found[j] = true;
                break;
            }
        }

        if (j >= length) {
            StringBuilder buffer = new StringBuilder();
            for (j = 0; j < length; ++j) {
                if (!found[j]) {
                    if (j != 0)
                        buffer.append(",");
                    buffer.append(" ");
                    buffer.append(list[j]);
                }
            }
            String msg2 = format(RESOURCES.getString("flabFoundErrMsg"), msg, str, buffer.toString());
            throw new IOException(msg2);
        }

        return j;
    }

    /**
     * Method that reads in the data for a single 1D table from the SAV file.
     */
    private static void read1DTableData(StreamTokenizer tokenizer, String tblName, String aoaName,
            int numAOABP, Float64Table table) throws IOException {

        //  Read in the parameter count to make sure it jives.
        int count = (int)TRUtils.nextNumber(tokenizer);
        if (count != numAOABP)
            throw new IOException(format(RESOURCES.getString("flabWrongNumBPMsg"), tblName, aoaName));
        count = (int)TRUtils.nextNumber(tokenizer);
        if (count != 1)
            throw new IOException(format(RESOURCES.getString("flabWrongNumDimMsg"), tblName));

        //  Skip to the next line.
        tokenizer.eolIsSignificant(true);
        TRUtils.nextLine(tokenizer);
        tokenizer.eolIsSignificant(false);

        //  Read in the dependent data.
        for (int j = 0; j < numAOABP; ++j) {
            double value = TRUtils.nextNumber(tokenizer);
            table.set(j, value);
        }

    }

    /**
     * Method that reads in the data for a single 2D table from the SAV file.
     */
    private static void read2DTableData(StreamTokenizer tokenizer, String tblName, String aoaName,
            int numAOABP, int numMachBP, Float64Table table) throws IOException {

        //  Read in the parameter count to make sure it jives.
        int count = (int)TRUtils.nextNumber(tokenizer);
        if (count != numAOABP)
            throw new IOException(format(RESOURCES.getString("flabWrongNumBPMsg"), tblName, aoaName));
        count = (int)TRUtils.nextNumber(tokenizer);
        if (count != numMachBP)
            throw new IOException(format(RESOURCES.getString("flabWrongNumBPMsg"), tblName, "MACHT or FAOAT3"));

        //  Go to next line.
        tokenizer.eolIsSignificant(true);
        TRUtils.nextLine(tokenizer);
        tokenizer.eolIsSignificant(false);

        //  Read in the dependent data.
        FastTable<Integer64> pos = FastTable.newInstance();
        pos.add(Integer64.ZERO);
        pos.add(Integer64.ZERO);
        for (int i = 0; i < numMachBP; ++i) {
            pos.set(0, Integer64.valueOf(i));           //  pos[0] = i;
            for (int j = 0; j < numAOABP; ++j) {
                pos.set(1, Integer64.valueOf(j));       //  pos[1] = j;
                double value = TRUtils.nextNumber(tokenizer);
                table.set(pos, value);
            }
        }
        FastTable.recycle(pos);

    }

    /**
     * Check the tables that are to be written out to make sure that they meet the
     * requirements of either a 2D airfoil data table set or a fuselage 2D uniform table
     * set.
     *
     * @param tables A collection of tables that we are writing out.
     * @return The type of the table set we are dealing with. A value of 0 is returned if
     *         the tables do not meet any of our SAV sub-type format requirements.
     */
    private static int checkTables(FTableList tables) throws IOException {
        int type = UNKNOWN;

        int size = tables.size();
        if (size == 6) {
            //  We may have an airfoil high/low table set.

            //  Build up an ordered array of tables with the correct names.
            //  If the properly named table can't be found, bail out.
            Float64Table[] tblArray = new Float64Table[6];
            for (int i = 0; i < 6; ++i) {
                Float64Table tbl = tables.get(AF_LOW_HIGH[i]);
                if (tbl == null)
                    return UNKNOWN;
                tblArray[i] = tbl;
            }

            //  Make sure the 3 low alpha tables have the correct independents.
            for (int i = 0; i < 3; ++i) {
                Float64Table tbl = tblArray[i];
                if (tbl.getDimensions() != 2)
                    return UNKNOWN;
                List<Text> indeps = tbl.getIndepNames();
                if (!indeps.get(0).toString().equals("MACHT") || !indeps.get(1).toString().equals("AOATL"))
                    return UNKNOWN;
            }

            //  Make sure the 3 high alpha tables have the correct independents.
            for (int i = 3; i < 6; ++i) {
                Float64Table tbl = tblArray[i];
                if (tbl.getDimensions() != 1)
                    return UNKNOWN;
                List<Text> indeps = tbl.getIndepNames();
                if (!indeps.get(0).toString().equals("AOATH"))
                    return UNKNOWN;
            }

            //  We must have an airfoil low/high alpha table set.
            type = HIGH_LOW_AIRFOIL;

        } else if (size == 18) {
            //  We may have a fuselage 2D uniform table set.

            //  Make sure we have all the required tables.
            for (int i = 0; i < 18; ++i) {
                if (!tables.containsName(FUSELAGE2D_U[i]))
                    return UNKNOWN;
            }

            //  Check the independents of each table.
            Float64Table tbl = tables.get("CLFAL");
            List<Text> indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT1"))
                return UNKNOWN;

            tbl = tables.get("CLFAH");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT2"))
                return UNKNOWN;

            tbl = tables.get("CLFBAL");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT3") || !indeps.get(1).toString().equals("FAOST1"))
                return UNKNOWN;

            tbl = tables.get("CLFBAH");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT3") || !indeps.get(1).toString().equals("FAOST2"))
                return UNKNOWN;

            tbl = tables.get("CDFAL");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT1"))
                return UNKNOWN;

            tbl = tables.get("CDFAH");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT2"))
                return UNKNOWN;

            tbl = tables.get("CMFAL");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT1"))
                return UNKNOWN;

            tbl = tables.get("CDFAH");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT2"))
                return UNKNOWN;

            tbl = tables.get("CDFBL");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOST1"))
                return UNKNOWN;

            tbl = tables.get("CDFBH");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOST2"))
                return UNKNOWN;

            tbl = tables.get("CMFBAL");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT3") || !indeps.get(1).toString().equals("FAOST1"))
                return UNKNOWN;

            tbl = tables.get("CMFBAH");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT3") || !indeps.get(1).toString().equals("FAOST2"))
                return UNKNOWN;

            tbl = tables.get("CYFBAL");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT3") || !indeps.get(1).toString().equals("FAOST1"))
                return UNKNOWN;

            tbl = tables.get("CYFBAH");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT3") || !indeps.get(1).toString().equals("FAOST2"))
                return UNKNOWN;

            tbl = tables.get("CNFBAL");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT3") || !indeps.get(1).toString().equals("FAOST1"))
                return UNKNOWN;

            tbl = tables.get("CNFBAH");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT3") || !indeps.get(1).toString().equals("FAOST2"))
                return UNKNOWN;

            tbl = tables.get("CRFBAL");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT3") || !indeps.get(1).toString().equals("FAOST1"))
                return UNKNOWN;

            tbl = tables.get("CRFBAH");
            indeps = tbl.getIndepNames();
            if (!indeps.get(0).toString().equals("FAOAT3") || !indeps.get(1).toString().equals("FAOST2"))
                return UNKNOWN;

            //  We must have a fuselage 2D uniform table set.
            type = FUSELAGE2D_UNIFORM;

        }

        return type;
    }

    /**
     * Write out a FlightLab SAV airfoil table with low/high alpha uniform increment
     * formatted multi-table stream and create a new table database.
     *
     * @param output A buffered output stream writer to the output file.
     * @param tabels The database of tables to be written out to the file.
     */
    private static void writeHighLowAirfoil(BufferedWriter output, FTableList tables) throws IOException {

        //  Create an exponential number format for the numbers we are writing out.
        ExponentialFormat nf = new ExponentialFormat();
        nf.setMaximumFractionDigits(16);
        nf.setMinimumFractionDigits(16);
        nf.setMaximumIntegerDigits(1);
        nf.setMinimumExponentDigits(2);

        //  Write out the comments to the header of the file.
        writeFileComments(output, tables);

        //  Write out the names of the tables.
        output.write("#Contents:");
        output.newLine();
        output.write("# AOATL");
        output.newLine();
        output.write("# AOATH");
        output.newLine();
        output.write("# MACHT");
        output.newLine();

        for (int i = 0; i < 6; ++i) {
            output.write("# " + AF_LOW_HIGH[i]);
            output.newLine();
        }

        //  Write out the info for the AOATL independent variable.
        Float64Table table = tables.get("CLL");
        Float64Vector bp = table.getBreakpoints(1);
        double min = bp.getValue(0);
        double max = bp.getValue(bp.getDimension() - 1);
        double step = bp.getValue(1) - min;
        output.write("AOATL 3 1 0");
        output.newLine();
        output.write(formatNumber(24, nf, min) + formatNumber(25, nf, max) + formatNumber(25, nf, step));
        output.newLine();
        output.newLine();

        //  Write out the info for the AOATH independent variable.
        table = tables.get("CLH");
        bp = table.getBreakpoints(0);
        min = bp.getValue(0);
        max = bp.getValue(bp.getDimension() - 1);
        step = bp.getValue(1) - min;
        output.write("AOATH 3 1 0");
        output.newLine();
        output.write(formatNumber(24, nf, min) + formatNumber(25, nf, max) + formatNumber(25, nf, step));
        output.newLine();
        output.newLine();

        //  Write out the info for the MACHT independent variable.
        table = tables.get("CLL");
        bp = table.getBreakpoints(0);
        min = bp.getValue(0);
        max = bp.getValue(bp.getDimension() - 1);
        step = bp.getValue(1) - min;
        output.write("MACHT 3 1 0");
        output.newLine();
        output.write(formatNumber(24, nf, min) + formatNumber(25, nf, max) + formatNumber(25, nf, step));
        output.newLine();
        output.newLine();

        //  Write out the CLL table.
        table = tables.get("CLL");
        write2DTable(table, nf, output);

        //  Write out the CDL table.
        table = tables.get("CDL");
        write2DTable(table, nf, output);

        //  Write out the CML table.
        table = tables.get("CML");
        write2DTable(table, nf, output);

        //  Write out the CLH table.
        table = tables.get("CLH");
        write1DTable(table, nf, output);

        //  Write out the CDH table.
        table = tables.get("CDH");
        write1DTable(table, nf, output);

        //  Write out the CMH table.
        table = tables.get("CMH");
        write1DTable(table, nf, output);

    }

    /**
     * Write out a FlightLab SAV fuselage 2D uniform increment formatted multi-table
     * stream and create a new table database.
     *
     * @param output A buffered output stream writer to the output file.
     * @param tabels The database of tables to be written out to the file.
     */
    private static void writeFuselage2DUniform(BufferedWriter output, FTableList tables) throws IOException {

        //  Create an exponential number format for the numbers we are writing out.
        ExponentialFormat nf = new ExponentialFormat();
        nf.setMaximumFractionDigits(16);
        nf.setMinimumFractionDigits(16);
        nf.setMaximumIntegerDigits(1);
        nf.setMinimumExponentDigits(2);

        //  Write out the comments to the header of the file.
        writeFileComments(output, tables);

        //  Write out the names of the tables.
        output.write("#Contents:");
        output.newLine();
        output.write("# FAOAT1");
        output.newLine();
        output.write("# FAOAT2");
        output.newLine();
        output.write("# FAOAT3");
        output.newLine();
        output.write("# FAOST1");
        output.newLine();
        output.write("# FAOST2");
        output.newLine();

        for (int i = 0; i < 18; ++i) {
            output.write("# " + FUSELAGE2D_U[i]);
            output.newLine();
        }

        //  Write out the info for the FAOAT1 independent variable.
        Float64Table table = tables.get("CLFAL");
        Float64Vector bp = table.getBreakpoints(0);
        double min = bp.getValue(0);
        double max = bp.getValue(bp.getDimension() - 1);
        double step = bp.getValue(1) - min;
        output.write("FAOAT1 3 1 0");
        output.newLine();
        output.write(formatNumber(24, nf, min) + formatNumber(25, nf, max) + formatNumber(25, nf, step));
        output.newLine();
        output.newLine();

        //  Write out the info for the FAOAT2 independent variable.
        table = tables.get("CLFAH");
        bp = table.getBreakpoints(0);
        min = bp.getValue(0);
        max = bp.getValue(bp.getDimension() - 1);
        step = bp.getValue(1) - min;
        output.write("FAOAT2 3 1 0");
        output.newLine();
        output.write(formatNumber(24, nf, min) + formatNumber(25, nf, max) + formatNumber(25, nf, step));
        output.newLine();
        output.newLine();

        //  Write out the info for the FAOAT3 independent variable.
        table = tables.get("CLFBAH");
        bp = table.getBreakpoints(0);
        min = bp.getValue(0);
        max = bp.getValue(bp.getDimension() - 1);
        step = bp.getValue(1) - min;
        output.write("FAOAT3 3 1 0");
        output.newLine();
        output.write(formatNumber(24, nf, min) + formatNumber(25, nf, max) + formatNumber(25, nf, step));
        output.newLine();
        output.newLine();

        //  Write out the info for the FAOST1 independent variable.
        table = tables.get("CDFBL");
        bp = table.getBreakpoints(0);
        min = bp.getValue(0);
        max = bp.getValue(bp.getDimension() - 1);
        step = bp.getValue(1) - min;
        output.write("FAOST1 3 1 0");
        output.newLine();
        output.write(formatNumber(24, nf, min) + formatNumber(25, nf, max) + formatNumber(25, nf, step));
        output.newLine();
        output.newLine();

        //  Write out the info for the FAOST2 independent variable.
        table = tables.get("CDFBH");
        bp = table.getBreakpoints(0);
        min = bp.getValue(0);
        max = bp.getValue(bp.getDimension() - 1);
        step = bp.getValue(1) - min;
        output.write("FAOST2 3 1 0");
        output.newLine();
        output.write(formatNumber(24, nf, min) + formatNumber(25, nf, max) + formatNumber(25, nf, step));
        output.newLine();
        output.newLine();

        //  Write out the CDFAH table.
        table = tables.get("CDFAH");
        write1DTable(table, nf, output);

        //  Write out the CDFAL table.
        table = tables.get("CDFAL");
        write1DTable(table, nf, output);

        //  Write out the CDFBH table.
        table = tables.get("CDFBH");
        write1DTable(table, nf, output);

        //  Write out the CDFBL table.
        table = tables.get("CDFBL");
        write1DTable(table, nf, output);

        //  Write out the CLFAH table.
        table = tables.get("CLFAH");
        write1DTable(table, nf, output);

        //  Write out the CLFAL table.
        table = tables.get("CLFAL");
        write1DTable(table, nf, output);

        //  Write out the CLFBAH table.
        table = tables.get("CLFBAH");
        write2DTable(table, nf, output);

        //  Write out the CLFBAL table.
        table = tables.get("CLFBAL");
        write2DTable(table, nf, output);

        //  Write out the CMFAL table.
        table = tables.get("CMFAL");
        write1DTable(table, nf, output);

        //  Write out the CMFAH table.
        table = tables.get("CMFAH");
        write1DTable(table, nf, output);

        //  Write out the CMFBAH table.
        table = tables.get("CMFBAH");
        write2DTable(table, nf, output);

        //  Write out the CMFBAL table.
        table = tables.get("CMFBAL");
        write2DTable(table, nf, output);

        //  Write out the CNFBAH table.
        table = tables.get("CNFBAH");
        write2DTable(table, nf, output);

        //  Write out the CNFBAL table.
        table = tables.get("CNFBAL");
        write2DTable(table, nf, output);

        //  Write out the CRFBAH table.
        table = tables.get("CRFBAH");
        write2DTable(table, nf, output);

        //  Write out the CRFBAL table.
        table = tables.get("CRFBAL");
        write2DTable(table, nf, output);

        //  Write out the CYFBAH table.
        table = tables.get("CYFBAH");
        write2DTable(table, nf, output);

        //  Write out the CYFBAL table.
        table = tables.get("CYFBAL");
        write2DTable(table, nf, output);

    }

    /**
     * Write out any database notes as comments to the header of the file.
     *
     * @param output A writer used to output the data.
     * @param tables The table database we are writing out.
     */
    private static void writeFileComments(BufferedWriter output, FTableList tables) throws IOException {

        // Write out the table notes as the header of the file.
        int numNotes = tables.numberOfNotes();
        for (int i = 0; i < numNotes; ++i) {
            Text note = tables.getNote(i);
            output.write("# ");
            note.print(output);
            output.newLine();
        }

        // Write out some generic table comments into the header lines.
        output.write("#.dat");
        output.newLine();
        output.write("# Date:, ");
        Date theDate = new Date();
        output.write(DateFormat.getDateInstance(DateFormat.SHORT).format(theDate));
        output.write(", ");
        output.write(DateFormat.getTimeInstance().format(theDate));
        output.newLine();
        output.write("# File created by: " + System.getProperty("user.name"));
        output.newLine();

    }

    /**
     * Method that writes out a single 1D table to a file.
     *
     * @param table  The table to be written out (must be 1D).
     * @param nf     The NumberFormat object used to format the output.
     * @param output A writer that writes out the data to a file.
     */
    private static void write1DTable(Float64Table table, NumberFormat nf, BufferedWriter output) throws IOException {

        String tblName = table.getName().toString();
        int numBP = table.getNumBreakpoints(0);
        output.write(tblName + " " + numBP + " 1 0");
        output.newLine();

        int count = 0;
        for (int j = 0; j < numBP; ++j) {
            double value = table.getValue(j);
            output.write(formatNumber(24, nf, value));
            ++count;
            if (count == 3) {
                count = 0;
                output.newLine();
            } else
                output.write(" ");
        }
        output.newLine();
//      if (count != 0)
//          output.newLine();
    }

    /**
     * Method that writes out a single 2D table to a file.
     *
     * @param table  The table to be written out (must be 2D).
     * @param nf     The NumberFormat object used to format the output.
     * @param output A writer that writes out the data to a file.
     */
    private static void write2DTable(Float64Table table, NumberFormat nf, BufferedWriter output) throws IOException {

        String tblName = table.getName().toString();
        int numBP1 = table.getNumBreakpoints(1);
        int numBP2 = table.getNumBreakpoints(0);
        output.write(tblName + " " + numBP1 + " " + numBP2 + " 0");
        output.newLine();

        int count = 0;
        FastTable<Integer64> pos = FastTable.newInstance();
        pos.add(Integer64.ZERO);
        pos.add(Integer64.ZERO);
        for (int i = 0; i < numBP2; ++i) {
            pos.set(0, Integer64.valueOf(i));           //  pos[0] = i;
            for (int j = 0; j < numBP1; ++j) {
                pos.set(1, Integer64.valueOf(j));       //  pos[1] = j;
                double value = table.getValue(pos);
                output.write(formatNumber(24, nf, value));
                ++count;
                if (count == 3) {
                    count = 0;
                    output.newLine();
                } else
                    output.write(" ");
            }
        }
        output.newLine();
    }

    /**
     * Format a number for output to a table file. This method formats the given number
     * using the supplied NumberFormat object. It then adds spaces to the start of the
     * formatted number until the string reaches the specified length.
     *
     * @param size   The overall length of the formatted number including the decimal
     *               point, minus sign, "E" notation, etc.
     * @param nf     The NumberFormat to use when formatting this number.
     * @param number The number to be formatted.
     */
    private static String formatNumber(int size, NumberFormat nf, double number) {
        //  Remove a bunch of extraneous digits from the double's converted to doubles
        //  by rounding them properly.
        number = MathTools.roundToPlace(number, -8);

        StringBuffer buffer = new StringBuffer(nf.format(number));
        TRUtils.addSpaces(size, buffer);
        return buffer.toString();
    }

}
