/*
 * Float64Table -- A table of functional values of arbitrary dimension.
 *
 * Copyright (C) 1999-2024, by Joseph A. Huwaldt.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *   
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.tables;

import jahuwaldt.tools.math.MathTools;
import java.text.MessageFormat;
import static java.util.Objects.isNull;
import static java.util.Objects.requireNonNull;
import java.util.*;
import javolution.context.ArrayFactory;
import javolution.context.ObjectFactory;
import javolution.lang.Realtime;
import javolution.text.Text;
import javolution.text.TextBuilder;
import javolution.util.FastMap;
import javolution.util.FastTable;
import javolution.xml.XMLFormat;
import javolution.xml.XMLSerializable;
import javolution.xml.stream.XMLStreamException;
import org.jscience.mathematics.number.Float64;
import org.jscience.mathematics.number.Integer64;
import org.jscience.mathematics.vector.Float64Vector;

/**
 * A table of functional values of arbitrary dimension. This class is used to contain a
 * table of functional data <code>value = fn(x,y,z,...,n)</code>. You can then do lookups
 * into this table to get values for arbitrary values of x,y,z, etc. This includes the
 * ability to linearly interpolate between breakpoints.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: July 17, 1999
 * @version January 21, 2024
 */
public final class Float64Table implements Realtime, XMLSerializable {

    private static final long serialVersionUID = -934363198544701471L;

    /**
     * The resource bundle for this package.
     */
    static final ResourceBundle RESOURCES
            = ResourceBundle.getBundle("jahuwaldt.js.tables.Float64Table", java.util.Locale.getDefault());

    /**
     * The name of this data table.
     */
    private CharSequence _tblName = Text.EMPTY;

    /**
     * The name of each independent variable (breakpoint table).
     */
    private FastTable<Text> _independentNames;

    /**
     * The breakpoints (independent variables) for each dimension of this table.
     */
    private FastTable<Float64Vector> _breakpoints;

    /**
     * The n-dimensional dependent data table.
     */
    private Float64NDimArray _depData;

    /**
     * The database notes stored as one note (CharSequence) per entry.
     */
    private FastTable<Text> _notes;

    //  Reference data for this element.
    private FastMap<Object, Object> _metaData = null;

    // The following are used by the recursive iteration routine in the "lookup()" method.
    private transient List<Float64> _indep;
    private transient double[][] _indValues;
    private transient int[][] _indPos;
    private transient FastTable<Integer64> _pos;

    /**
     * Stack of HI/LO independent array indexes used to indicate which value to read from
     * the n-dimensional data table.
     */
    private transient FastTable<Integer64> _idxStack;
    private static final Integer64 LO = Integer64.ZERO;
    private static final Integer64 HI = Integer64.ONE;
    private static final int LO_VALUE = 0;
    private static final int HI_VALUE = 1;

    //-----------------------------------------------------------------------------
    /**
     * Returns a <code>Float64Table</code> object with the given name and an arbitrary
     * number of dimensions. The array will be filled with zeros. Breakpoints MUST be
     * stored in ascending order!
     *
     * @param name       The name of the new table.
     * @param indepNames A list of names for the independent variables.
     * @param brkpoints  A list of breakpoint vectors for each independent variable.
     * @param dData      An n-dimensional array of dependent data for this table. Pass
     *                   null to have an empty table created.
     * @return A Float64Table object initially filled with zeros.
      */
    public static Float64Table valueOf(CharSequence name, List<? extends CharSequence> indepNames,
            List<Float64Vector> brkpoints, Float64NDimArray dData) {
        String paramNullErr = RESOURCES.getString("paramNullErr");
        requireNonNull(name, MessageFormat.format(paramNullErr, "name"));
        requireNonNull(indepNames, MessageFormat.format(paramNullErr, "indepNames"));
        requireNonNull(brkpoints, MessageFormat.format(paramNullErr, "brkpoints"));
        
        // Check for consistent inputs.
        int numDims = indepNames.size();
        if (brkpoints.size() != numDims
                || (!isNull(dData) && dData.getDimensions() != numDims)) {
            throw new ArrayIndexOutOfBoundsException(RESOURCES.getString("f64tIncDimensionsErr"));
        }

        //  Get an instance of a table.
        Float64Table tbl = Float64Table.newInstance(name, indepNames);

        // Fill in the breakpoints.
        tbl._breakpoints.addAll(brkpoints);

        if (isNull(dData)) {
            //  Create an nD array full of zero values.
            List<Integer64> elemPerDim = tbl.getNumBreakpoints();
            tbl._depData = Float64NDimArray.newInstance(elemPerDim);
            FastTable.recycle((FastTable<?>)elemPerDim);

        } else
            tbl._depData = dData.copy();

        return tbl;
    }

    /**
     * Returns a <code>Float64Table</code> object with the given name and an arbitrary
     * number of dimensions. The array will be filled with zeros. Breakpoints MUST be
     * stored in ascending order! This is a convenience method for translating simple Java
     * arrays into a table.
     *
     * @param name       The name of the new table.
     * @param indepNames A array of names for the independent variables.
     * @param brkpoints  A array of breakpoint arrays for each independent variable.
     * @param dData      An n-dimensional array of dependent data for this table. Pass
     *                   null to have an empty table created.
     * @return A Float64Table object initially filled with zeros.
     * @throws ArrayIndexOutOfBoundsException
     */
    public static Float64Table valueOf(String name, String[] indepNames, double[][] brkpoints, Float64NDimArray dData) {
        String paramNullErr = RESOURCES.getString("paramNullErr");
        requireNonNull(name, MessageFormat.format(paramNullErr, "name"));
        requireNonNull(indepNames, MessageFormat.format(paramNullErr, "indepNames"));
        requireNonNull(brkpoints, MessageFormat.format(paramNullErr, "brkpoints"));
        
        //  Convert the Java arrays into lists.
        FastTable<String> indepNamesList = FastTable.newInstance();
        indepNamesList.addAll(Arrays.asList(indepNames));

        FastTable<Float64Vector> brkpointsList = FastTable.newInstance();
        for (double[] data : brkpoints)
            brkpointsList.add(Float64Vector.valueOf(data));

        Float64Table table = valueOf(name, indepNamesList, brkpointsList, dData);

        //  Recycle tables.
        FastTable.recycle(indepNamesList);
        FastTable.recycle(brkpointsList);

        return table;
    }

    /**
     * Returns a 1D <code>Float64Table</code> object using the given vectors as
     * independent and dependent variables. Breakpoints MUST be sorted in ascending order!
     *
     * @param name      The name to be given to this table.
     * @param indepName The name of the independent variable.
     * @param indep     The vector of independent variable breakpoints.
     * @param dependent The vector of dependent variable values. May pass null to create
     *                  an table filled with zero values.
     * @return A Float64Table object using the given independent and dependent arrays of
     *         data.
      */
    public static Float64Table valueOf(CharSequence name, CharSequence indepName, Float64Vector indep, Float64Vector dependent) {
        String paramNullErr = RESOURCES.getString("paramNullErr");
        requireNonNull(name, MessageFormat.format(paramNullErr, "name"));
        requireNonNull(indepName, MessageFormat.format(paramNullErr, "indepName"));
        requireNonNull(indep, MessageFormat.format(paramNullErr, "indep"));
        
        int length = indep.getDimension();
        if (!isNull(dependent) && dependent.getDimension() != length)
            throw new ArrayIndexOutOfBoundsException(RESOURCES.getString("f64tIncDimensionsErr"));

        //  Get an instance of a table.
        Float64Table tbl = Float64Table.newInstance(name, indepName);

        // Fill in the breakpoints.
        tbl._breakpoints.add(indep);

        if (isNull(dependent)) {
            //  Create a vector of zeros.
            FastTable<Float64> data = FastTable.newInstance();
            for (int i = 0; i < length; ++i)
                data.add(Float64.ZERO);
            dependent = Float64Vector.valueOf(data);
            FastTable.recycle(data);
        }

        tbl._depData = Float64NDimArray.valueOf(dependent);

        return tbl;
    }

    /**
     * Returns a 1D <code>Float64Table</code> object using the given Java arrays as
     * independent and dependent variables. Breakpoints MUST be sorted in ascending order!
     * This is a convenience method for translating simple Java arrays into a table.
     *
     * @param name      The name to be given to this table.
     * @param indepName The name of the independent variable.
     * @param indep     The array of independent variable breakpoints.
     * @param dependent The array of dependent variable values. May pass null to create an
     *                  table filled with zero values.
     * @return A Float64Table object using the given independent and dependent arrays of
     *         data.
      */
    public static Float64Table valueOf(String name, String indepName, double[] indep, double[] dependent) {
        String paramNullErr = RESOURCES.getString("paramNullErr");
        requireNonNull(name, MessageFormat.format(paramNullErr, "name"));
        requireNonNull(indepName, MessageFormat.format(paramNullErr, "indepName"));
        requireNonNull(indep, MessageFormat.format(paramNullErr, "indep"));
        
        //  Convert arrays to vectors.
        Float64Vector indepV = Float64Vector.valueOf(indep);
        Float64Vector dependentV = null;
        if (!isNull(dependent))
            dependentV = Float64Vector.valueOf(dependent);

        return valueOf(name, indepName, indepV, dependentV);
    }

    /**
     * Creates a 2D <code>Float64Table</code> object using the given vector as independent
     * variable and list of rows as dependent variable. Breakpoints MUST be sorted in
     * ascending order!
     *
     * @param name       The name to be given to this table.
     * @param indepNames List of names of independent variables.
     * @param indep1     The vector of 1st independent variable breakpoints. This is the
     *                   independent that cycles the slowest.
     * @param indep2     The vector of 2nd independent variable breakpoints. This is the
     *                   independent that cycles the fastest.
     * @param dependent  The list of rows of dependent variable values (rows of columns).
     *                   May pass null to create a table filled with zero values.
     * @return A Float64Table object using the given independent and dependent arrays of
     *         data.
     */
    public static Float64Table valueOf(CharSequence name, List<? extends CharSequence> indepNames, Float64Vector indep1,
            Float64Vector indep2, List<Float64Vector> dependent) {
        String paramNullErr = RESOURCES.getString("paramNullErr");
        requireNonNull(name, MessageFormat.format(paramNullErr, "name"));
        requireNonNull(indepNames, MessageFormat.format(paramNullErr, "indepNames"));
        requireNonNull(indep1, MessageFormat.format(paramNullErr, "indep1"));
        requireNonNull(indep2, MessageFormat.format(paramNullErr, "indep2"));
        
        int length1 = indep1.getDimension();
        int length2 = indep2.getDimension();
        if (!isNull(dependent) && (dependent.size() != length1
                || dependent.get(0).getDimension() != length2)) {
            throw new ArrayIndexOutOfBoundsException(RESOURCES.getString("f64tIncDimensionsErr"));
        }

        //  Get an instance of a table.
        Float64Table tbl = Float64Table.newInstance(name, indepNames);

        // Fill in the breakpoints.
        tbl._breakpoints.add(indep1);
        tbl._breakpoints.add(indep2);

        if (!isNull(dependent))
            tbl._depData = Float64NDimArray.valueOf(dependent);

        else {
            //  Create an nD array full of zero values.
            List<Integer64> elemPerDim = tbl.getNumBreakpoints();
            tbl._depData = Float64NDimArray.newInstance(elemPerDim);
            FastTable.recycle((FastTable<?>)elemPerDim);
        }

        return tbl;
    }

    /**
     * Creates a 2D <code>Float64Table</code> object using the given Java arrays as
     * independent and dependent variables. Breakpoints MUST be sorted in ascending order!
     * This is a convenience method for translating simple Java arrays into a table.
     *
     * @param name       The name to be given to this table.
     * @param indepNames List of names of independent variables.
     * @param indep1     The array of 1st independent variable breakpoints. This is the
     *                   independent that cycles the slowest.
     * @param indep2     The array of 2nd independent variable breakpoints. This is the
     *                   independent that cycles the fastest.
     * @param dependent  The array dependent variable values (rows of columns). May pass
     *                   null to create a table filled with zero values.
     * @return A Float64Table object using the given independent and dependent arrays of
     *         data.
     */
    public static Float64Table valueOf(String name, String[] indepNames, double[] indep1,
            double[] indep2, double[][] dependent) {
        String paramNullErr = RESOURCES.getString("paramNullErr");
        requireNonNull(name, MessageFormat.format(paramNullErr, "name"));
        requireNonNull(indepNames, MessageFormat.format(paramNullErr, "indepNames"));
        requireNonNull(indep1, MessageFormat.format(paramNullErr, "indep1"));
        requireNonNull(indep2, MessageFormat.format(paramNullErr, "indep2"));
        
        //  Convert the Java arrays into lists.
        FastTable<String> indepNamesList = FastTable.newInstance();
        indepNamesList.addAll(Arrays.asList((indepNames)));

        Float64Vector indep1V = Float64Vector.valueOf(indep1);
        Float64Vector indep2V = Float64Vector.valueOf(indep2);

        FastTable<Float64Vector> depData = null;
        if (!isNull(dependent)) {
            depData = FastTable.newInstance();
            for (double[] data : dependent) {
                depData.add(Float64Vector.valueOf(data));
            }
        }

        Float64Table table = valueOf(name, indepNamesList, indep1V, indep2V, depData);

        //  Recycle the tables.
        if (!isNull(depData))
            FastTable.recycle(depData);
        FastTable.recycle(indepNamesList);

        return table;
    }

    /**
     * Creates a 3D Float64Table object using the given lists of independent and dependent
     * variables. Breakpoints MUST be sorted in ascending order!
     *
     * @param name       The name to be given to this table.
     * @param indepNames List of names of independent variables.
     * @param indep1     The vector of 1st independent variable breakpoints. This is the
     *                   independent that cycles the slowest.
     * @param indep2     The vector of 2nd independent variable breakpoints.
     * @param indep3     The vector of 3rd independent variable breakpoints. This is the
     *                   independent that cycles the fastest.
     * @param dependent  The 3D array of dependent variable values (ranks of rows of
     *                   columns). May pass null to create a table filled with zero
     *                   values.
     * @return A Float64Table object using the given independent and dependent arrays of
     *         data.
     */
    public static Float64Table valueOf(CharSequence name, List<? extends CharSequence> indepNames, Float64Vector indep1,
            Float64Vector indep2, Float64Vector indep3, List<List<Float64Vector>> dependent) {
        String paramNullErr = RESOURCES.getString("paramNullErr");
        requireNonNull(name, MessageFormat.format(paramNullErr, "name"));
        requireNonNull(indepNames, MessageFormat.format(paramNullErr, "indepNames"));
        requireNonNull(indep1, MessageFormat.format(paramNullErr, "indep1"));
        requireNonNull(indep2, MessageFormat.format(paramNullErr, "indep2"));
        requireNonNull(indep3, MessageFormat.format(paramNullErr, "indep3"));

        int length1 = indep1.getDimension();
        int length2 = indep2.getDimension();
        int length3 = indep3.getDimension();
        if (!isNull(dependent) && (dependent.size() != length1 || dependent.get(0).size() != length2
                || dependent.get(0).get(0).getDimension() != length3)) {
            throw new ArrayIndexOutOfBoundsException(RESOURCES.getString("f64tIncDimensionsErr"));
        }

        //  Get an instance of a table.
        Float64Table tbl = Float64Table.newInstance(name, indepNames);

        // Fill in the breakpoints.
        tbl._breakpoints.add(indep1);
        tbl._breakpoints.add(indep2);
        tbl._breakpoints.add(indep3);

        if (!isNull(dependent))
            tbl._depData = Float64NDimArray.valueOf3D(dependent);

        else {
            //  Create an nD array full of zero values.
            List<Integer64> elemPerDim = tbl.getNumBreakpoints();
            tbl._depData = Float64NDimArray.newInstance(elemPerDim);
            FastTable.recycle((FastTable<?>)elemPerDim);
        }

        return tbl;
    }

    /**
     * Creates a 3D Float64Table object using the given Java arrays as independent and
     * dependent variables. Breakpoints MUST be sorted in ascending order! This is a
     * convenience method for translating simple Java arrays into a table.
     *
     * @param name       The name to be given to this table.
     * @param indepNames List of names of independent variables.
     * @param indep1     The array of 1st independent variable breakpoints. This is the
     *                   independent that cycles the slowest.
     * @param indep2     The array of 2nd independent variable breakpoints.
     * @param indep3     The array of 3rd independent variable breakpoints. This is the
     *                   independent that cycles the fastest.
     * @param dependent  The 3D array of dependent variable values (ranks of rows of
     *                   columns). May pass null to create a table filled with zero
     *                   values.
     * @return A Float64Table object using the given independent and dependent arrays of
     *         data.
     */
    @SuppressWarnings("rawtypes")
    public static Float64Table valueOf(String name, String[] indepNames, double[] indep1,
            double[] indep2, double[] indep3, double[][][] dependent) {
        String paramNullErr = RESOURCES.getString("paramNullErr");
        requireNonNull(name, MessageFormat.format(paramNullErr, "name"));
        requireNonNull(indepNames, MessageFormat.format(paramNullErr, "indepNames"));
        requireNonNull(indep1, MessageFormat.format(paramNullErr, "indep1"));
        requireNonNull(indep2, MessageFormat.format(paramNullErr, "indep2"));
        requireNonNull(indep3, MessageFormat.format(paramNullErr, "indep3"));

        //  Convert the Java arrays into lists.
        FastTable<String> indepNamesList = FastTable.newInstance();
        indepNamesList.addAll(Arrays.asList(indepNames));

        Float64Vector indep1V = Float64Vector.valueOf(indep1);
        Float64Vector indep2V = Float64Vector.valueOf(indep2);
        Float64Vector indep3V = Float64Vector.valueOf(indep3);

        FastTable<List<Float64Vector>> depData = null;
        if (!isNull(dependent)) {
            depData = FastTable.newInstance();
            for (double[][] row : dependent) {
                List<Float64Vector> rowList = FastTable.newInstance();
                depData.add(rowList);
                for (double[] column : row) {
                    rowList.add(Float64Vector.valueOf(column));
                }
            }
        }

        Float64Table table = valueOf(name, indepNamesList, indep1V, indep2V, indep3V, depData);

        //  Recycle the tables.
        if (depData != null) {
            for (List<Float64Vector> row : depData)
                FastTable.recycle((FastTable)(row));
            FastTable.recycle((FastTable)depData);
        }
        FastTable.recycle(indepNamesList);

        return table;
    }

    //-----------------------------------------------------------------------------
    /**
     * Returns the number of dimensions (the number of independent variables) contained in
     * this table.
     *
     * @return Returns the number of dimensions in this table.
     */
    public int getDimensions() {
        return _breakpoints.size();
    }

    /**
     * Returns the total number of elements in the dependent data for this table.
     *
     * @return Returns the total number of elements in the dependent data array for this
     *         table.
     */
    public int size() {
        return _depData.size();
    }

    /**
     * Sets this table name to the specified name.
     *
     * @param name The new name to give this table. May not be null.
     */
    public void setTableName(CharSequence name) {
        _tblName = Text.valueOf(requireNonNull(name, MessageFormat.format(RESOURCES.getString("paramNullErr"), "name")));
    }

    /**
     * Returns the table name.
     *
     * @return Returns the name of this table.
     */
    public CharSequence getName() {
        return _tblName;
    }

    /**
     * Sets the names of the independent variables to those in the given list.
     *
     * @param names The list of new names for the table independent variables.
     */
    private void setIndepNames(List<? extends CharSequence> names) {
        requireNonNull(names, MessageFormat.format(RESOURCES.getString("paramNullErr"), "names"));
        _independentNames.clear();
        for (CharSequence text : names) {
            if (!isNull(text))
                _independentNames.add(Text.valueOf(text));
        }
    }

    /**
     * Returns the names of the independent variables as an unmodifiable list.
     *
     * @return Returns an unmodifiable list of independent variable names.
     */
    public List<Text> getIndepNames() {
        return _independentNames.unmodifiable();
    }

    /**
     * Set the name of the specified independent variable.
     *
     * @param dim  The index of the independent variable to set the name of. A value of 0
     *             indicates the independent that varies the least, a value of
     *             getDimensions()-1 indicates the independent that varies the most.
     * @param name The new name to give the specified independent variable. Can be null
     * for no name.
     */
    public void setIndepName(int dim, CharSequence name) {
        _independentNames.set(dim, (name != null ? Text.valueOf(name) : Text.EMPTY));
    }

    /**
     * Returns the name of the specified independent variable.
     *
     * @param dim The index of the independent variable to return the name of. A value of
     *            0 indicates the independent that varies the least, a value of
     *            getDimensions()-1 indicates the independent that varies the most.
     * @return Returns the name of the specified independent variable.
     */
    public Text getIndepName(int dim) {
        return _independentNames.get(dim);
    }

    /**
     * Returns the number of breakpoints in each of the dimensions of this nD array as a
     * list of <code>Integer64</code> values.
     *
     * @return Returns a list of <code>Integer64</code> values representing the number of
     *         breakpoints in each dimension of this nD array.
     */
    public List<Integer64> getNumBreakpoints() {
        FastTable<Integer64> elemPerDim = FastTable.newInstance();
        int numDims = _breakpoints.size();
        for (int i = 0; i < numDims; ++i) {
            int dimLength = _breakpoints.get(i).getDimension();
            elemPerDim.add(Integer64.valueOf(dimLength));
        }
        return elemPerDim;
    }

    /**
     * Returns the number of breakpoints in a specified dimension.
     *
     * @param dim The index of the independent variable to return the number of
     *            breakpoints for. A value of 0 indicates the independent that varies the
     *            least, a value of getDimensions()-1 indicates the independent that
     *            varies the most.
     * @return Returns the number of breakpoints in a specified dimension.
     */
    public int getNumBreakpoints(int dim) {
        return _breakpoints.get(dim).getDimension();
    }

    /**
     * Returns the array of breakpoints corresponding to the specified dimension
     * (independent variable).
     *
     * @param dim The index of the independent variable to return the breakpoints for. A
     *            value of 0 indicates the independent that varies the least, a value of
     *            getDimensions()-1 indicates the independent that varies the most.
     * @return Returns a reference to the array of breakpoints for the specified
     *         dimension.
     */
    public Float64Vector getBreakpoints(int dim) {
        return _breakpoints.get(dim);
    }

    /**
     * Get the value of a particular table breakpoint.
     *
     * @param dim   The index of the independent variable or dimension to return. A value
     *              of 0 indicates the independent that varies the least, a value of
     *              getDimensions()-1 indicates the independent that varies the most.
     * @param index Index of the breakpoint to return in the specified dimension.
     * @return The value of a particular table breakpoint.
     * @returns The value of the specified breakpoint.
     */
    public Float64 getBreakpoint(int dim, int index) {
        return _breakpoints.get(dim).get(index);
    }

    /**
     * Get the value of a particular table breakpoint as a <code>double</code>.
     *
     * @param dim   The index of the independent variable or dimension to return. A value
     *              of 0 indicates the independent that varies the least, a value of
     *              getDimensions()-1 indicates the independent that varies the most.
     * @param index Index of the breakpoint to return in the specified dimension.
     * @return The value of a particular table breakpoint.
     * @returns The value of the specified breakpoint as a <code>double</code>.
     */
    public double getBreakpointValue(int dim, int index) {
        return _breakpoints.get(dim).getValue(index);
    }

    /**
     * Sets the vector of breakpoints corresponding to the specified dimension
     * (independent variable).
     *
     * @param dim            The index of the independent variable to set the breakpoints
     *                       for. A value of 0 indicates the independent that varies the
     *                       least, a value of getDimensions()-1 indicates the independent
     *                       that varies the most.
     * @param newBreakpoints An array of breakpoints for the specified dimension.
     */
    public void setBreakpoints(int dim, Float64Vector newBreakpoints) {
        int length = newBreakpoints.getDimension();
        if (_breakpoints.get(dim).getDimension() != length)
            throw new ArrayIndexOutOfBoundsException(RESOURCES.getString("f64tNumBreakPointsErr"));

        _breakpoints.set(dim, newBreakpoints);
    }

    /**
     * Set the value of a particular table breakpoint.
     *
     * @param dim   The index of the independent variable or dimension to change a
     *              breakpoint for. A value of 0 indicates the independent that varies the
     *              least, a value of getDimensions()-1 indicates the independent that
     *              varies the most.
     * @param index Index of the breakpoint to modify in the specified dimension.
     * @param value The new value of the specified breakpoint.
     */
    public void setBreakpoint(int dim, int index, Float64 value) {
        if (value.equals(getBreakpoint(dim, index)))
            return;

        //  First create a list of the contents of the breakpoints of the specified dimension.
        Float64Vector vector = _breakpoints.get(dim);
        FastTable<Float64> list = FastTable.newInstance();
        int numElements = vector.getDimension();
        for (int i = 0; i < numElements; ++i)
            list.add(vector.get(i));

        //  Modify the specified breakpoint.
        list.set(index, value);

        //  Turn back into a Float64Vector.
        vector = Float64Vector.valueOf(list);

        //  Store back in the breakpoints list.
        _breakpoints.set(dim, vector);

        //  Recycle the list.
        FastTable.recycle(list);
    }

    /**
     * Set the value of a particular table breakpoint.
     *
     * @param dim   The index of the independent variable or dimension to change a
     *              breakpoint for. A value of 0 indicates the independent that varies the
     *              least, a value of getDimensions()-1 indicates the independent that
     *              varies the most.
     * @param index Index of the breakpoint to modify in the specified dimension.
     * @param value The new value of the specified breakpoint.
     */
    public void setBreakpoint(int dim, int index, double value) {
        setBreakpoint(dim, index, Float64.valueOf(value));
    }

    /**
     * Set the value of a particular element in the table.
     *
     * @param position A list of indexes into the table of the value to be set. For
     *                 example, for a 3D table the position could be { 0, 3, 2 }. The 1st
     *                 index is for the independent that varies the least, the last for
     *                 the one that varies the most.
     * @param value    The value to be stored in the table.
     * @throws ArrayIndexOutOfBoundsException if the input position exceeds the table
     * dimensions.
     */
    public void set(List<Integer64> position, Float64 value) {
        requireNonNull(position, MessageFormat.format(RESOURCES.getString("paramNullErr"), "position"));
        requireNonNull(value, MessageFormat.format(RESOURCES.getString("paramNullErr"), "value"));
        _depData.set(position, value);
    }

    /**
     * Set the value of a particular element in the table.
     *
     * @param position A list of indexes into the table of the value to be set. For
     *                 example, for a 3D table the position could be { 0, 3, 2 }. The 1st
     *                 index is for the independent that varies the least, the last for
     *                 the one that varies the most.
     * @param value    The value to be stored in the table.
     * @throws ArrayIndexOutOfBoundsException if the input position exceeds the table
     * dimensions.
     */
    public void set(List<Integer64> position, double value) {
        _depData.set(requireNonNull(position, MessageFormat.format(RESOURCES.getString("paramNullErr"), "position")), value);
    }

    /**
     * Get the value of a particular element in the table.
     *
     * @param position A list of indexes into the table of the value to be retrieved. For
     *                 example, for a 3D table the position could be { 0, 3, 2 }. The 1st
     *                 index is for the independent that varies the least, the last for
     *                 the one that varies the most.
     * @return Returns the specified value in the table.
     * @throws ArrayIndexOutOfBoundsException if the input position exceeds the table
     * dimensions.
     */
    public Float64 get(List<Integer64> position) {
        return _depData.get(requireNonNull(position, MessageFormat.format(RESOURCES.getString("paramNullErr"), "position")));
    }

    /**
     * Get the value of a particular element in the table as a <code>double</code>.
     *
     * @param position A list of indexes into the table of the value to be retrieved. For
     *                 example, for a 3D table the position could be { 0, 3, 2 }. The 1st
     *                 index is for the independent that varies the least, the last for
     *                 the one that varies the most.
     * @return Returns the specified value in the table as a <code>double</code>.
     * @throws ArrayIndexOutOfBoundsException if the input position exceeds the table
     * dimensions.
     */
    public double getValue(List<Integer64> position) {
        return _depData.getValue(requireNonNull(position, MessageFormat.format(RESOURCES.getString("paramNullErr"), "position")));
    }

    /**
     * Set the value of a particular element in a 1-dimensional table. This is a special
     * version of set() that works only with a 1-dimensional table (1 independent
     * variable). This version is provided for increased performance in this special (but
     * common) case.
     *
     * @param position The index into the 1D table of the value to be retrieved.
     * @param value    The value to be stored in the table.
     * @throws ArrayIndexOutOfBoundsException if this table is not 1-dimensional or if
     * position exceeds the limits of the table.
     */
    public void set(int position, Float64 value) {
        _depData.set(position, requireNonNull(value, MessageFormat.format(RESOURCES.getString("paramNullErr"), "value")));
    }

    /**
     * Set the value of a particular element in a 1-dimensional table. This is a special
     * version of set() that works only with a 1-dimensional table (1 independent
     * variable). This version is provided for increased performance in this special (but
     * common) case.
     *
     * @param position The index into the 1D table of the value to be retrieved.
     * @param value    The value to be stored in the table.
     * @throws ArrayIndexOutOfBoundsException if this table is not 1-dimensional or if
     * position exceeds the limits of the table.
     */
    public void set(int position, double value) {
        _depData.set(position, value);
    }

    /**
     * Get the value of a particular element in a 1-dimensional table. This is a special
     * version of get() that works only with a 1-dimensional table (1 independent
     * variable). This version is provided for increased performance in this special (but
     * common) case.
     *
     * @param position The index into the 1D table of the value to be retrieved.
     * @return Returns the specified value in the table.
     * @throws ArrayIndexOutOfBoundsException if this table is not 1-dimensional or if
     * position exceeds the limits of the table.
     */
    public Float64 get(int position) {
        return _depData.get(position);
    }

    /**
     * Get the value of a particular element in a 1-dimensional table and return as a
     * <code>double</code>. This is a special version of getValue() that works only with a
     * 1-dimensional table (1 independent variable). This version is provided for
     * increased performance in this special (but common) case.
     *
     * @param position The index into the 1D table of the value to be retrieved.
     * @return The specified value in the table as a <code>double</code>.
     * @throws ArrayIndexOutOfBoundsException if this table is not 1-dimensional or if
     * position exceeds the limits of the table.
     */
    public double getValue(int position) {
        return _depData.getValue(position);
    }

    /**
     * Return a copy of all the dependent data in the table.
     *
     * @return Returns a copy of all the dependent data in the table as a
     *         <code>Float64NDimArray</code> object.
     */
    public Float64NDimArray getAll() {
        return _depData.copy();
    }

    /**
     * <p>
     * Look up and return a value in the table. Linear interpolation is used to find
     * values between the breakpoints contained in the table.
     * </p>
     * <p>
     * WARNING: If the independent values supplied exceed the table limits, the value
     * returned will be extrapolated from the last two entries in the table.
     * </p>
     *
     * @param independents The list of <code>Float64</code> n-dimensional independent
     *                     variables to use for the table look up.
     * @return Returns the interpolated value of the table at the given independent
     *         variables.
     */
    public Float64 lookup(List<Float64> independents) {
        double value = lookupValue(independents);
        return Float64.valueOf(value);
    }

    /**
     * <p>
     * Look up a value in the table and return as a <code>double</code>. Linear
     * interpolation is used to find values between the breakpoints contained in the
     * table.
     * </p>
     * <p>
     * WARNING: If the independent values supplied exceed the table limits, the value
     * returned will be extrapolated from the last two entries in the table.
     * </p>
     *
     * @param independents The list of <code>Float64</code> n-dimensional independent
     *                     variables to use for the table look up.
     * @return Returns the interpolated value of the table at the given independent
     *         variables as a <code>double</code>.
     */
    public double lookupValue(List<Float64> independents) {
        int numDims = getDimensions();
        if (isNull(independents) || independents.size() != numDims)
            throw new ArrayIndexOutOfBoundsException(RESOURCES.getString("f64tNumIndepErr"));

        _indep = independents;

        // If 1st time through, allocate memory used by temporary arrays in the recursive
        // interpolation algorithm.
        if (_indValues == null) {
            _indValues = DOUBLES_2D_FACTORY.array(numDims);
            _indPos = INTS_2D_FACTORY.array(numDims);
            _pos = FastTable.newInstance();
            _pos.setSize(numDims);
            _idxStack = FastTable.newInstance();
            _idxStack.setSize(numDims + 1);
        }

        // Determine indexes of table points surrounding the independents
        // that the user has supplied.
        for (int i = 0; i < numDims; ++i) {
            Float64Vector bpVector = _breakpoints.get(i);
            int idx = findHi(independents.get(i).doubleValue(), bpVector);
            double value = bpVector.getValue(idx);
            _indValues[HI_VALUE][i] = value;
            _indPos[HI_VALUE][i] = idx;
            --idx;
            if (idx < 0) {
                _indValues[LO_VALUE][i] = value;
                _indPos[LO_VALUE][i] = 0;
            } else {
                _indValues[LO_VALUE][i] = bpVector.getValue(idx);
                _indPos[LO_VALUE][i] = idx;
            }
        }

        // Recursively interpolate the functional value.
        double value = rInterp(0);
        _indep = null;      //  Let go of the input list of independents.

        return value;
    }

    /**
     * <p>
     * Look up a value in the table and return as a <code>double</code>. Linear
     * interpolation is used to find values between the breakpoints contained in the
     * table.
     * </p>
     * <p>
     * WARNING: If the independent values supplied exceed the table limits, the value
     * returned will be extrapolated from the last two entries in the table.
     * </p>
     *
     * @param independents The array of double n-dimensional independent variables to use
     *                     for the table look up.
     * @return Returns the interpolated value of the table at the given independent
     *         variables as a <code>double</code>.
     */
    public double lookupValue(double... independents) {
        requireNonNull(independents, MessageFormat.format(RESOURCES.getString("paramNullErr"), "independents"));
        FastTable<Float64> indep = FastTable.newInstance();
        for (double value : independents)
            indep.add(Float64.valueOf(value));

        double value = lookupValue(indep);

        FastTable.recycle(indep);

        return value;
    }

    /**
     * <p>
     * Look up a value in a 1D table. Linear interpolation is used to find value between
     * the breakpoints contained in the table. This is a special version of lookup() that
     * works only with a 1-dimensional table (1 independent variable). This version is
     * provided for increased performance in this special (but common) case.
     * </p>
     * <p>
     * WARNING: If the independent value supplied exceeds the table limits, the value
     * returned will be extrapolated from the table.
     * </p>
     *
     * @param independent The independent variable to use for the table look up.
     * @return Returns the interpolated value of the 1D table at the given independent
     *         variable.
     */
    public Float64 lookup(Float64 independent) {
        requireNonNull(independent, MessageFormat.format(RESOURCES.getString("paramNullErr"), "independent"));
        int numDims = getDimensions();
        if (numDims != 1)
            throw new ArrayIndexOutOfBoundsException(RESOURCES.getString("f64tNumIndepErr"));

        // Determine indexes of the table points surrounding the independent
        // that the user has supplied.
        Float64Vector breakpoints = _breakpoints.get(0);
        double indepValue = independent.doubleValue();
        int indPosHI = findHi(indepValue, breakpoints);
        int indPosLO = indPosHI - 1;
        if (indPosLO < 0)
            indPosLO = 0;

        // Extract the independent values from the breakpoint array.
        double indValHI = breakpoints.getValue(indPosHI);
        double indValLO = breakpoints.getValue(indPosLO);

        // Extract the dependent values from the dependent data array.
        double depValHI = _depData.getValue(indPosHI);
        double depValLO = _depData.getValue(indPosLO);

        // Do the linear interpolation.
        double value = interp(indValLO, depValLO, indValHI, depValHI, indepValue);

        return Float64.valueOf(value);
    }

    /**
     * <p>
     * Look up a value in a 1D table. Linear interpolation is used to find value between
     * the breakpoints contained in the table. This is a special version of lookup() that
     * works only with a 1-dimensional table (1 independent variable). This version is
     * provided for increased performance in this special (but common) case.
     * </p>
     * <p>
     * WARNING: If the independent value supplied exceeds the table limits, the value
     * returned will be extrapolated from the table.
     * </p>
     *
     * @param independent The independent variable to use for the table look up.
     * @return Returns the interpolated value of the 1D table at the given independent
     *         variable.
     */
    public double lookupValue(double independent) {
        int numDims = getDimensions();
        if (numDims != 1)
            throw new ArrayIndexOutOfBoundsException(RESOURCES.getString("f64tNumIndepErr"));

        // Determine indexes of the table points surrounding the independent
        // that the user has supplied.
        Float64Vector breakpoints = _breakpoints.get(0);
        int indPosHI = findHi(independent, breakpoints);
        int indPosLO = indPosHI - 1;
        if (indPosLO < 0)
            indPosLO = 0;

        // Extract the independent values from the breakpoint array.
        double indValHI = breakpoints.getValue(indPosHI);
        double indValLO = breakpoints.getValue(indPosLO);

        // Extract the dependent values from the dependent data array.
        double depValHI = _depData.getValue(indPosHI);
        double depValLO = _depData.getValue(indPosLO);

        // Do the linear interpolation.
        double value = interp(indValLO, depValLO, indValHI, depValHI, independent);

        return value;
    }

    /**
     * Linear interpolation routine. Returns the value linearly interpolated between the
     * given dependent values for a given independent value. If the independent value we
     * are interpolating with is outside the range of low and high independents, then the
     * return value is extrapolated.
     *
     * @param ind1  Low independent value.
     * @param dep1  Low dependent value.
     * @param ind2  High independent value.
     * @param dep2  High dependent value.
     * @param indep Independent value we want to interpolate for.
     * @return Returns the value resulting from linear interpolation.
     */
    public static final double interp(double ind1, double dep1, double ind2, double dep2, double indep) {
        return ((dep2 - dep1) / (ind2 - ind1) * (indep - ind1) + dep1);
    }

    /**
     * Returns the number of notes associated with this table.
     *
     * @return The number of notes associated with this table.
     */
    public int numberOfNotes() {
        return _notes.size();
    }

    /**
     * Add a new note to the list of notes associated with this table.
     *
     * @param note The new note to add to this table.
     */
    public void addNote(CharSequence note) {
         _notes.add(Text.valueOf(requireNonNull(note,MessageFormat.format(RESOURCES.getString("paramNullErr"), "note"))));

    }

    /**
     * Add all the notes in the specified list of notes to the list of notes associated
     * with this table.
     *
     * @param noteList The list of new notes to add to this table.
     */
    public void addAllNotes(List<? extends CharSequence> noteList) {
        requireNonNull(noteList, MessageFormat.format(RESOURCES.getString("paramNullErr"), "noteList"));
        for (CharSequence note : noteList)
            addNote(note);
    }

    /**
     * Return a specified note from the list of notes associated with this table.
     *
     * @param index The index of the note to retrieve.
     * @return The note at the specified index in the list of notes.
     */
    public Text getNote(int index) {
        return _notes.get(index);
    }

    /**
     * Return all the notes associated with this table as an unmodifiable list.
     *
     * @return All the notes associated with this table.
     */
    public List<Text> getAllNotes() {
        return _notes.unmodifiable();
    }

    /**
     * Remove a specified note from the list of notes associated with this table.
     *
     * @param index The index of the note to remove.
     * @return The removed note.
     */
    public Text removeNote(int index) {
        return _notes.remove(index);
    }

    /**
     * Removes all notes associated with this table.
     */
    public void clearNotes() {
        _notes.clear();
    }

    /**
     * Returns the text representation of this table by outputting the table name and the
     * names of it's independent variables.
     *
     * @return the text representation of this table.
     */
    @Override
    public Text toText() {
        TextBuilder tmp = TextBuilder.newInstance();
        tmp.append(getName());
        tmp.append('(');
        int numDims = getDimensions();
        for (int i = 0; i < numDims; ++i) {
            tmp.append(getIndepName(i));
            if (i < numDims - 1)
                tmp.append(",");
        }
        tmp.append(')');
        Text txt = tmp.toText();
        TextBuilder.recycle(tmp);
        return txt;
    }

    /**
     * Output the name of the table and the names of it's independent variables.
     *
     * @return Returns a String containing the name of the table and it's independent
     *         variables.
     */
    @Override
    public final String toString() {
        return toText().toString();
    }

    /**
     * Compares this <code>Float64Table</code> against the specified object for strict
     * equality (same name, independent labels, dimensions and same values).
     *
     * @param obj the object to compare with.
     * @return <code>true</code> if this table is identical to that table;
     *         <code>false</code> otherwise.
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if ((obj == null) || (obj.getClass() != this.getClass()))
            return false;

        Float64Table that = (Float64Table)obj;
        if (!this._tblName.equals(that._tblName))
            return false;
        if (!this._independentNames.equals(that._independentNames))
            return false;
        if (!this._breakpoints.equals(that._breakpoints))
            return false;
        if (!this._depData.equals(that._depData))
            return false;

        return this._notes.equals(that._notes);
    }

    /**
     * Returns the hash code for this <code>Float64Table</code>.
     *
     * @return the hash code value.
     */
    @Override
    public int hashCode() {
        int hash = 7;

        hash = hash * 31 + _tblName.hashCode();
        hash = hash * 31 + _independentNames.hashCode();
        hash = hash * 31 + _breakpoints.hashCode();
        hash = hash * 31 + _depData.hashCode();
        hash = hash * 31 + _notes.hashCode();

        return hash;
    }

    /**
     * Returns a copy of this table {@link javolution.context.AllocatorContext allocated}
     * by the calling thread (possibly on the stack).
     *
     * @return an identical and independent copy of this table.
     */
    public Float64Table copy() {
        return copyOf(this);
    }

    /**
     * Return any user defined object or meta-data associated with this table and the
     * specified key. If there is no meta-data with the specified key, then
     * <code>null</code> is returned.
     *
     * @param key the key whose associated value is to be returned or <code>null</code> if
     *            the specified key could not be found. May not be null.
     * @return Any user defined object or meta-data associated with this table and the
     *         specified key.
     */
    public Object getMetaData(Object key) {
        requireNonNull(key, MessageFormat.format(RESOURCES.getString("paramNullErr"), "key"));
        if (isNull(_metaData))
             return null;
        return _metaData.get(key);
    }

    /**
     * Returns a new Map containing all the user objects or meta-data associated with this
     * table.
     *
     * @return A Map of all the user objects or meta-data associated with this table.
     */
    public Map<Object, Object> getAllMetaData() {
        if (isNull(_metaData))
            return FastMap.newInstance();
        FastMap<Object, Object> map = FastMap.newInstance();
        map.putAll(_metaData);
        return map;
    }

    /**
     * Set the user defined object or meta-data associated with this table and the
     * specified key. This can be used to store any type of information with a table that
     * could be useful. The meta-data does not impact the function of the
     * <code>equals</code> or <code>hashCode</code> methods and the meta-data does not get
     * deep copied when the table is cloned.
     *
     * @param key   the key with which the specified value is to be associated. May not be
     *              null.
     * @param value the value to be associated with the specified key. May not be null.
     */
    public void putMetaData(Object key, Object value) {
        requireNonNull(key, MessageFormat.format(RESOURCES.getString("paramNullErr"), "key"));
        requireNonNull(value, MessageFormat.format(RESOURCES.getString("paramNullErr"), "value"));
        if (isNull(_metaData))
            _metaData = FastMap.newInstance();
        _metaData.put(key, value);
    }

    /**
     * Add all the user defined or meta-data in the supplied Map to this tables map of
     * meta-data.
     *
     * @param data The Map of meta-data to be added to this object's map of meta-data. May
     *             not be null.
     */
    public void putAllMetaData(Map data) {
        requireNonNull(data, MessageFormat.format(RESOURCES.getString("paramNullErr"), "data"));
        if (data.isEmpty())
            return;
        Set keys = data.keySet();
        for (Object key : keys)
            putMetaData(key, data.get(key));
    }

    /**
     * Removes the entry for the specified meta-data object key if present in the
     * meta-data associated with this table.
     *
     * @param key the key whose mapping is to be removed from the meta-data in this table.
     *            May not be null.
     */
    public void removeMetaData(Object key) {
        requireNonNull(key, MessageFormat.format(RESOURCES.getString("paramNullErr"), "key"));
        if (isNull(_metaData)) {
             return;
        }
        _metaData.remove(key);
    }

    /**
     * Recycles a <code>Float64Table</code> instance immediately (on the stack when
     * executing in a <code>StackContext</code>).
     *
     * @param instance The instance to be recycled.
     */
    public static void recycle(Float64Table instance) {
        FACTORY.recycle(requireNonNull(instance, MessageFormat.format(RESOURCES.getString("paramNullErr"), "instance")));
    }

    /**
     * Returns the index of the next breakpoint greater than the value passed in. If the
     * value is greater than the largest breakpoint, then the 2nd to the last breakpoint
     * index is returned. If the value is less than the smallest breakpoint, then the 2nd
     * index is returned.
     *
     * @param value  The value to find the next greater breakpoint than.
     * @param vector The vector of breakpoints to be searched.
     * @return Returns index of the next breakpoint greater than the value passed in unless
     * there is only a single breakpoint in which case 0 is returned.
     */
    private static int findHi(double value, Float64Vector vector) {

        int length = vector.getDimension() - 1;
        if (length == 0)
            return 0;
        int i;
        for (i = 1; i < length; ++i) {
            if (vector.getValue(i) > value)
                break;
        }

        return i;
    }

    /**
     * <p>
     * Recursive interpolation algorithm for n dimensional data tables. This routine is
     * called recursively until the highest dimension is reached. Interpolated values are
     * then passed back up and the tree is walked until the entire problem has been
     * solved.
     * </p>
     *
     * <p> Written by: Joseph A. Huwaldt, Date: July 17, 1999 </p>
     *
     * @param dim The current dimension (used to indicate when the highest dimension has
     *            been reached).
     * @return Returns the result of an interpolation at a particular dimension.
     */
    private double rInterp(int dim) {

        if (dim != _breakpoints.size()) {
            int ndim = dim;
            ++dim;

            if (MathTools.isApproxEqual(_indep.get(ndim).doubleValue(), _indValues[LO_VALUE][ndim])) {
                //  Independent variable is exactly the same as low breakpoint value at this dimension.
                //  So, return breakpoint value only (don't need to interpolate in this dimension).
                _idxStack.set(ndim, LO);
                return rInterp(dim);

            } else {
                // Retrieve the dependent values surrounding the independent value
                // at this dimension.
                _idxStack.set(ndim, LO);
                double depValLO = rInterp(dim);
                _idxStack.set(ndim, HI);
                double depValHI = rInterp(dim);

                // Find the breakpoint values surrounding this independent value
                // at this dimension.
                double indVal = _indep.get(ndim).doubleValue();
                double indValLO = _indValues[LO_VALUE][ndim];
                double indValHI = _indValues[HI_VALUE][ndim];

                // Do linear interpolation and return the dependent at this dimension.
                return interp(indValLO, depValLO, indValHI, depValHI, indVal);
            }

        } else {
            // dim == numDims,  we are at the greatest dimension in the table.
            // Determine the address of the needed data point in the table.
            for (int i = 0; i < dim; ++i) {
                int index = _indPos[_idxStack.get(i).intValue()][i];
                _pos.set(i, Integer64.valueOf(index));
            }

            // Return the data point in the table.
            return _depData.getValue(_pos);
        }
    }

    //-----------------------------------------------------------------------------
    /**
     * Holds the default XML representation. For example:
     * <pre>
     *    &lt;Float64Table name="table1"&gt;
     *        &lt;IndependentNames&gt;
     *            &lt;Text value="indep1" /&gt;
     *        &lt;/IndependentNames&gt;
     *        &lt;Breakpoints&gt;
     *            &lt;Float64Vector dimension="9"&gt;
     *                &lt;Float64 value="0.0" /&gt;
     *                &lt;Float64 value="1.0" /&gt;
     *                &lt;Float64 value="2.0" /&gt;
     *                &lt;Float64 value="3.0" /&gt;
     *                &lt;Float64 value="4.0" /&gt;
     *                &lt;Float64 value="5.0" /&gt;
     *                &lt;Float64 value="6.0" /&gt;
     *                &lt;Float64 value="7.0" /&gt;
     *                &lt;Float64 value="8.0" /&gt;
     *            &lt;/Float64Vector&gt;
     *        &lt;/Breakpoints&gt;
     *        &lt;DependentData&gt;
     *            &lt;Dimensions&gt;
     *                &lt;Integer64 value="9" /&gt;
     *            &lt;/Dimensions&gt;
     *            &lt;Data dimension="9"&gt;
     *                &lt;Float64 value="0.0" /&gt;
     *                &lt;Float64 value="10.0" /&gt;
     *                &lt;Float64 value="20.0" /&gt;
     *                &lt;Float64 value="30.0" /&gt;
     *                &lt;Float64 value="40.0" /&gt;
     *                &lt;Float64 value="50.0" /&gt;
     *                &lt;Float64 value="60.0" /&gt;
     *                &lt;Float64 value="70.0" /&gt;
     *                &lt;Float64 value="80.0" /&gt;
     *            &lt;/Data&gt;
     *        &lt;/DependentData&gt;
     *        &lt;Notes&gt;
     *            &lt;Text value="This is a sample table." /&gt;
     *        &lt;/Notes&gt;
     * </pre>
     */
    protected static final XMLFormat<Float64Table> XML = new XMLFormat<Float64Table>(Float64Table.class) {

        @SuppressWarnings("unchecked")
        @Override
        public Float64Table newInstance(Class<Float64Table> cls, InputElement xml) throws XMLStreamException {

            //  Read the data from the XML element.
            String tblName = xml.getAttribute("name", "Unknown");
            FastTable<Text> indepNames = xml.get("IndependentNames", FastTable.class);
            FastTable<Float64Vector> breakpoints = xml.get("Breakpoints", FastTable.class);
            Float64NDimArray depData = xml.get("DependentData", Float64NDimArray.class);
            FastTable<Text> notes = xml.get("Notes", FastTable.class);

            if (xml.hasNext()) {
                FastTable.recycle(indepNames);
                FastTable.recycle(breakpoints);
                FastTable.recycle(notes);
                Float64NDimArray.recycle(depData);
                throw new XMLStreamException("Too many elements");
            }

            //  Create the table.
            Float64Table T = Float64Table.valueOf(tblName, indepNames, breakpoints, depData);
            T.addAllNotes(notes);

            return T;
        }

        @Override
        public void read(InputElement xml, Float64Table obj) throws XMLStreamException {
            FastMap userData = xml.get("MetaData", FastMap.class);
            if (!isNull(userData)) {
                obj.putAllMetaData(userData);
                FastMap.recycle(userData);
            }
        }

        @Override
        public void write(Float64Table obj, OutputElement xml) throws XMLStreamException {

            xml.setAttribute("name", obj._tblName);
            xml.add(obj._independentNames, "IndependentNames", FastTable.class);
            xml.add(obj._breakpoints, "Breakpoints", FastTable.class);
            xml.add(obj._depData, "DependentData", Float64NDimArray.class);
            xml.add(obj._notes, "Notes", FastTable.class);

            FastMap userData = (FastMap)obj.getAllMetaData();
            if (!userData.isEmpty())
                xml.add(userData, "MetaData", FastMap.class);
        }
    };

    //////////////////////
    // Factory Creation //
    //////////////////////
    private static Float64Table newInstance(CharSequence name, CharSequence indepName) {
        Float64Table o = FACTORY.object();
        o._tblName = name;
        o._independentNames = FastTable.newInstance();
        o._independentNames.add(Text.valueOf(indepName));
        o._breakpoints = FastTable.newInstance();
        o._notes = FastTable.newInstance();
        return o;
    }

    private static Float64Table newInstance(CharSequence name, List<? extends CharSequence> indepNames) {
        Float64Table o = FACTORY.object();
        o._tblName = name;
        o._independentNames = FastTable.newInstance();
        o.setIndepNames(indepNames);
        o._breakpoints = FastTable.newInstance();
        o._notes = FastTable.newInstance();
        return o;
    }

    private static Float64Table copyOf(Float64Table original) {
        Float64Table o = FACTORY.object();
        o._tblName = original._tblName;
        o._independentNames = FastTable.newInstance();
        o._independentNames.addAll(original._independentNames);
        o._breakpoints = FastTable.newInstance();
        o._breakpoints.addAll(original._breakpoints);
        o._depData = original._depData.copy();
        o._notes = FastTable.newInstance();
        o._notes.addAll(original._notes);
        if (!isNull(original._metaData))
            o.putAllMetaData(original._metaData);
        return o;
    }

    private static final ObjectFactory<Float64Table> FACTORY = new ObjectFactory<Float64Table>() {
        @Override
        protected Float64Table create() {
            return new Float64Table();
        }

        @Override
        protected void cleanup(Float64Table instance) {

            //  First recycle the elements of this table.
            FastTable.recycle(instance._independentNames);
            instance._independentNames = null;
            FastTable.recycle(instance._breakpoints);
            instance._breakpoints = null;
            Float64NDimArray.recycle(instance._depData);
            instance._depData = null;
            FastTable.recycle(instance._notes);
            instance._notes = null;

            if (!isNull(instance._metaData)) {
                instance._metaData.clear();
                FastMap.recycle(instance._metaData);
                instance._metaData = null;
            }

            //  Recycle arrays used for table lookups (if they exist);
            if (!isNull(instance._indValues)) {
                DOUBLES_2D_FACTORY.recycle(instance._indValues);
                instance._indValues = null;
                INTS_2D_FACTORY.recycle(instance._indPos);
                instance._indPos = null;
                FastTable.recycle(instance._pos);
                instance._pos = null;
                FastTable.recycle(instance._idxStack);
                instance._idxStack = null;
            }
        }
    };

    private Float64Table() { }

    /**
     * A factory that returns a 2D double array with the 1st dimension of 2 and the 2nd
     * dimension variable.
     */
    private static final ArrayFactory<double[][]> DOUBLES_2D_FACTORY = new ArrayFactory<double[][]>() {
        @Override
        protected double[][] create(int size) {
            return new double[2][size];
        }
    };

    /**
     * A factory that returns a 2D double array with the 1st dimension of 2 and the 2nd
     * dimension variable.
     */
    private static final ArrayFactory<int[][]> INTS_2D_FACTORY = new ArrayFactory<int[][]>() {
        @Override
        protected int[][] create(int size) {
            return new int[2][size];
        }
    };

    /**
     * Tests the methods in this class.
     *
     * @param args Command-line arguments (ignored).
     */
    public static void main(String args[]) {
        System.out.println("Testing Float64Table:");

        FastTable<Integer64> pos = new FastTable<>();
        FastTable<Float64> indep = new FastTable<>();

        //  Test a 1D table.
        Float64Vector indep1 = Float64Vector.valueOf(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10);
        Float64Vector vector1 = Float64Vector.valueOf(0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100);
        Float64Table table1 = Float64Table.valueOf("table1", "indep1", indep1, vector1);
        System.out.println("indep1 = " + indep1);
        System.out.println("vector1 = " + vector1);
        System.out.println("table1(indep1, vector1) = " + table1);
        System.out.println("  table1.getDimensions() = " + table1.getDimensions());
        System.out.println("  table1.size() = " + table1.size());
        System.out.println("  table1.getBreakpointValue(0, 3) = " + table1.getBreakpointValue(0, 3));
        System.out.println("table1.setBreakpoint(0, 3, 3.1)");
        table1.setBreakpoint(0, 3, 3.1);
        System.out.println("  table1.getBreakpoints(0) = " + table1.getBreakpoints(0));
        System.out.println("  table1.getAll() = " + table1.getAll());
        System.out.println("  table1.getValue(5) = " + table1.getValue(5));
        pos.clear();
        pos.add(Integer64.valueOf(5));
        System.out.println("pos = " + pos);
        System.out.println("table1.setValue(pos, 51.1)");
        table1.set(pos, 51.1);
        System.out.println("  table1.getAll() = " + table1.getAll());

        indep.clear();
        indep.add(Float64.valueOf(2));
        System.out.println("indep = " + indep);
        System.out.println("  table1.lookup(indep) = " + table1.lookup(indep));
        System.out.println("  table1.lookupValue(2.0) = " + table1.lookupValue(2.0));
        System.out.println("  table1.lookupValue(1.5) = " + table1.lookupValue(1.5));

        //  Test a 2D table.
        Float64Vector indep2 = Float64Vector.valueOf(100, 200, 300);
        List<Float64Vector> dep2D = FastTable.newInstance();
        dep2D.add(vector1);
        dep2D.add(Float64Vector.valueOf(0, 100, 200, 300, 400, 500, 600, 700, 800, 900, 1000));
        dep2D.add(Float64Vector.valueOf(0, 1000, 2000, 3000, 4000, 5000, 6000, 7000, 8000, 9000, 10000));
        System.out.println("\nindep1 = " + indep1);
        System.out.println("indep2 = " + indep2);
        System.out.println("dep2D = " + dep2D);
        List<String> indepNames = new FastTable<>();
        indepNames.add("indep2");
        indepNames.add("indep1");
        Float64Table table2 = Float64Table.valueOf(Text.valueOf("table2"), indepNames, indep2, indep1, dep2D);
        System.out.println("table2(indep2, indep1, dep2D) = " + table2);
        System.out.println("  table2.getDimensions() = " + table2.getDimensions());
        System.out.println("  table2.getNumBreakpoints() = " + table2.getNumBreakpoints());
        System.out.println("  table2.size() = " + table2.size());

        pos.clear();
        pos.add(Integer64.valueOf(1));
        pos.add(Integer64.valueOf(2));
        System.out.println("pos = " + pos);
        System.out.println("  table2.get(pos) = " + table2.get(pos));
        System.out.println("table2.setValue(pos, 251.1)");
        table2.set(pos, 251.1);
        System.out.println("  table2.getAll() = " + table2.getAll());

        indep.clear();
        indep.add(Float64.valueOf(200));
        indep.add(Float64.valueOf(1));
        System.out.println("indep = " + indep);
        System.out.println("  table2.lookup(indep) = " + table2.lookup(indep));

        indep.clear();
        indep.add(Float64.valueOf(150));
        indep.add(Float64.valueOf(0.5));
        System.out.println("indep = " + indep);
        System.out.println("  table2.lookup(indep) = " + table2.lookup(indep));

        //  Write out XML data.
        try {
            System.out.println();
            javolution.xml.XMLObjectWriter writer = javolution.xml.XMLObjectWriter.newInstance(System.out);
            writer.setIndentation("    ");
            writer.write(table1, "Float64Table", Float64Table.class);
            writer.flush();

        } catch (Exception e) {
            e.printStackTrace();
        }

    }
}
