/*
 *   NASAReader -- A class that can read and write NASA 2,3 or 4 argument multi-table files.
 *
 *   Copyright (C) 2000-2024, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.tables;

import jahuwaldt.io.FileUtils;
import static jahuwaldt.js.tables.FTableReader.RESOURCES;
import java.io.*;
import static java.text.MessageFormat.format;
import java.text.*;
import java.util.*;
import static java.util.Objects.requireNonNull;
import javolution.text.Text;
import javolution.util.FastTable;
import org.jscience.mathematics.number.Integer64;
import org.jscience.mathematics.vector.Float64Vector;

/**
 * A class that provides methods for reading or writing a NASA 2, 3, or 4 argument
 * multi-table file.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: April 3, 2000
 * @version January 15, 2024
 */
public class NASAReader implements FTableReader {

    //  The preferred file extension for files of this reader's type.
    private static final String EXTENSION = "nasa";

    private static final String TEN_SPACES = "          ";

    /**
     * Returns a string representation of the object. This will return a brief description
     * of the format read by this reader.
     */
    @Override
    public String toString() {
        return RESOURCES.getString("nasaDescription");
    }

    /**
     * Returns the preferred file extension (not including the ".") for files of this
     * reader's type.
     *
     * @return The file extension "nasa".
     */
    @Override
    public String getExtension() {
        return EXTENSION;
    }

    /**
     * Method that determines if this reader can read data from the specified input
     * stream. Will return FTableReader.MAYBE if the file name has the extension "nasa".
     *
     * @param name  The name of the file.
     * @param input An input stream containing the data to be read. Any methods that read
     *              from this stream must first set a mark and then reset back to that
     *              mark before the method returns (even if it returns with an exception).
     * @return FTableReader.NO if the file is not recognized at all or FTableReader.MAYBE
     *         if the filename extension is "*.nasa".
     */
    @Override
    public int canReadData(String name, BufferedInputStream input) throws IOException {
        requireNonNull(name, MessageFormat.format(RESOURCES.getString("paramNullErr"), "name"));
        
        int response = NO;
        String ext = FileUtils.getExtension(name);
        if (EXTENSION.equals(ext))
            response = MAYBE;

        return response;
    }

    /**
     * Returns true. This class can write data to a NASA multi-table formatted file.
     */
    @Override
    public boolean canWriteData() {
        return true;
    }

    /**
     * Compares this object with the specified object for order based on the
     * <code>toString().compareTo(o.toString())</code> method. Returns a negative integer,
     * zero, or a positive integer as this object is less than, equal to, or greater than
     * the specified object.
     */
    @Override
    public int compareTo(FTableReader o) {
        requireNonNull(o, MessageFormat.format(RESOURCES.getString("paramNullErr"), "o"));
        return this.toString().compareTo(o.toString());
    }

    /**
     * Method that reads in NASA 2,3, or 4 argument table formatted data from the
     * specified input stream and returns that data as an FTableList object.
     *
     * @param input An input stream containing the NASA formatted data.
     * @return An FTableList object that contains the data read in from the specified
     *         stream.
     * @throws IOException If there is a problem reading the specified stream.
     */
    @Override
    public FTableList read(InputStream input) throws IOException {
        requireNonNull(input, MessageFormat.format(RESOURCES.getString("paramNullErr"), "input"));
        
        //  NACA 2,3,4 files are required to be in ASCII with U.S. style number formatting.
        //  Get the default locale and save it.  Then change the default locale to US.
        Locale defLocale = Locale.getDefault();
        Locale.setDefault(Locale.US);

        try {
            // Get a reference to the file and create a stream reader.
            LineNumberReader reader = new LineNumberReader(new InputStreamReader(input));

            // Create an empty table database.
            FTableList db = FTableList.newInstance();

            // Read in the comment lines (if there are any) and add them
            // as database notes to the table database.
            readComments(reader, db);

            // Read in the tables in the file, one after the other.
            String aLine;
            do {
                Float64Table table = readNASATable(reader);
                db.add(table);

                // Is there more data?
                reader.mark(8192);
                aLine = reader.readLine();
                reader.reset();

            } while (aLine != null);

            return db;
            
        } finally {
            //  Restore the default locale.
            Locale.setDefault(defLocale);
        }
    }

    /**
     * Method for writing out all the data stored in the specified list of Float64Table
     * objects to the specified output stream in NASA 2,3, or 4 argument multi-table
     * format.
     *
     * @param output The output stream to which the data is to be written.
     * @param tables The list of Float64Table objects to be written out.
     * @throws IOException If there is a problem writing to the specified stream.
     */
    @Override
    public void write(OutputStream output, FTableList tables) throws IOException {
        requireNonNull(output, MessageFormat.format(RESOURCES.getString("paramNullErr"), "output"));
        requireNonNull(tables, MessageFormat.format(RESOURCES.getString("paramNullErr"), "tables"));
        if (tables.isEmpty())
            return;

        // Check for consistant tables.
        checkTables(tables);

        //  NACA 2,3,4 files are required to be in ASCII with U.S. style number formatting.
        //  Get the default locale and save it.  Then change the default locale to US.
        Locale defLocale = Locale.getDefault();
        Locale.setDefault(Locale.US);

        try {
            // Get a reference to the output file.
            BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(output));

            // Write out the tables.
            writeNASAFile(writer, tables);

            //  Flush the output.
            writer.flush();

        } finally {
            //  Restore the default locale.
            Locale.setDefault(defLocale);
        }
    }

    /**
     * Read in a single NASA 2, 3 or 4 argument table and return the result.
     *
     * @param in Reader for the file being read in.
     * @return A table containing the data read in from the file.
     */
    private static Float64Table readNASATable(LineNumberReader in) throws IOException {
        requireNonNull(in, MessageFormat.format(RESOURCES.getString("paramNullErr"), "in"));
        Float64Table table = null;
        try {
            // Skip over possible comments.
            String line0 = null;
            String line1 = in.readLine();
            String line2 = in.readLine();
            String line3 = in.readLine();
            while (line3 != null && !line3.trim().startsWith("&")) {
                line0 = line1;
                line1 = line2;
                line2 = line3;
                line3 = in.readLine();
            }
            if (line3 == null)
                throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));

            // Line 0 may contain the names of the independent variables.
            // Try to extract them if they are there.
            String[] indepNames = null;
            if (line0 != null) {

                // Independent names are surrounded by parentheses.
                int pos1 = line0.indexOf("(");
                int posn = line0.indexOf(")");
                if (pos1 >= 0 && posn > pos1) {
                    List list = new ArrayList();
                    String trimmedLine = line0.substring(pos1 + 1, posn);
                    pos1 = 0;
                    posn = trimmedLine.length();

                    // Independent names in the list are separated by commas.
                    while (pos1 < posn) {
                        int pos2 = trimmedLine.indexOf(",", pos1);
                        if (pos2 < 0)
                            pos2 = posn;

                        String subStr = trimmedLine.substring(pos1, pos2);
                        list.add(subStr.trim());
                        pos1 = pos2 + 1;
                    }

                    // Create an array of independent names.
                    int size = list.size();
                    indepNames = new String[size];
                    for (int i = 0; i < size; ++i)
                        indepNames[i] = (String)list.get(i);
                }
            }

            // Line 1 contains the table name (or number, but we'll use it as a name).
            String tblName = line1.trim();

            // Line 2 has the total number of elements in the table (which we ignore).
            // Read in 1st independent variable breakpoints.
            List indepVect = new ArrayList();
            DecimalFormat nf = (DecimalFormat)NumberFormat.getInstance();
            String aLine = in.readLine();
            if (aLine == null)
                throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));

            double[] bp1;
            int pos = 10;
            String subStr = aLine.substring(pos, pos + 9).trim();
            int numBP1 = (int)nf.parse(subStr).doubleValue();
            if (numBP1 > 1) {
                bp1 = new double[numBP1];
                indepVect.add(bp1);
                readNASA_BP(in, aLine, bp1, pos, nf);
            }

            // Read in the 2nd set of breakpoints.
            aLine = in.readLine();
            if (aLine == null)
                throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));

            pos = 10;
            subStr = aLine.substring(pos, pos + 9).trim();
            numBP1 = (int)nf.parse(subStr).doubleValue();
            if (numBP1 > 1) {
                bp1 = new double[numBP1];
                indepVect.add(bp1);
                readNASA_BP(in, aLine, bp1, pos, nf);
            }

            // Read in the 3rd & 4th set of breakpoints.
            // Mark this spot so we can return to it when we read in the dependent data.
            in.mark(8192);
            aLine = in.readLine();
            if (aLine == null)
                throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));

            pos = 10;
            subStr = aLine.substring(pos, pos + 9).trim();
            numBP1 = (int)nf.parse(subStr).doubleValue();
            pos += 10;
            subStr = aLine.substring(pos, pos + 9).trim();
            int numBP2 = (int)nf.parse(subStr).doubleValue();
            if (numBP1 < 2 || numBP2 < 2)
                throw new IOException(format(RESOURCES.getString("invNumBPErrMsg"), in.getLineNumber()));

            // Read in the 3rd & 4th breakpoint values and return the number
            // of lines that they take up.
            bp1 = new double[numBP1];
            double[] bp2 = new double[numBP2];
            indepVect.add(bp1);
            indepVect.add(bp2);
            int numBPLines = readNASA_BP2(in, aLine, bp1, bp2, pos, nf);

            // Create an array of breakpoint arrays.
            int numDims = indepVect.size();
            boolean haveNames = true;
            if (indepNames == null || indepNames.length != numDims) {
                indepNames = new String[numDims];
                haveNames = false;
            }
            double[][] breakpoints = new double[numDims][];
            for (int i = 0; i < numDims; ++i) {
                double[] bpArr = (double[])indepVect.get(i);
                breakpoints[i] = bpArr;
                if (haveNames == false)
                    indepNames[i] = "Indep" + (i + 1);
            }

            // Create an empty table with the breakpoints just read in.
            table = Float64Table.valueOf(tblName, indepNames, breakpoints, null);

            // Begin reading in the dependent data.
            in.reset();
            FastTable<Integer64> indep = FastTable.newInstance();
            for (int i = 0; i < numDims; ++i)
                indep.add(Integer64.ZERO);
            rReadNASA(0, in, numBPLines, indep, table, nf);

            FastTable.recycle(indep);

            // Read in &END line.
            in.readLine();

        } catch (ParseException | NumberFormatException e) {
            throw new IOException(format(RESOURCES.getString("parseErrMsg"), in.getLineNumber()), e);
        }

        return table;
    }

    /**
     * Recursively read in arrays of dependent data from a NASA 2,3, or 4 argument table
     * file into the parameter depData.
     *
     * @param dim        The dimension (independent) we are currently looping through.
     *                   Should initially be set to 0.
     * @param in         Reader for our table file stream.
     * @param numBPLines The number of lines taken up by the repeated breakpoints.
     * @param indep      A list of indexes into the data table (independents).
     * @param table      The table we are reading into.
     * @param nf         The number format object used to parse the numbers read in.
     */
    private static void rReadNASA(int dim, LineNumberReader in, int numBPLines,
            List<Integer64> indep, Float64Table table, NumberFormat nf)
            throws IOException, ParseException {
        if (dim >= indep.size() - 2) {
            // Reached the highest two dimensions, indep[] is now ready.
            // Read in the data.
            readNASADataArray(in, indep, table, numBPLines, nf);

        } else {

            // Loop over all the breakpoints in this dimension.
            int length = table.getNumBreakpoints(dim);
            for (int i = 0; i < length; ++i) {

                // Fill in the required value for indep[] at this dimension.
                indep.set(dim, Integer64.valueOf(i));       //  indep[dim] = i;

                // Go on to the next higher dimension.
                rReadNASA(dim + 1, in, numBPLines, indep, table, nf);
            }
        }
    }

    /**
     * Read in the dependent data associated with the last two breakpoints in a NASA 2,3,
     * or 4 argument table file.
     *
     * @param in         Reader for the input stream.
     * @param indep      A list of indexes into the data table (independents).
     * @param table      The table we are reading into.
     * @param numBPLines The number of lines taken up by the repeated breakpoints.
     * @param nf         The number format object used to parse the numbers read in.
     */
    private static void readNASADataArray(LineNumberReader in, List<Integer64> indep,
            Float64Table table, int numBPLines, NumberFormat nf)
            throws IOException, ParseException {
        // Skip over the repeated breakpoint lines.
        for (int i = 0; i < numBPLines; ++i)
            in.readLine();

        // Read in the 1st line.
        String aLine = in.readLine();
        if (aLine == null)
            throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));

        int pos = 0;
        int count = 0;
        int numDims = indep.size();
        int length1 = table.getNumBreakpoints(numDims - 2);
        int length2 = table.getNumBreakpoints(numDims - 1);

        // Loop over all the breakpoints in the last 2 indepenents
        // and read them in.
        for (int i = 0; i < length1; ++i) {
            indep.set(numDims - 2, Integer64.valueOf(i));                   //  indep[numDims - 2] = i;
            for (int j = 0; j < length2; ++j) {
                indep.set(numDims - 1, Integer64.valueOf(j));               //  indep[numDims - 1] = j;
                if (count == 6 || pos + 10 >= aLine.length()) {
                    aLine = in.readLine();
                    if (aLine == null)
                        throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));
                    pos = 0;
                    count = 0;
                }
                pos += 10;
                ++count;
                String subStr = aLine.substring(pos, pos + 9).trim();
                double value = nf.parse(subStr).doubleValue();
                table.set(indep, value);
            }
        }
    }

    /**
     * Read in a set of breakpoints for independent variables 3 & 4 (if they have any
     * breakpoints) in a NASA 2,3 or 4 argument table file.
     *
     * @param in    Reader for the input file stream.
     * @param aLine The current line being processed from the input file.
     * @param bpArr The array to contain the breakpoint values.
     * @param pos   The current position on the line "aLine".
     * @param nf    A number format for parsing the values as they are read in.
     */
    private static int readNASA_BP(LineNumberReader in, String aLine, double[] bpArr,
            int pos, NumberFormat nf) throws IOException, ParseException {
        int numBP = bpArr.length;
        int count = pos / 10;
        for (int i = 0; i < numBP; ++i) {
            if (count == 6) {
                aLine = in.readLine();
                if (aLine == null)
                    throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));

                pos = 0;
                count = 0;
            }
            pos += 10;
            ++count;
            String subStr = aLine.substring(pos, pos + 9).trim();
            double BP = nf.parse(subStr).doubleValue();
            bpArr[i] = BP;
        }

        return pos;
    }

    /**
     * Read in a set of breakpoints for independent variables 1 & 2 in a NASA 2,3 or 4
     * argument table file.
     *
     * @param in     Reader for the input file stream.
     * @param aLine  The current line being processed from the input file.
     * @param bpArr2 The array to contain the breakpoint values for indep #2.
     * @param bpArr1 The array to contain the breakpoint values for indep #1.
     * @param pos    The current position on the line "aLine".
     * @param nf     A number format for parsing the values as they are read in.
     */
    private static int readNASA_BP2(LineNumberReader in, String aLine,
            double[] bpArr2, double[] bpArr1, int pos, NumberFormat nf)
            throws IOException, ParseException {
        int numBP2 = bpArr2.length;
        int numBP1 = bpArr1.length;
        int count = pos / 10;
        int numLines = 1;
        for (int i = 0; i < numBP2; ++i) {
            if (count == 6) {
                aLine = in.readLine();
                if (aLine == null)
                    throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));

                pos = 0;
                count = 0;
                ++numLines;
            }
            pos += 10;
            ++count;
            String subStr = aLine.substring(pos, pos + 9).trim();
            double BP = nf.parse(subStr).doubleValue();
            bpArr2[i] = BP;
        }

        for (int i = 0; i < numBP1; ++i) {
            if (count == 6) {
                aLine = in.readLine();
                if (aLine == null)
                    throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));

                pos = 0;
                count = 0;
                ++numLines;
            }
            pos += 10;
            ++count;
            String subStr = aLine.substring(pos, pos + 9).trim();
            double BP = nf.parse(subStr).doubleValue();
            bpArr1[i] = BP;
        }

        return numLines;
    }

    /**
     * Read in the comment lines from a NASA 2, 3, or 4 argument table file and add them
     * as notes to the specified table database.
     *
     * @param tdb The table database to add the comments to.
     */
    private static void readComments(LineNumberReader in, FTableList tdb) throws IOException {
        if (tdb != null && in != null) {
            in.mark(8192);

            // Read in lines until we encounter an "&" character.
            String line1 = in.readLine();
            String line2 = in.readLine();
            String line3 = in.readLine();
            while (line3 != null && !line3.startsWith(" &") && !line3.startsWith("&")) {

                // Since line1 wasn't part of the NASA table structure,
                // it must be a comment.
                tdb.addNote(line1.trim());
                line1 = line2;
                line2 = line3;
                line3 = in.readLine();
            }
            if (line3 == null)
                throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));

            in.reset();
        }
    }

    /**
     * Check the tables that are to be written out to make sure that they have the
     * required number of independent variables. NASA tables must have between 2 and 4
     * independent variables. Also, each dimension must have at least 2 breakpoint values.
     *
     * @param tables A collection of tables that we are writing out.
     * @throws IOException if the tables do not have the required number of independent
     * variables.
     */
    private static void checkTables(FTableList tables) throws IOException {
        // Loop over all the tables in this collection and check each.
        for (Float64Table aTable : tables) {
            int dim = aTable.getDimensions();
            if (dim < 2 || dim > 4)
                throw new IOException(format(RESOURCES.getString("nasaIncDimErrMsg"), aTable.getName()));

            // Each dimension must have more than one breakpoint.
            for (int j = 0; j < dim; ++j) {
                int numBP = aTable.getNumBreakpoints(j);
                if (numBP < 2)
                    throw new IOException(format(RESOURCES.getString("nasaBPErrMsg"), aTable.getName()));
            }
        }
    }

    /**
     * Write out a collection of Float64Table objects as a NASA 2, 3, or 4 argument
     * multi-table file.
     *
     * @param output Reference to a file to write the tables to.
     * @param tables Collection of tables to be written to the file.
     */
    private static void writeNASAFile(BufferedWriter output, FTableList tables)
            throws IOException {
        DecimalFormat nf = (DecimalFormat)NumberFormat.getInstance();
        nf.setMinimumIntegerDigits(0);
        nf.setGroupingUsed(false);

        // Write out the database notes as table file comments (if there are any).
        writeFileComments(output, tables);

        // Loop over all the tables writing each one out in turn.
        int pos = 0;
        for (Iterator i = tables.iterator(); i.hasNext(); ++pos) {
            Float64Table theTable = (Float64Table)i.next();

            // Write out comment lines (1st already contains file comments).
            if (pos != 0) {
                output.newLine();
                output.newLine();
            }

            // Write out independent variables.
            output.write(theTable.toString());
            output.newLine();

            // Write out the table name.
            output.write(theTable.getName().toString());
            output.newLine();

            // Calculate and write out the number of data elements.
            String buffer = " " + calcNumElements(theTable);
            output.write(buffer);
            output.newLine();

            // Write out NASA card.
            output.write(" &TABLE DNASA=");
            output.newLine();

            // Write out fillers for non-existant independent variables.
            int numDim = theTable.getDimensions();
            switch (numDim) {
                case 2:
                    output.write("              1.000,    0.000,");
                    output.newLine();

                case 3:
                    output.write("              1.000,    0.000,");
                    output.newLine();

                default:
                    break;
            }

            // Set up the number format for writing out breakpoints.
            nf.setMaximumFractionDigits(3);
            nf.setMinimumFractionDigits(3);

            // Write out breakpoints for dimensions > 2.
            int dim = 0;
            while (dim < numDim - 2) {
                // Get the breakpoints for this dimension.
                Float64Vector breakpoints = theTable.getBreakpoints(dim);
                int numBP = breakpoints.getDimension();

                output.write(TEN_SPACES);

                // Write out the number of breakpoints in this dimension.
                output.write(formatNumber(10, nf, numBP));

                // Write out each breakpoint value.
                int count = 1;
                for (int j = 0; j < numBP; ++j) {
                    if (count == 6) {
                        output.newLine();
                        output.write(TEN_SPACES);
                        count = 0;
                    }
                    output.write(formatNumber(10, nf, breakpoints.getValue(j)));
                    ++count;
                }
                output.newLine();
                ++dim;
            }

            // Recursively write out all the arrays of dependent data.
            FastTable<Integer64> indep = FastTable.newInstance();
            for (int j = 0; j < numDim; ++j)
                indep.add(Integer64.ZERO);
            rWriteNASA(0, output, indep, theTable, nf);

            // Write out NASA END card.
            output.write(" &END");
            output.newLine();
        }
    }

    /**
     * Calculate the total number of elements in a NASA table including repeated and
     * non-repeated breakpoints.
     *
     * @param table The table we are writing out.
     * @return The total number of elements in the NADA table array (everything between
     *         the starting "&TABLE" and "&END").
     */
    private static int calcNumElements(Float64Table table) {

        int nElements = 1;
        int numDim = table.getDimensions();

        // Calculate the total number of dependent data points.
        for (int i = 0; i < numDim; ++i) {
            int num = table.getNumBreakpoints(i);
            nElements *= num;
        }

        int numBP1 = table.getNumBreakpoints(numDim - 2);
        int numBP2 = table.getNumBreakpoints(numDim - 1);

        // Add in the number of 3rd and 4th independent breakpoints (which arn't
        // repeated).
        int count = 1;
        for (int i = 0; i < numDim - 2; ++i) {
            int num = table.getNumBreakpoints(i);
            nElements += num + 1;
            count *= num;
        }

        // Add in the 1st & 2nd breakpoint lines (which are repeated).
        int num1p2 = numBP1 + numBP2 + 2;
        nElements += num1p2 * count;

        // Finally, add in the 1,0 elements for missing dimensions.
        while (numDim < 4) {
            nElements += 2;
            ++numDim;
        }

        return nElements;
    }

    /**
     * Recursively write out arrays of dependent data for a NASA 2,3, or 4 argument table
     * file into the specified stream.
     *
     * @param dim   The dimension (independent) we are currently looping through. Should
     *              initially be set to 0.
     * @param out   Writer for our table file stream.
     * @param indep A list of indexes into the data table (independents).
     * @param table The table we are writing out.
     * @param nf    The number format object used to format the numbers written out.
     */
    private static void rWriteNASA(int dim, BufferedWriter out, List<Integer64> indep,
            Float64Table table, NumberFormat nf) throws IOException {
        if (dim >= indep.size() - 2) {
            // Reached the highest 2 dimensions, indep[] is now ready.
            // Write out the data array.
            writeNASADataArray(out, indep, table, nf);

        } else {

            // Loop over all the breakpoints in this dimension.
            int length = table.getNumBreakpoints(dim);
            for (int i = 0; i < length; ++i) {
                // Fill in the required value for indep[] at this dimension.
                indep.set(dim, Integer64.valueOf(i));       //  indep[dim] = i;

                // Go on to the next higher dimension.
                rWriteNASA(dim + 1, out, indep, table, nf);
            }
        }
    }

    /**
     * Write out the dependent data associated with the last two breakpoints in a NASA
     * 2,3, or 4 argument table file.
     *
     * @param out   Writer for the output stream.
     * @param indep A list of indexes into the data table (independents).
     * @param table The table we are writing out to.
     * @param nf    The number format object used to format the numbers written out.
     */
    private static void writeNASADataArray(BufferedWriter out, List<Integer64> indep,
            Float64Table table, NumberFormat nf) throws IOException {
        // Write out the repeated breakpoint lines.
        writeBPLines(out, table, nf);

        // Setup the number format for formatting data array numbers.
        nf.setMaximumFractionDigits(5);
        nf.setMinimumFractionDigits(5);
        int count = 0;
        int numDims = indep.size();
        int length1 = table.getNumBreakpoints(numDims - 2);
        int length2 = table.getNumBreakpoints(numDims - 1);

        // Loop over all the breakpoints in the last 2 indepenents
        // and write the dependent data out.
        out.write(TEN_SPACES);
        for (int i = 0; i < length1; ++i) {
            indep.set(numDims - 2, Integer64.valueOf(i));           //  indep[numDims - 2] = i;
            for (int j = 0; j < length2; ++j) {
                indep.set(numDims - 1, Integer64.valueOf(j));       //  indep[numDims - 1] = j;
                if (count == 6) {
                    out.newLine();
                    out.write(TEN_SPACES);
                    count = 0;
                }
                ++count;
                out.write(formatNumber(10, nf, table.getValue(indep)));
            }
            out.newLine();
            if (i < length1 - 1)
                out.write(TEN_SPACES);
            count = 0;
        }
    }

    /**
     * Write out the repeated breakpoint lines for each data array.
     *
     * @param out   Writer used to write to the output stream.
     * @param table The table being written out.
     * @param nf    The number format used to format numbers for writing out.
     */
    private static void writeBPLines(BufferedWriter out, Float64Table table, NumberFormat nf)
            throws IOException {
        // Set up the number format for breakpoints.
        nf.setMaximumFractionDigits(3);
        nf.setMinimumFractionDigits(3);
        int numDim = table.getDimensions();

        // Extract the breakpoints for the last 2 dimensions.
        Float64Vector bp2 = table.getBreakpoints(numDim - 1);
        int numBP2 = bp2.getDimension();
        Float64Vector bp1 = table.getBreakpoints(numDim - 2);
        int numBP1 = bp1.getDimension();

        // Write out the number of breakpoints.
        out.write(TEN_SPACES);
        out.write(formatNumber(10, nf, numBP1));
        out.write(formatNumber(10, nf, numBP2));

        // Loop over the breakpoint arrays and write them out.
        int count = 2;
        for (int i = 0; i < numBP1; ++i) {
            if (count == 6) {
                out.newLine();
                out.write(TEN_SPACES);
                count = 0;
            }
            ++count;
            out.write(formatNumber(10, nf, bp1.getValue(i)));
        }
        for (int i = 0; i < numBP2; ++i) {
            if (count == 6) {
                out.newLine();
                out.write(TEN_SPACES);
                count = 0;
            }
            ++count;
            out.write(formatNumber(10, nf, bp2.getValue(i)));
        }
        out.newLine();
    }

    /**
     * Format a number for output to the NASA table file. Numbers in the NASA table file
     * are of a fixed size and include a comma separator. This method formats the given
     * number using the supplied NumberFormat object. It then adds a comma to the end of
     * the formatted number and finally adds spaces to the start of the formatted number
     * until it reaches the specified length.
     *
     * @param size   The overall size of the formatted number including the separating
     *               comma and decimal point.
     * @param nf     The NumberFormat to use when formatting this number.
     * @param number The number to be formatted.
     */
    private static String formatNumber(int size, NumberFormat nf, double number) {
        StringBuffer buffer = new StringBuffer(nf.format(number));
        buffer.append(",");
        TRUtils.addSpaces(size, buffer);
        return buffer.toString();
    }

    /**
     * Write out any database notes as table comments to the NASA file. A NASA table file
     * may have up to 3 lines of comments per table, but one line is, by convention,
     * reserved for a listing of the table independent variables.
     *
     * @param output A writer used to output the data.
     * @param tables The table database we are writing out.
     */
    private static void writeFileComments(BufferedWriter output, FTableList tables)
            throws IOException {
        int numNotes = tables.numberOfNotes();
        if (numNotes == 0) {

            // Write out some generic table comments into the header lines.
            output.write(tables.toString());
            output.newLine();
            output.write("Date:, ");
            Date theDate = new Date();
            output.write(DateFormat.getDateInstance(DateFormat.SHORT).format(theDate));
            output.write(", ");
            output.write(DateFormat.getTimeInstance().format(theDate));
            output.newLine();

        } else {
            // Write out the table notes as the header of the file.
            if (numNotes > 2)
                numNotes = 2;

            for (int i = 0; i < numNotes; ++i) {
                Text note = tables.getNote(i);
                note.print(output);
                output.newLine();
            }

            if (numNotes == 1) {
                Date theDate = new Date();
                output.write(DateFormat.getDateInstance(DateFormat.SHORT).format(theDate));
                output.write(", ");
                output.write(DateFormat.getTimeInstance().format(theDate));
                output.newLine();
            }
        }
    }

}
