/*
 *  sweepLEfn_xrle_xtle_b - Method for calculating LE sweep as a function of the X location of
 *       the root and tip chord leading edges and the span.
 *
 *  Copyright (C) 1998 - 2015 by Joseph A. Huwaldt
 *  All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */

import jahuwaldt.js.ParamTools.AbstractParamMethod;
import jahuwaldt.js.param.Parameter;
import javax.measure.quantity.Angle;
import javax.measure.unit.SI;

/**
 * Calculates leading edge wing sweep angle based on the X location of the root
 * and tip chord leading edges and the span.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: May 23, 2002
 * @version June 19, 2015
 */
public class sweepLEfn_xrle_xtle_b extends AbstractParamMethod<Angle> {

    /**
     * Constructor for this method.
     *
     * @param xrle Label for the "X coord of root chord leading edge" parameter.
     * @param xtle Label for the "X coord of tip chord leading edge" parameter.
     * @param b Label for the "wing span" parameter.
     */
    public sweepLEfn_xrle_xtle_b(String xrle, String xtle, String b) {
        _inputLabels = new String[3];
        _inputLabels[0] = xrle;
        _inputLabels[1] = xtle;
        _inputLabels[2] = b;
    }

    /**
     * Calculates leading edge wing sweep angle based on the X location of the
     * root and tip chord leading edges and the span. Method:
     *      sweepLE = atan((xtle-xrle)/b )
     *
     * @param inputs Array of input parameters. [0=X location of root chord LE (xrle),
     * 1=X location of tip chord LE (xtle), 2=wing span (b)].
     * @return Wing sweep angle at the leading edge in angular units.
     */
    @Override
    public Parameter<Angle> calc(Parameter[] inputs) {
        double xrle = inputs[0].doubleValue(SI.METER);
        double xtle = inputs[1].doubleValue(SI.METER);
        double b = inputs[2].doubleValue(SI.METER);
        double sweepLE = Math.atan((xtle - xrle) / b);
        return Parameter.valueOf(sweepLE, SI.RADIAN);
    }

}
